<?php
// $Header: /home/CVS/rps/includes/project/pps.inc,v 1.169 2008/05/19 01:59:00 yoxel Exp $
/*
 * ----START-LICENCE----
 * Copyright 2002-2008 Yoxel Systems, Santa Clara, CA, USA (info@yoxel.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * -----END-LICENCE-----
 */

require_once('includes/project/calendar.inc');
require_once('includes/params/user_filters.inc');

define('PROJ_OPT_MANUAL', 0x004);

define('PROJ_HISTORY_MAIN', 0x001);
define('PROJ_HISTORY_PLAN', 0x002);
define('PROJ_HISTORY_DECOMMIT', 0x003);
define('PROJ_HISTORY_HLINK', 0x004);

define('PROJ_MENU_MAIN', 0x001);
define('PROJ_MENU_HISTORY', 0x002);
define('PROJ_MENU_EST', 0x003);
define('PROJ_MENU_TS', 0x004);

define('WH_WASTE_IGNORE',0);
define('WH_WASTE_USE',1);
define('WH_WASTE_REGISTER',2);

class ProjectPlan {

	var $project_info=array();
	var $proj_id=0;
	var $log=array();
	var $rid2row=array();
	var $rid2dep=array();
	var $dep2rid=array();
	var $wcal; // Work Calendar
	var $inproj='';
	var $object_params;
	var $sql_done_states_str=array();
	var $class2str=array();
	var $history_buf='';

	function check_object_params(&$object_params){

		if(!isset($object_params['class_tag']))
			_die(__FILE__.':'.__LINE__." object parameter class_tag is not defined.");

		if(!isset($object_params['history_tbl']))
			_die(__FILE__.':'.__LINE__." object parameter history_tbl is not defined.");
		if(!isset($object_params['history_proj_decommit_type']))
			_die(__FILE__.':'.__LINE__." object parameter history_proj_decommit_type is not defined.");
		if(!isset($object_params['history_proj_plan_type']))
			_die(__FILE__.':'.__LINE__." object parameter history_proj_plan_type is not defined.");

		if(!isset($object_params['pnets_tbl']))
			_die(__FILE__.':'.__LINE__." object parameter pnets_tbl is not defined.");
		if(!isset($object_params['pnet_type_pri']))
			_die(__FILE__.':'.__LINE__." object parameter pnet_type_pri is not defined.");
		if(!isset($object_params['pnet_type_ts']))
			_die(__FILE__.':'.__LINE__." object parameter pnet_type_ts is not defined.");

		if(!isset($object_params['pq_tbl']))
			_die(__FILE__.':'.__LINE__." object parameter pq_tbl is not defined.");
		if(!isset($object_params['pq_fn']))
			_die(__FILE__.':'.__LINE__." object parameter pq_fn is not defined.");

		if(!isset($object_params['proj_id']))
			_die(__FILE__.':'.__LINE__." object parameter proj_id is not defined.");
		if(!isset($object_params['proj_tbl']))
			_die(__FILE__.':'.__LINE__." object parameter proj_tbl is not defined.");
		if(!isset($object_params['proj_class_mask']))
			_die(__FILE__.':'.__LINE__." object parameter proj_class_mask is not defined.");
		if(!isset($object_params['proj_deadline_fn']))
			_die(__FILE__.':'.__LINE__." object parameter proj_deadline_fn is not defined.");
		if(!isset($object_params['proj_id_fn']))
			_die(__FILE__.':'.__LINE__." object parameter proj_id_fn is not defined.");
		if(!isset($object_params['proj_mem_type']))
			_die(__FILE__.':'.__LINE__." object parameter proj_mem_type is not defined.");
		if(!isset($object_params['proj_history_tbl']))
			_die(__FILE__.':'.__LINE__." object parameter proj_history_tbl is not defined.");
		if(!isset($object_params['proj_title']))
			_die(__FILE__.':'.__LINE__." object parameter proj_title is not defined.");

		if(!isset($object_params['requests_tbl']))
			_die(__FILE__.':'.__LINE__." object parameter requests_tbl is not defined.");
		if(!isset($object_params['req_in_proj_fn']))
			_die(__FILE__.':'.__LINE__." object parameter req_in_proj_fn is not defined.");
		if(!isset($object_params['req_proj_id_fn']))
			_die(__FILE__.':'.__LINE__." object parameter req_proj_id_fn is not defined.");
		if(!isset($object_params['req_proj_pri_fn']))
			_die(__FILE__.':'.__LINE__." object parameter req_proj_pri_fn is not defined.");
		if(!isset($object_params['req_classes']))
			_die(__FILE__.':'.__LINE__." object parameter req_classes is not defined.");
		if(!isset($object_params['req_last_state_mask']))
			_die(__FILE__.':'.__LINE__." object parameter req_last_state_mask is not defined.");

		$c=count($object_params['pm_phases']);
		if(($c!=1) && ($c!=2))
			_die(__FILE__.':'.__LINE__." Currently supporting only 1 and 2 phase projects.");

		$i=0;
		foreach($object_params['pm_phases'] as $phase){
			$i++;
			if(!isset($phase['proj_state']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter proj_state is not defined.");
			if(!isset($phase['mgr_fn']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter mgr_fn is not defined.");
			if(!isset($phase['implementer_fn']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter implementer_fn is not defined.");
			if(!isset($phase['implementer_spec']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter implementer_spec is not defined.");
			if(!isset($phase['dflt_fix_fn']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter dflt_fix_fn is not defined.");
			if(!isset($phase['start_state_mask']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter start_state_mask is not defined.");
			if(!isset($phase['done_state_mask']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter done_state_mask is not defined.");
			if(!isset($phase['plan_start_fn']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter plan_start_fn is not defined.");
			if(!isset($phase['plan_done_fn']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter plan_done_fn is not defined.");
			if(!isset($phase['start_fn']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter plan_start_fn is not defined.");
			if(!isset($phase['done_fn']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter plan_done_fn is not defined.");
			if(!isset($phase['pnet_type']))
                        	_die(__FILE__.':'.__LINE__." phase$i object parameter pnet_type is not defined.");
		}

	}

	function projectplan(&$object_params, $full_mode=TRUE, $ts_overlap_mode=FALSE){

		$this->check_object_params($object_params);
		$this->object_params=$object_params;
		$proj_id=$this->proj_id=$object_params['proj_id'];
		if(!$proj_id)
			return;

		$proj_tbl=$object_params['proj_tbl'];
		$proj_id_fn=$object_params['proj_id_fn'];
		$proj_cm=$object_params['proj_class_mask'];
		$pq_fn=$object_params['pq_fn'];
		$req_classes=$object_params['req_classes'];
		$req_last_sm=$object_params['req_last_state_mask'];
		if(!isset($object_params['proj_success_phase']))
			$psp=count($object_params['pm_phases'])-1;
		else
			$psp=$object_params['proj_success_phase'];
		$this->object_params['req_success_state_mask']=&$this->object_params['pm_phases'][$psp]['done_state_mask'];
		$req_success_sm=&$this->object_params['req_success_state_mask'];

		$query = "SELECT * FROM $proj_tbl WHERE $proj_id_fn=$proj_id";
                if(!($result=@mysql_query($query)))
                        _die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl.");

                if(mysql_num_rows($result)<1)
			_quit_error("Wrong proj_id=$proj_id.");

		$this->project_info=mysql_fetch_array($result, MYSQL_ASSOC);
		if($this->project_info['state']=='plan')
			$this->inproj='plan';
		else
			$this->inproj='yes';

		if(isset($this->project_info['label']) && $this->project_info['label']){
                        $lbl=$this->project_info['label'];
                        $pqid=$this->project_info[$pq_fn];
                        $pqname=$this->rts->pqs_2d['name'][$pqid];
                        $_SESSION['pps_proj_title']="$pqname '$lbl'";
                        $_SESSION['pps_proj_url_param']='?proj_id='.$proj_id;
                }

                $query = "SELECT * FROM $proj_tbl WHERE $pq_fn=".$this->project_info[$pq_fn].($proj_id?" AND $proj_id_fn!=".$proj_id:'')." AND state!='done' ORDER BY $proj_id_fn DESC LIMIT 1";
                //echo "$query <br>\n";
                if(!($result=@mysql_query($query)))
                        _die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl.");
                if(mysql_num_rows($result)>0)
                        $this->prev_project_info=mysql_fetch_array($result, MYSQL_ASSOC);


		$this->wcal=new Calendar();

		$sql=array();$success_sql=array();
		foreach($req_classes['options'] as $c => $o){
			if(!($c&$proj_cm)) continue;
			if($this->project_info['state']=='plan'){
				array_push($sql,"(r.class=$c AND (r.state&".$req_last_sm[$c].")=0)");
				$this->class2str[$c]=$req_classes['name'][$c];
			}
			array_push($success_sql,"(r.class=$c AND r.state&".$req_success_sm[$c].")");
		}
		if(count($sql)<1)
			$this->sql_states_str='1';
		elseif(count($sql)<2)
			$this->sql_states_str=$sql[0];
		else
			$this->sql_states_str="(".implode(' OR ',$sql).")";

		if(count($success_sql)<1)
			$this->sql_success_states_str='1';
		elseif(count($success_sql)<2)
			$this->sql_success_states_str=$success_sql[0];
		else
			$this->sql_success_states_str="(".implode(' OR ',$success_sql).")";
		//echo $this->sql_success_states_str."<br>\n";

		foreach($object_params['pm_phases'] as $phase){
			$ph_done_sql=array();
			foreach($req_classes['options'] as $c => $o){
				if(!($c&$proj_cm)) continue;
				array_push($ph_done_sql,"(r.class=$c AND r.state&".$phase['done_state_mask'][$c].")");
			}
			if(count($ph_done_sql)<1)
				$this->sql_done_states_str[]='0';
			elseif(count($ph_done_sql)<2)
				$this->sql_done_states_str[]=$ph_done_sql[0];
			else
				$this->sql_done_states_str[]="(".implode(' OR ',$ph_done_sql).")";
		}
		$this->init_dep_hash();
		$this->init_proj_members($full_mode);
		$this->init_timeshare($ts_overlap_mode);
	}

	function get_proj_title(){
		return $this->object_params['proj_title'];
	}

	function proj_lbl_exists($proj_lbl,$pqid=FALSE,$except_proj_id=FALSE){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_fn=$this->object_params['pq_fn'];

		$query = "SELECT $proj_id_fn FROM $proj_tbl WHERE label='".escape_data($proj_lbl)."'".($pqid?" AND $pq_fn=$pqid":'').($except_proj_id?" AND $proj_id_fn!=$except_proj_id":'');
		//echo "$query <br>\n";
		if(!($result= @mysql_query($query)))
			die2(__FILE__, __LINE__," Could not select from $proj_tbl table.");

		if(mysql_num_rows($result)>0)
			return TRUE;

		return FALSE;
	}

	function proj_lbl_add($proj_lbl,$pqid){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_fn=$this->object_params['pq_fn'];
		$proj_rts_opt=$this->rts->get_proj_rts_opt();

		$query = "INSERT INTO $proj_tbl ($pq_fn, state, label,  done_date, options) VALUES ($pqid, 'done', '".escape_data($proj_lbl)."', NOW(), $proj_rts_opt)";
		if(!@mysql_query ($query))
			_die2(__FILE__, __LINE__," Could not insert into $proj_tbl table.");
		return mysql_insert_id();
	}

	function is_rts_proj($proj_id){
		$proj_rts_opt=$this->rts->get_proj_rts_opt();
		return $this->projects_2d['row'][$proj_id]['options']&$proj_rts_opt;
	}

	function proj_has_requests($proj_id){
		$req_tbl=$this->object_params['requests_tbl'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];

		return simple_query("SELECT COUNT(rid) FROM $req_tbl WHERE $req_proj_id_fn=$proj_id");
	}

	function proj_rts_remove($proj_id){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_rts_opt=$this->rts->get_proj_rts_opt();

		$query="DELETE FROM $proj_tbl WHERE options&$proj_rts_opt AND $proj_id_fn=$proj_id LIMIT 1";
		if(!@mysql_query($query))
			die2(__FILE__.':'.__LINE__." Could not delete from $proj_tbl table. ");

	}

	function proj_rts_update($proj_id,$proj_lbl,$pqid=FALSE){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_fn=$this->object_params['pq_fn'];
		$proj_rts_opt=$this->rts->get_proj_rts_opt();

		$query = "UPDATE $proj_tbl SET label='".escape_data($proj_lbl)."'".($pqid?", $pq_fn=$pqid":'')." WHERE $proj_id_fn=$proj_id AND options&$proj_rts_opt LIMIT 1";
		//echo "$query <br>";
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not update  $proj_tbl table.");

	}

	function &cc_list(){
		$ccl=$this->proj_members;
		$other_cc=&$this->rts->proj_cc_list($this->proj_id);
		foreach(array_keys($other_cc) as $cck)
			arr_push($ccl,$cck);

		return $ccl;
	}

	function add_cc($u){
		$this->rts->proj_add_cc($this->proj_id,$u);
	}

	function delete_cc($u){
		$this->rts->proj_delete_cc($this->proj_id,$u);
	}

	var $projects_2d=array();
	function &init_projects($pqid = FALSE, $all = FALSE,  $with_rts = TRUE){

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_fn=$this->object_params['pq_fn'];
		$proj_rts_opt=$this->rts->get_proj_rts_opt();

		$ref=&$this->projects_2d;
		$this->projects_2d['lbl']=array();
		$this->projects_2d['state']=array();
		$this->projects_2d['pqid']=array();

		$query = "SELECT $proj_id_fn, $pq_fn, state, label, options, CONCAT(MONTH(plan_start_date),'/',DAYOFMONTH(plan_start_date),'/',YEAR(plan_start_date)) AS plan_start_date, CONCAT(MONTH(implement_start_date),'/',DAYOFMONTH(implement_start_date),'/',YEAR(implement_start_date)) AS ph1_start_date".($this->get_phase_num()>1?", CONCAT(MONTH(test_start_date),'/',DAYOFMONTH(test_start_date),'/',YEAR(test_start_date)) AS ph2_start_date":'').", CONCAT(MONTH(done_date),'/',DAYOFMONTH(done_date),'/',YEAR(done_date)) AS done_date FROM $proj_tbl WHERE 1";

		if($pqid)
			$query .= " AND $pq_fn=$pqid".($all?'':" AND state='done'");
		else
			$query .= ($all?'':" AND state='done'");

		if(!$with_rts)
			$query .= " AND !(options&$proj_rts_opt)";

		$query .=" ORDER BY $proj_id_fn DESC";
		//echo "$query <br>\n";

		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl table.");

		while($row=mysql_fetch_array($result, MYSQL_ASSOC)){
			if(!($rel_id=$row[$proj_id_fn]))
				continue;

			//print_r($row);echo "<br>\n";
			$this->projects_2d['row'][$rel_id]=$row;
			$this->projects_2d['lbl'][$rel_id]=$row['label'];
			$this->projects_2d['state'][$rel_id]=$row['state'];
			$this->projects_2d['pqid'][$rel_id]=$row[$pq_fn];
			$this->projects_2d['done_date'][$rel_id]=$row['done_date'];
		}
		asort($this->projects_2d['lbl']);
		//krsort($this->projects_2d['state']);

		if(count($this->projects_2d['state'])<1)
			prereq_check_set('pps_no_projects');

		if(!in_array('plan',$this->projects_2d['state']))
			prereq_check_set('pps_no_projects_in_plan');
		else
			prereq_check_set('pps_project_in_plan');

		$this->projects_2d['qid']=&$this->projects_2d['pqid'];
		$this->projects_2d['pid']=&$this->projects_2d['pqid'];
		return $ref;
	}

	function get_phase_num(){
		return count($this->object_params['pm_phases']);
	}

	function get_phase_info($phnum,$tag){
		if(($phnum<1) || ($phnum>$this->get_phase_num()))
			_die("Incorrect phase number specified: $phnum.");
		if(!isset($this->object_params['pm_phases'][$phnum-1][$tag]))
			_die("Unknown phase info tag: $tag.");
		return $this->object_params['pm_phases'][$phnum-1][$tag];
	}

	function find_last_project($pqid){

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$proj_cm=$this->object_params['proj_class_mask'];
		$pq_fn=$this->object_params['pq_fn'];

		$sql_sel=array(); $ph=0;
		foreach($this->object_params['pm_phases'] as $phase){
			$ph++;
			$ph_dflt_fn=$phase['dflt_fix_fn'];
			$sql_sel[]="$ph_dflt_fn AS ph{$ph}_dflt";
		}
		$sql_str=implode(', ',$sql_sel);

		$query = "SELECT $proj_id_fn, state, label, done_date, $proj_deadline_fn AS proj_deadline, plan_start_date, good_fix_rate ".($ph>1?', good_test_rate':'').", estimated_fix_end ".($ph>1?', estimated_test_end':'').", ct_pct, avg_fix_hrs ".($ph>1?', avg_test_hrs':'').", $sql_str FROM $proj_tbl WHERE $pq_fn=$pqid ORDER BY $proj_id_fn DESC LIMIT 10";

		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__.' Could not select from releases.');

		$fn=$tn=$ft=$tt=$n=0;
		$rel_info=array();
		while($row=mysql_fetch_array($result, MYSQL_ASSOC)){
			if($row['state'] == 'plan' )
				return(FALSE);
			if($n==0){
				$rel_info=$row;
				if(($row['state'] == 'implement'))
				// Makes sense for implemented or done releases
					$rel_info['ct_pct']=0;
			}
			if(isset($row['ph1_dflt']) && $row['ph1_dflt']){
				$ft+=$row['ph1_dflt'];
				$fn++;
			}
			if(isset($row['ph2_dflt']) && $row['ph2_dflt']){
				$tt+=$row['ph2_dflt'];
				$tn++;
			}
			$n++;
		}
		if(count($rel_info)<1){
			$rel_info['ph1_dflt']=8;
			if($ph>1)
				$rel_info['ph2_dflt']=8;
			$rel_info['ct_pct']=0;
		}

		if($fn){
			$h=$ft%$fn;
			$rel_info['ph1_dflt']=($ft-$h)/$fn;
		}
		if($tn){
			$h=$tt%$tn;
			$rel_info['ph2_dflt']=($tt-$h)/$tn;
		}
		return $rel_info;
	}

	function num_of_planable_requests($pqid){
		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$req_tbl=$this->object_params['requests_tbl'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_last_sm=&$this->object_params['req_last_state_mask'];

		$lsql=array();
		foreach($req_last_sm as $c => $m)
			array_push($lsql,"(r.class=$c AND r.state&$m)");
		$lsql_str="(".implode(' OR ',$lsql).")";

		$proj_ids_str=$this->valid_proj_ids_str($pqid);

		$query= "SELECT COUNT(rid) FROM $req_tbl AS r INNER JOIN $pq_tbl AS p ON r.$pq_fn=p.$pq_fn WHERE (r.$pq_fn=$pqid OR p.parent=$pqid) ".($proj_ids_str?" AND (r.$req_in_proj_fn!='yes' OR !(r.$req_proj_id_fn IN ($proj_ids_str)))":'')." AND !($lsql_str) AND r.customer_pri!=".REQ_PRI_CRITICAL;
		//echo "$query <br>\n";
		return(simple_query($query));
	}

	function project_exists($pqid,$lbl){

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_fn=$this->object_params['pq_fn'];

		$lbl = escape_data($lbl);
		$query = "SELECT $proj_id_fn FROM $proj_tbl WHERE $pq_fn=$pqid AND label='$lbl'";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl.");
		return(mysql_num_rows($result));
	}

	function prev_proj_deadline($pqid){

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$proj_rts_opt=$this->rts->get_proj_rts_opt();
		$pq_fn=$this->object_params['pq_fn'];


		return(simple_query("SELECT IF(state='done',done_date,$proj_deadline_fn) FROM $proj_tbl WHERE $pq_fn=$pqid AND (options&$proj_rts_opt)=0 LIMIT 1"));
	}

	function add_new_project($pqid,$lbl,$descr,$rld,$ct_pct,$mems,$dft=8,$dtt=8){
		global $uid;

		if(count($mems)<1)
			_die(__FILE__.':'.__LINE__." No members specified.");

                $req_classes=$this->object_params['req_classes'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$proj_cm=$this->object_params['proj_class_mask'];
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$proj_mem_type=$this->object_params['proj_mem_type'];
		$pq_fn=$this->object_params['pq_fn'];
		$ph=count($this->object_params['pm_phases']);

		$lbl=escape_data($lbl);
		$descr=escape_data($descr);
		$rld=escape_data($rld);
		$ct_pct=escape_data($ct_pct);
		if($dft===FALSE)
			$dft=8;
		if($dtt===FALSE)
			$dtt=8;

		$query = "INSERT INTO $proj_tbl ($pq_fn, state, label, description, $proj_deadline_fn, added_by, ct_pct, plan_start_date, dflt_fix_time  ".($ph>1?', dflt_test_time':'').") VALUES ($pqid, 'plan', '$lbl', '$descr', '$rld', $uid, '$ct_pct', NOW(), $dft".($ph>1?", $dtt":'').")";
		//echo "$query <br>\n";
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not insert into $proj_tbl table.");
		$proj_id=mysql_insert_id();

		if(!in_array($uid,$mems))
			array_push($mems,$uid);

		$ins_sql_str='';
		foreach ($mems as $v)
			$ins_sql_str.=", ($proj_id, $v, $proj_mem_type)";

		$query = "INSERT INTO members (xid, uid, type) VALUES ".substr($ins_sql_str,2);
		//echo "$query <br>\n";
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__.' Could not register project members.');

		$filters=new UserFilters($this->rts->wfid());
		$fid=$filters->create_new_filter("__project_requests $lbl");
		$rf=$filters->get_filter_by_id($fid);
		$cstr='';
		foreach($req_classes['options'] as $c => $o){
                        if(!($c&$proj_cm))
				continue;
			$cstr.=",$c";
		}
		$rf->pset('class',substr($cstr,1));
		$rf->pset($req_proj_id_fn,$proj_id);
		$rf->pset($req_proj_id_fn.'_extra_sql',"r.$req_in_proj_fn=IF(proj.state='plan','plan','yes')");
		$rf->pset('_type_select',"class,$req_proj_id_fn");
		$rf->pset('_proj_tag',$proj_id);
		$rf->save();

		if(!$this->rts->have_extra())
			return $proj_id;

		$this->rts->x_add_proj_templates($proj_id,$lbl,$fid);
		return $proj_id;
	}

	function init_dep_hash(){
		$pq_tbl=$this->object_params['pq_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_fn=$this->object_params['pq_fn'];
		$req_tbl=$this->object_params['requests_tbl'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_success_sm=$this->object_params['req_success_state_mask'];
		$req_last_sm=$this->object_params['req_last_state_mask'];
		$ph1_fixer_fn=$this->object_params['pm_phases'][0]['implementer_fn'];
		$ph_two=count($this->object_params['pm_phases'])>1;
		if($ph_two)
			$ph2_fixer_fn=$this->object_params['pm_phases'][1]['implementer_fn'];

		if(count($this->rid2dep)) return;

		$reread=FALSE;
		$looping=0;
		// Do this loop to include all blockers into the project
		do {
			$this->rid2dep=array();
			$this->dep2rid=array();
		   	if($looping++>10)
		   		_die(__FILE__.':'.__LINE__.' Loop detected ...');

			$query="SELECT r.rid, r.depends_on, r.phase_of_rid, IF(r.$req_proj_id_fn=".$this->project_info[$proj_id_fn]." AND r.$req_in_proj_fn='".$this->inproj."',1,0), d.class, d.state, IF((d.$req_in_proj_fn='plan' OR d.$req_in_proj_fn='yes') AND d.$req_proj_id_fn,1,0), d.$req_proj_id_fn FROM $req_tbl AS r LEFT JOIN $req_tbl AS d ON (r.depends_on=d.rid AND d.customer_pri!=".REQ_PRI_CRITICAL.") LEFT JOIN $pq_tbl as q ON r.$pq_fn=q.$pq_fn WHERE (r.$pq_fn=".$this->project_info[$pq_fn]." OR q.parent=".$this->project_info[$pq_fn].") AND (r.depends_on!=0 OR r.phase_of_rid!=0) AND ".$this->sql_states_str;
			//echo "$query <br>\n";
			if(!($result=@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
			$reread=FALSE;
			$add_parent_rid=FALSE;
			while($row=mysql_fetch_row($result)){
				//echo $row[0]." (in plan =".$row[3]."), depends on ".$row[1]."<br>\n";
				if($row[1] && !$row[4]){
				// Sanity check that blocker really exists
					simple_query("UPDATE $req_tbl SET depends_on=0 WHERE rid=".$row[0]);
					$row[1]=0;
				}
				if($row[1] && $row[3]){
				// Do consistency check
					//echo "row: "; print_r($row); echo "<br>\n";
					$req_closed=($row[5]&$req_success_sm[$row[4]])&$req_last_sm[$row[4]];

					// Blocker is in a future project and not closed
					if($row[6] && ($this->project_info[$proj_id_fn]<$row[7]) && !$req_closed)
						_die(__FILE__.':'.__LINE__.' Blocker RID'.$row[1].' for RID'.$row[0].' is in a future project.');

					// Blocker is not in any planned project and not closed
					if(!$row[6] && !$req_closed){
						// Add the blocker to the current project
						$query="SELECT rid, IF((($req_in_proj_fn='plan' OR $req_in_proj_fn='yes') AND $req_proj_id_fn),1,0), $req_proj_id_fn, class, state FROM $req_tbl WHERE phase_of_rid=$row[1] LIMIT 500";
						//echo "$query <br>\n";
						if(!($SQL=@mysql_query($query)))
							_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

						$update_sql_str="$req_in_proj_fn='".$this->inproj."', $req_proj_id_fn=".$this->project_info[$proj_id_fn].", plan_$ph1_fixer_fn=$ph1_fixer_fn".($ph_two?", plan_$ph2_fixer_fn=$ph2_fixer_fn":'');
						// Do this loop in case it has sub-requests
						for($num=0;$r=mysql_fetch_row($SQL);$num++){
							// Add all phases
							$cl=($r[4]&$req_success_sm[$r[3]])&$req_last_sm[$r[3]];
							if(($r[1] && ($this->project_info[$proj_id_fn]>=$r[2])) || $cl) continue;
							_warning(' Blocker RID'.$row[0].' for RID'.$row[0].' does not belong to any project yet.');
							_warning(' Adding it to the current project automatically.');
							$query="UPDATE $req_tbl SET $update_sql_str WHERE rid=$r[0]";
							//echo "$query <br>\n";
							if(!@mysql_query($query))
								_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
							$reread=TRUE;
						}

						if(!$num){
							_warning(' Blocker RID'.$row[1].' for RID'.$row[0].' does not belong to any project yet.');
							_warning(' Adding it to the current project automatically.');
							$query="UPDATE $req_tbl SET $update_sql_str WHERE rid=$row[1]";
							//echo "$query <br>\n";
							if(!@mysql_query($query))
								_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
							$reread=TRUE;
							break;
						}elseif($reread)
							break;
					}

					if($row[1]==$row[2])
						$row[1]=0;

					if($req_closed)
						$row[1]=0;
				}
				$this->rid2dep[$row[0]]=array($row[1],$row[2]);
				if($row[1]){
					if(!isset($this->dep2rid[$row[1]]))
						$this->dep2rid[$row[1]]=array();
					array_push($this->dep2rid[$row[1]],$row[0]);
				}
			}
			mysql_free_result($result);
		} while($reread);
	}

	function get_dependency_list($rid){
		$this->init_dep_hash();
		//echo "*** rid2dep:";print_r($this->rid2dep);echo "<br>\n";
		// figure out dependency list here
		$loop=0;
		$depends_on=array();
		$l=array($rid);
		while($r=array_shift($l)){
			if(!isset($this->rid2dep[$r][0])) continue;
			if(!($dep=$this->rid2dep[$r][0])) continue;
			array_push($depends_on,$dep);array_push($l,$dep);
			foreach($this->rid2dep as $k => $v){
				if($dep==$v[1]){
					array_push($depends_on,$k);
					array_push($l,$k);
				}
			}
			if($loop++>100000)
				_die(__FILE__.':'.__LINE__." Seems like a loop.");;
		}
		//echo "*** $rid depends on:";print_r($depends_on);echo "<br>\n";
		return($depends_on);
	}

	function get_blocking_list($rid){
		$this->init_dep_hash();
		//echo "*** dep2rid:";print_r($this->dep2rid);echo "<br>\n";
		// figure out dependency list here
		$blocking=array();
		$l=array($rid);
		while($b=array_shift($l)){
			if(!isset($this->dep2rid[$b])) continue;
			foreach($this->dep2rid[$b] as $junk => $r){
				array_push($blocking,$r);array_push($l,$r);
				foreach($this->rid2dep as $k => $v){
					if($r==$v[1]){
						array_push($blocking,$k);
						array_push($l,$k);
					}
				}
			}
		}
		//echo "<br>\n*** $rid depends on:";print_r($depends_on);echo "<br>\n";
		return($blocking);
	}

	function gen_sql_for_booked_times(){

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$pqid=$this->project_info[$pq_fn];
		$proj_id=$this->proj_id;

		$ph1_fixer_fn=$this->object_params['pm_phases'][0]['implementer_fn'];
		$done_sm=&$this->object_params['pm_phases'][0]['done_state_mask'];
		$dsql=array();
		foreach(array_keys($this->rts->classes['options']) as $c)
			array_push($dsql,"(r.class=$c AND r.state&".$done_sm[$c].")");

		$ph1dsql_str="(".implode(' OR ',$dsql).") AS ph1_is_done";

		$ph2=FALSE;
		if(count($this->object_params['pm_phases'])>1){
			$ph2=TRUE;
			$ph2_fixer_fn=$this->object_params['pm_phases'][1]['implementer_fn'];
			$done_sm=&$this->object_params['pm_phases'][1]['done_state_mask'];
		}

		$dsql=array();
		foreach(array_keys($this->rts->classes['options']) as $c)
			array_push($dsql,"(r.class=$c AND r.state&".$done_sm[$c].")");

		$dsql_str="(".implode(' OR ',$dsql).")";

		$query="SELECT r.$ph1_fixer_fn AS ph1_fixer, UNIX_TIMESTAMP(r.cal_fix_start) AS ph1_start, UNIX_TIMESTAMP(r.cal_fix_end) AS ph1_end, $ph1dsql_str,".
		($ph2?" r.$ph2_fixer_fn AS ph2_fixer, UNIX_TIMESTAMP(r.cal_test_start) AS ph2_start, UNIX_TIMESTAMP(r.cal_test_end) AS ph2_end,":'')." r.rid FROM $proj_tbl AS rl, $req_tbl AS r WHERE !$dsql_str AND rl.$pq_fn=$pqid AND (rl.state='implement' OR rl.state='test') AND rl.$proj_id_fn=r.$req_proj_id_fn AND r.$req_in_proj_fn='yes' AND rl.$proj_id_fn<$proj_id";

		return($query);
	}

	function init($start_date=FALSE, $end_date=FALSE){
		global $uid;

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$pq_fn=$this->object_params['pq_fn'];

		$proj_id=$this->project_info[$proj_id_fn];
		$proj_in_plan=$this->project_info['state']=='plan';
		$fix_and_test=count($this->object_params['pm_phases'])>1;

		$ok=TRUE;

		$start_ch=$this->wcal->date2ch($start_date);
		if(!$end_date){
			$end_ch=$this->wcal->date2ch(date_sql2norm($this->project_info[$proj_deadline_fn]));
		}else
			$end_ch=$this->wcal->date2ch($end_date);

		//echo $this->wcal->ch2date("m/j/Y",$start_ch).
		//	" ($end_date) end: ".$this->wcal->ch2date("m/j/Y",$end_ch)." <br>\n";
		if($start_ch>$end_ch){
			if($proj_in_plan)
				_warning('Warning: Project deadline is before its start date. Please correct.');
			$end_ch=$start_ch;
			$ok=FALSE;
		}

		$eday=intval($end_ch/HOURS_IN_DAY);
		$sday=intval($start_ch/HOURS_IN_DAY);
		if($eday>($sday+$this->wcal->dlimit())){
			_warning("Warning: Project periods over ".$this->wcal->dlimit()." days are discouraged (".$this->wcal->ch2date("m/j/Y",$start_ch).','.$this->wcal->ch2date("m/j/Y",$end_ch).'). Please correct.');

			$end_date='';
			$ok=FALSE;
			$end_ch=$this->wcal->limit_ch($start_ch);
		}

		if(!$ok){
			$npd=$this->wcal->ch2date("Y-m-j",$end_ch);
			$this->project_info[$proj_deadline_fn]=$npd;
			$npd=date_sql2norm($npd);
			if($proj_in_plan)
				_warning('Warning: Assuming '.$npd.' for current calculations.');
		}

		$wcal_users=array();
		if($fix_and_test){
			$this->init_testers();
			foreach(array_keys($this->all_testers) as $k)
				$wcal_users[$k]=TRUE;
		}

		$this->init_fixers(TRUE);
		foreach(array_keys($this->all_fixers) as $k)
			$wcal_users[$k]=TRUE;

		if(count($wcal_users)==0)
			array_push($wcal_users,$uid);

		$SQL=$this->gen_sql_for_booked_times();
		$this->wcal->init($start_ch,$end_ch,array_keys($wcal_users),$SQL);

		$this->project_info['proj_deadline_ch']=$this->wcal->date2ch(date_sql2norm($this->project_info[$proj_deadline_fn]));

		$this->init_hps();
		$this->init_ets();
		//echo "init <br>";
		return($ok);
	}

	var $proj_members=array();
	var $mem_pri=array();
	var $members_vote=array();
	// mode:
	//		FALSE - initialize proj_members array only
	//		TRUE - initialize both arrays
	function init_proj_members($full_mode=TRUE){
		//echo "init_proj_members: $full_mode <br>\n";

		$req_tbl=$this->object_params['requests_tbl'];
		$proj_mem_type=$this->object_params['proj_mem_type']; //MEMB_TYPE_ITPROJ
		$pnets_tbl=$this->object_params['pnets_tbl'];
		$pnet_type_pri=$this->object_params['pnet_type_pri'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

	//	echo "In init_proj_members: $this->proj_id, $full_mode <br>";

		if(count($this->proj_members) && !$full_mode)
			return;

		$query = "SELECT uid FROM members WHERE xid=".$this->proj_id." AND type=$proj_mem_type";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__.' Could not select from members table.');

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "$row[0] <br>\n";
			array_push($this->proj_members,$row[0]);
		}

		if(count($this->proj_members)==0)
			_quit_error('No members found for this project (id='.$this->proj_id.'). Must be corrupted database.');

		if(!$full_mode)
			return;

		$mem_sql=array();
		foreach($this->proj_members as $k => $v){
			//echo "proj_mem: $k -> $v <br>";
			array_push($mem_sql,$v);
		}

		$sql_where='u.uid IN ('.implode(',',$mem_sql).')';
		$query = "SELECT r.rid, u.uid, e.xval, r.phase_of_rid FROM ($req_tbl AS r, users AS u) LEFT JOIN $pnets_tbl AS e ON (r.rid=e.rid AND u.uid=e.uid AND e.type=$pnet_type_pri) WHERE $sql_where AND r.$req_in_proj_fn='".$this->inproj."' AND r.$req_proj_id_fn=".$this->proj_id." AND ".$this->sql_states_str." ORDER BY r.phase_of_rid DESC";
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		$mvoted=array();
		$max_rvote=array();
		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "$row[0], $row[1], $row[2], $row[3] <br>";
			if($row[2]){
				//echo "pri(".$row[0].")=".$row[2]."<br>\n";
				$this->mem_pri[$row[1]][$row[0]]=$row[2];
				$mvoted[$row[0]][$row[1]]=TRUE;
				if(!isset($max_rvote[$row[0]]))
					$max_rvote[$row[0]]=0;
				if($row[2]>$max_rvote[$row[0]])
					$max_rvote[$row[0]]=$row[2];
			}else
				$this->mem_pri[$row[1]][$row[0]]=0;
			if(!isset($this->members_vote[$row[0]]))
				$this->members_vote[$row[0]]=0;
			$this->members_vote[$row[0]]+=$this->mem_pri[$row[1]][$row[0]];
		}

		$nmem=count($this->proj_members);
		$step=0.8/$nmem;
		foreach(array_keys($this->members_vote) as $r){
			if(!isset($mvoted[$r])) continue;
			$sum=$this->members_vote[$r];
			$nvoted=count($mvoted[$r]);
			$avg=$sum/$nvoted;
			// a little more sophisticated than a simple average :)
			// with an extra for participation
			//$this->members_vote[$r]=$avg+$sum/($nmem*$max_rvote[$r]);
			$this->members_vote[$r]=$avg+$step*$nvoted/$nmem;
			//echo "$nmem mem_vote($r)=".$this->members_vote[$r]." avg=".$avg."<br>\n";
		}
	}

	function is_proj_member($uid){
		if(in_array($uid,$this->proj_members))
			return TRUE;
		return FALSE;
	}

	function is_proj_owner($uid){
		return $this->project_info['added_by']==$uid;
	}

	function proj_owner(){
		return $this->project_info['added_by'];
	}

	function get_members_vote($rid){
		return isset($this->members_vote[$rid])?$this->members_vote[$rid]:0;
	}

	function &members_without_vote($uids=FALSE){
		$uid2rids=array();

		$this->init_proj_members();
		if(!$uids)
			$uids=array_keys($this->mem_pri);

		foreach ($this->mem_pri as $u => $upris) {
			if(!in_array($u,$uids))
				continue;

			$tmp=array();
			foreach ($upris as $rid => $v) {
				if($v) continue;
				array_push($tmp,$rid);
			}
			if(count($tmp))
				$uid2rids[$u]=$tmp;

		}
		$ref=&$uid2rids;
		return $ref;
	}

	function rid_has_all_members_vote($rid){
		$this->init_proj_members(FALSE);
		foreach ($this->mem_pri as $u => $upris)
			if(!isset($upris[$rid]) || !$upris[$rid])
				return FALSE;
		return TRUE;
	}

	function pick_safe_owner($phase,$pqid){
		global $uid;

		$pqs_2d=&$this->pqs_2d;
		$ph_mgr_fn=$this->object_params['pm_phases'][$phase-1]['mgr_fn'];

		if(YxlUsers::exists($pqs_2d[$ph_mgr_fn][$pqid]))
			return $pqs_2d[$ph_mgr_fn][$pqid];

		if($p=$pqs_2d['parent'][$pqid]){
			if(YxlUsers::exists($pqs_2d[$ph_mgr_fn][$p]))
				return $pqs_2d[$ph_mgr_fn][$p];
		}

		return $uid;
	}

	var $qid2fixer_2d=array();
	var $fixer2qid_2d=array();
	var $qid2rid_2d=array();
	var $rid2qid_2d=array();
	var $fixer2rid_thru_qid_2d=array();
	var $all_fixers=array();
	function init_fixers($fixer_list_only=FALSE){
		global $user_spec2str;

		if(count($this->all_fixers) && $fixer_list_only)
			return;

		$ignore_mask=(USER_IS_CUSTOMER|USER_IS_HIDDEN);

		//echo "init_fixers $fixer_list_only<br>\n";

		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$ph1_mgr_fn=$this->object_params['pm_phases'][0]['mgr_fn'];
		$ph1_fixer_fn=$this->object_params['pm_phases'][0]['implementer_fn'];
		$ph1_fixer_spec=$this->object_params['pm_phases'][0]['implementer_spec'];
		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$inproj=($this->inproj=='yes');
		$qid=$this->project_info[$pq_fn];

		$query = "SELECT q.$pq_fn, u.uid, u.username FROM $pq_tbl AS q INNER JOIN users AS u ON (q.$ph1_mgr_fn=u.uid OR q.$ph1_mgr_fn=u.report_to) AND !(u.user_is&".($ignore_mask|USER_IS_LOCKED).") WHERE (q.$pq_fn=$qid OR q.parent=$qid)";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $pq_tbl table.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "init_fixers: qid=$row[0] uid=$row[1]. $row[2] <br>";
			if(!$this->ts_user_share($row[1])) continue;
			$this->qid2fixer_2d[$row[0]][$row[1]]=$row[1];
			$this->fixer2qid_2d[$row[1]][$row[0]]=$row[0];
			$this->all_fixers[$row[1]]=$row[1];
		}

		$query = "SELECT uid, username FROM users WHERE spec=$ph1_fixer_spec AND !(user_is&".($ignore_mask|USER_IS_LOCKED).")";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from users table.");

		while($row=mysql_fetch_array($result, MYSQL_NUM))
			$this->all_fixers[$row[0]]=$row[0];

		$ffn=$inproj?"r.$ph1_fixer_fn":"r.plan_$ph1_fixer_fn";
		$query = "SELECT u.uid, r.$pq_fn, u.first_name, u.last_name, u.spec, u.user_is FROM $req_tbl AS r INNER JOIN users AS u ON $ffn=u.uid AND !(u.user_is&$ignore_mask) WHERE r.$req_in_proj_fn='".$this->inproj."' AND ".$this->sql_states_str." AND r.$req_proj_id_fn=".$this->proj_id;
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		$add_fixers=array();
		$locked=array();
		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "init_fixer2: uid=$row[0] qid=$row[1] $row[2]<br>\n";
			if(($row[4]!=$ph1_fixer_spec) && !isset($this->all_fixers[$row[0]]))
				_warning('Warning: Fixer '."'$row[2] $row[3]'".' does not belong to '.$user_spec2str[$ph1_fixer_spec].' group!');
			if($row[5]&USER_IS_LOCKED && !isset($locked[$row[0]])){
				_warning('Warning: Fixer '."'$row[2] $row[3]'".' is locked and will not be able to implement requests currently assigned to him!');
				$locked[$row[0]]=TRUE;
			}

			$this->qid2fixer_2d[$row[1]][$row[0]]=$row[0];
			$this->fixer2qid_2d[$row[0]][$row[1]]=$row[1];
			$this->all_fixers[$row[0]]=$row[0];
		}

		if($fixer_list_only) return TRUE;

		// Initialize the arrays
		$query = "SELECT $pq_fn FROM $pq_tbl WHERE $pq_fn=$qid OR parent=$qid";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $pq_tbl table.");
		if(mysql_num_rows($result) == 0)
			_die2(__FILE__.':'.__LINE__.' Empty $pq_tbl table.');
		while($row=mysql_fetch_array($result, MYSQL_NUM))
			$this->qid2rid_2d[$row[0]]=array();

		// Fill the arrays
		$query = "SELECT q.$pq_fn, r.rid FROM $pq_tbl AS q, $req_tbl AS r WHERE q.$pq_fn=r.$pq_fn AND (q.$pq_fn=$qid OR q.parent=$qid) AND (".$this->sql_states_str." AND r.$req_in_proj_fn='".$this->inproj."' AND r.$req_proj_id_fn=".$this->proj_id.")";
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $pq_tbl table.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "qid=$row[0], rid=$row[1], in_rel=$row[2]<br>";
			$this->qid2rid_2d[$row[0]][$row[1]]=$row[1];
			$this->rid2qid_2d[$row[1]][$row[0]]=$row[0];
		}

		foreach ($this->fixer2qid_2d as $u => $qids) {
			//echo "-- uid=$u qid=$qids <br>";
			if(!isset($this->fixer2rid_thru_qid_2d[$u]))
				$this->fixer2rid_thru_qid_2d[$u]=array();
			foreach ($qids as $c => $v) {
				if(isset($this->qid2rid_2d[$c])){
					//echo "--+ uid=$u qid=$c $this->qid2rid_2d[$c]<br>";
					foreach ( $this->qid2rid_2d[$c] as $r => $rid) {
						//echo "uid=$u qid=$c, rid=$r <br>";
						$this->fixer2rid_thru_qid_2d[$u][$r]=$r;
					}
					//echo "-- <br>";
				}
			}
		}
		return TRUE;
	}

	var $all_testers=array();
	function init_testers(){
		global $user_spec2str;

		if(count($this->all_testers))
			return;

		$ignore_mask=(USER_IS_CUSTOMER|USER_IS_HIDDEN);

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		if(count($this->object_params['pm_phases'])<2) return;

		$ph2_mgr_fn=$this->object_params['pm_phases'][1]['mgr_fn'];
		$ph2_fixer_fn=$this->object_params['pm_phases'][1]['implementer_fn'];
		$ph2_fixer_spec=$this->object_params['pm_phases'][1]['implementer_spec'];
		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$inproj=($this->inproj=='yes');

		$query = "SELECT uid FROM users WHERE (spec=".USPEC_CAE." OR spec=$ph2_fixer_spec) AND !(user_is&".($ignore_mask|USER_IS_LOCKED).")";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__.' Could not select from users table.');

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			if(!$this->ts_user_share($row[0])) continue;
			//echo "init_tester: uid=$row[0] <br>";
			$this->all_testers[$row[0]]=$row[0];
		}

		$ffn=$inproj?"r.$ph2_fixer_fn":"r.plan_$ph2_fixer_fn";
		$query = "SELECT u.uid, r.$pq_fn, u.first_name, u.last_name, u.spec, u.user_is FROM $req_tbl AS r, users AS u WHERE $ffn=u.uid AND r.$req_in_proj_fn='".$this->inproj."' AND !(u.user_is&$ignore_mask) AND ".$this->sql_states_str." AND r.$req_proj_id_fn=".$this->proj_id;
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could select from $req_tbl table.");

		$locked=array();
		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "init_tester2: uid=$row[0] <br>";
			if(($row[4]!=$ph2_fixer_spec) && ($row[4]!=USPEC_CAE) && !isset($this->all_testers[$row[0]]))
				_warning('Warning: Tester '."'$row[2] $row[3]'".' does not belong to '.$user_spec2str[$ph2_fixer_spec].' or '.$user_spec2str[USPEC_CAE].' group!');
			if($row[5]&USER_IS_LOCKED && !isset($locked[$row[0]])){
				_warning('Warning: Tester '."'$row[2] $row[3]'".' is locked and will not be able to test requests currently assigned to him!');
				$locked[$row[0]]=TRUE;
			}

			$this->all_testers[$row[0]]=$row[0];
		}

	}

	// _whp - work hour pointer
	// _wh - number of worked hours
	// All hour pointers start from 0 ...
	// So 0fix_whp points to the 1st fixers work hour
	// 0test_whp points to the 1st fixers work hour
	// Initialization puts them to -1.

	var $fixers_whp=array();
	var $fixers_waste=array();
	var $testers_required_hrs=array();
	var $testers_start_whp=array();
	function init_hps(){

		$fix_phase_only=count($this->object_params['pm_phases'])<2;

		$this->init_fixers(TRUE);
		foreach($this->all_fixers as $f => $v){
			//$this->fixers_whp[$f]=-1;
			$this->fixers_whp[$f]=0;
			$this->fixers_waste[$f]=array();
		}

		if($fix_phase_only) return;

		if(count($this->all_testers)==0)
			$this->init_testers();
		foreach($this->all_testers as $t => $v){
			//$this->testers_start_whp[$t]=-1;
			//echo "init_hps tester $t <br>\n";
			$this->testers_start_whp[$t]=0;
			$this->testers_required_hrs[$t]=0;

		}

	}

	function fixer_wh($uid){
		//echo "fixer_wh: $uid<br>\n";
		if(!isset($this->fixers_whp[$uid])){
			_warn_yoxel('Inconsistent DB: uid='.$uid.' is not a fixer. '.array_pop($this->log));
			return 0;
		}

		//print_r($this->fixers_whp); echo "<br>\n";
		return $this->fixers_whp[$uid];
	}

	function fixer_ch($uid){
		return $this->wcal->whp2ch($uid,$this->fixer_wh($uid));
	}

	function tester_wh($uid){
		if(!isset($this->testers_required_hrs[$uid])){
			_warn_yoxel('Inconsistent DB: uid='.$uid.' is not a tester. '.array_pop($this->log));
			return 0;
		}

		//echo "tester_wh: start_whp=".$this->testers_start_whp[$uid]." req_wh=".$this->testers_required_hrs[$uid]." <br>\n";
		return $this->testers_start_whp[$uid]+$this->testers_required_hrs[$uid];
	}

	function tester_ch($uid){
		return $this->wcal->whp2ch($uid,$this->tester_wh($uid));
	}

	function set_fixer_wh($uid,$wh){
		$this->fixers_whp[$uid]=$wh;
	}

	function get_max_fhp($fixers=FALSE){
		$max_fhp=$this->wcal->start_ch();
		foreach ($this->fixers_whp as $k => $v) {
			if($fixers && !in_array($k,$fixers))
				continue;
			if(!$v) // This fixer's time has not been used yet
				continue;
			$hr=$this->wcal->whp2ch($k, $v);
			//echo "get_max_fhp: uid=$k whp=$v hr=$hr <br>";
			if($hr>$max_fhp)
				$max_fhp=$hr;
		}
		return $max_fhp;
	}

	function get_max_thp(){
		$max_thp=$this->wcal->start_ch();
		foreach ($this->testers_required_hrs as $u => $h) {
			//echo "get_max_thp: tester=$u, start_whp=".$this->testers_start_whp[$u].", required=$h<br>\n";
			$hr=$this->wcal->whp2ch($u, $this->testers_start_whp[$u]+$h);
		   if($hr>$max_thp)
		   	$max_thp=$hr;
		}
		return $max_thp;
	}

	function fixer_wasted_hrs($uid){
		$waste=0;
		foreach($this->fixers_waste[$uid] as $l)
			$waste+=$l;

		return $waste;
	}

	function move_fixer_hp($uid, $hrs, $extend_cal=FALSE, $move_mode=WH_WASTE_USE){
		if($extend_cal)
			$this->wcal->extend($uid,$this->wcal->start_ch(),$hrs);

		$start_ch=$this->wcal->whp2ch($uid,$this->fixer_wh($uid));
		if($extend_cal)
			$this->wcal->extend($uid,$start_ch,$hrs);

		$current_fwh=$this->fixer_wh($uid);
		$exceed=$current_fwh+$hrs-$this->wcal->total_whs($uid);
		//echo "move_fixer_hp: uid=$uid hrs=$hrs whp=$current_fwh exceed=$exceed mvm=$move_mode<br>\n";

		if($hrs>=0){
		// Consume work hrs
			if($move_mode==WH_WASTE_REGISTER){
				if($exceed<=0)
					$this->fixers_waste[$uid][$current_fwh]=$hrs;
			}elseif($move_mode==WH_WASTE_USE){
				// First see if we could re-use wasted hrs
				foreach($this->fixers_waste[$uid] as $h => $l){
					if($hrs>$l)
						continue;

					// Found 1st long enough waste, use it!
					//echo "found waste: $h + $l<br>\n";
					unset($this->fixers_waste[$uid][$h]);
					$this->fixers_waste[$uid][$h+$hrs]=$l-$hrs;

					$start_ch=$this->wcal->whp2ch($uid,$h);
					$end_ch=$this->wcal->whp2ch($uid,$h+$hrs);
					return array(0,$start_ch,$end_ch,WH_WASTE_USE);
				}

				// Could not fit and re-use wasted hrs, use new hrs
			}

			$end_ch=FALSE;
			if($exceed<=0){
				$this->fixers_whp[$uid]+=$hrs;
				$end_ch=$this->wcal->whp2ch($uid,$this->fixer_wh($uid));
			}
			//echo "ignore waste <br>\n";
			return array($exceed,$start_ch,$end_ch,WH_WASTE_IGNORE);
		}else{
		// Release work hrs
			if($move_mode==WH_WASTE_USE){
				$wks=array_keys($this->fixers_waste[$uid]);
				if(!($h=array_pop($wks)))
					_die(__FILE__.':'.__LINE__." Did not find any hr-waste records.");

				$l=$this->fixers_waste[$uid][$h];
				unset($this->fixers_waste[$uid][$h]);
				$this->fixers_waste[$uid][$h+$hrs]=$l+abs($hrs);
				return array(0,$start_ch,$start_ch,WH_WASTE_USE);
			}elseif($move_mode==WH_WASTE_IGNORE){
				$this->fixers_whp[$uid]+=$hrs;
				$ewp=$this->fixer_wh($uid);
				foreach($this->fixers_waste[$uid] as $h => $l){
					if($h+$l>$ewp)
						_die(__FILE__.':'.__LINE__." Waste overlap $h + $l vs. $ewp.");
				}
				$end_ch=$this->wcal->whp2ch($uid,$ewp);
				return array(0,$end_ch,$end_ch,WH_WASTE_IGNORE);
			}else
				_die(__FILE__.':'.__LINE__." Unknown mode.");
		}

	}

	function move_tester_hp($uid, $hrs, $extend_cal=FALSE){
		// For now test_mode=0 only
		$ra=array();
		$ruid=$uid;
		$saved_tsw=array();

		// When does fix phase end?
		$max_fhp=$this->get_max_fhp();
		//echo "in move_tester_hp: uid=$uid hrs=$hrs max_fhp=$max_fhp max_fhp_date=".$this->wcal->ch2date("Y/m/j(H)",$max_fhp)."<br>";

		// Test-phase starts right after fix-phase so
		// shift the test start hour for all testers


		$restore=FALSE;
		foreach ($this->testers_required_hrs as $u => $h) {
			// This tester's time has not been used yet.
			if(($u!=$uid) && !$this->testers_start_whp[$u] && !$h)
				continue;

			$exceed=$h;
			$thp=$this->wcal->ch2whp($u, $max_fhp);
			//echo "*move_tester_hp1: tuid=$u twhp=$thp <br>";
			if($thp<0){
				//echo "move_tester_hp: A <br>";
				$exceed=$hrs;
				$ruid=$u;
				$restore=TRUE;
				break;
			}
			if($extend_cal)
				$this->wcal->extend($u,$max_fhp,$h+$hrs);

			$exceed=$thp+$h-$this->wcal->total_whs($u);
			$hr=$this->wcal->whp2ch($u, $thp+$h);
			//echo "*move_tester_hp2: uid=$u required_hrs=$h thp=".($thp+$h)." wcal->hr=$hr exceed=$exceed<br>";
			if($exceed>0){
				//echo "move_tester_hp: B exceed=$exceed<br>";
				$ruid=$u;
				$restore=TRUE;
				break;
			}
			//echo "move_tester_hp: **save: u=$u old_thp=".$this->testers_start_whp[$u]." new_thp=$thp <br>";
			$saved_tsw[$u]=$this->testers_start_whp[$u];
			$this->testers_start_whp[$u]=$thp;
		}

		if($restore){
			foreach ($saved_tsw as $u => $h) {
				//echo "move_tester_hp: **restore: u=$u h=$h thp=$thp <br>";
				$this->testers_start_whp[$u]=$saved_tsw[$u];
			}
			//return array($exceed+$hrs,$max_fhp+$h,$max_fhp+$h+$hrs);
			return array($exceed,FALSE,FALSE);
		}

		$swhp=$this->testers_start_whp[$uid]+$this->testers_required_hrs[$uid];
		//echo " move_tester_hp3: required_hrs=".$this->testers_required_hrs[$uid]." start_hp=".$this->testers_start_whp[$uid]." tt=$hrs, swhp=$swhp, total_hrs=".$this->wcal->total_whs($uid)." <br>";

		if($extend_cal)
			$this->wcal->extend($uid,$max_fhp,$this->testers_required_hrs[$uid]+$hrs);

		$exceed=$swhp+$hrs-$this->wcal->total_whs($uid);
		$start_ch=$this->wcal->whp2ch($uid,$swhp);
		$hr=$this->wcal->whp2ch($uid,$swhp+$hrs);
		if($exceed<=0)
			$this->testers_required_hrs[$uid]+=$hrs;
		else
			$hr=FALSE;
		return array($exceed,$start_ch,$hr);

	}

	function wh_debug(){
		echo "Fixers: <br>";
		foreach ($this->fixers_whp as $u => $h) {
			echo "uid=$u whp=$h whp_date=".$this->wcal->wh2date("m/j/Y(H)",$u, $h)." <br>";
		}
		echo "Testers: <br>";
		foreach ($this->testers_required_hrs as $u => $h) {
			echo "uid=$u hrs=$h <br>";
		}
	}

	function generate_plan_list($fid=FALSE){

		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$pnets_tbl=$this->object_params['pnets_tbl'];

		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$ph1_mgr_fn=$this->object_params['pm_phases'][0]['mgr_fn'];
		$ph1_fixer_fn=$this->object_params['pm_phases'][0]['implementer_fn'];
		$ph1_fixer_spec=$this->object_params['pm_phases'][0]['implementer_spec'];
		$ph1_dflt_fix_fn=$this->object_params['pm_phases'][0]['dflt_fix_fn'];
		$ph1_pnet_type=$this->object_params['pm_phases'][0]['pnet_type'];

		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$req_proj_pri_fn=$this->object_params['req_proj_pri_fn'];
		if(($fix_and_test=(count($this->object_params['pm_phases'])>1)))
			$ph2_fixer_fn=$this->object_params['pm_phases'][1]['implementer_fn'];

		$qid=$this->project_info[$pq_fn];
		$proj_deadline=$this->project_info[$proj_deadline_fn];
		$dft=$this->project_info[$ph1_dflt_fix_fn];
		$this->init_fixers(TRUE);
		$resource_num=count($this->all_fixers);

		$rid_deps=array();
		$now_tp=time();
		$proj_deadline_arr=explode('-',$proj_deadline);
		$proj_deadline_tp=mktime($this->wcal->work_day_start_hr,0,0,$proj_deadline_arr[1],$proj_deadline_arr[2],$proj_deadline_arr[0]);
		$avail_days=intval(($proj_deadline_tp-$now_tp)/SECS_IN_DAY);

		// Take half of all available work hours
		$avail_wrk_hrs=intval($avail_days*$this->wcal->work_day_len_hr*$resource_num/2);
		//echo "Total Available Work Hours: days=$avail_days fixers=$resource_num $avail_wrk_hrs <br>";

		$fwhere='';
		if($fid){
			$wfid=$this->rts->wfid();
			$filters=new UserFilters($wfid);
			$fwhere=$filters->filter2sql_where($fid);
		}
		//echo "$fwhere <br>\n";

		$proj_ids_str=$this->valid_proj_ids_str();

		$query = "SELECT r.rid, MIN(e.xval), r.customer_pri FROM $req_tbl AS r INNER JOIN $pq_tbl AS p ON r.$pq_fn=p.$pq_fn LEFT JOIN $pnets_tbl AS e ON r.rid=e.rid AND r.$ph1_fixer_fn=e.uid AND e.type=$ph1_pnet_type WHERE (r.$req_in_proj_fn!='yes' OR !(r.$req_proj_id_fn IN ($proj_ids_str))) $fwhere AND (p.parent=$qid OR r.$pq_fn=$qid) AND r.customer_pri!=".REQ_PRI_CRITICAL." AND ".$this->sql_states_str." GROUP BY r.rid ORDER BY r.customer_pri DESC, IF(r.class=".$this->rts->bug_class.",1,0) ASC";
		//echo "$query <br>\n"; die;
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		$total_used_hrs=0;
		$p=1;
		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			$rid=$row[0];
			foreach($this->get_dependency_list($rid) as $r)
				if(!in_array($r,$rid_deps))
					array_push($rid_deps,$r);
			$total_used_hrs+=$row[1]?$row[1]:$dft;
			//echo "$rid, $row[1], wrk_hrs=$total_used_hrs, avail_hrs=$avail_wrk_hrs. <br>";
			//echo "deplist: ";print_r($rid_deps);echo "<br>\n";
			if($total_used_hrs>$avail_wrk_hrs)
				break;
			$query = "UPDATE $req_tbl SET $req_in_proj_fn='plan', $req_proj_pri_fn=$p, plan_pri=0, $req_proj_id_fn=".$this->proj_id.", plan_$ph1_fixer_fn=$ph1_fixer_fn,".($fix_and_test?" plan_$ph2_fixer_fn=$ph2_fixer_fn,":'')." options=options|".REQ_OPT_RPSAUTO." WHERE rid=$rid";
			//echo "$query<br>\n";
			if(!(@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
			$p++;
		}

		// If no unincluded dependencies
		if(count($rid_deps)==0)
			return;
		//echo "deplist: ";print_r($rid_deps);echo "<br>\n";
		$dep_sql_str = '(r.rid='.implode(' OR r.rid=',$rid_deps).')';
		$query = "SELECT r.rid, r.$pq_fn, q.parent, r.customer_pri FROM $req_tbl AS r, $pq_tbl AS q WHERE r.$pq_fn=q.$pq_fn AND r.$req_in_proj_fn!='yes' AND $dep_sql_str AND ".$this->sql_states_str;
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//print_r($row);
			//echo "<br>\n";
			if(($row[1]!=$qid) && ($row[2]!=$qid))
				_die('InterPQ dependency!');
			if($row[3]==REQ_PRI_CRITICAL)
				_die('Dependency on a critical request!');
			$rid=$row[0];
			$query = "UPDATE $req_tbl SET $req_in_proj_fn='plan', $req_proj_pri_fn=$p, plan_pri=0, $req_proj_id_fn=".$this->proj_id.", plan_$ph1_fixer_fn=$ph1_fixer_fn,".($fix_and_test?" plan_$ph2_fixer_fn=$ph2_fixer_fn,":'')." options=options|".REQ_OPT_RPSAUTO." WHERE rid=$rid";
			//echo "$query <br>";
			if(!(@mysql_query($query)))
				_die(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
			$p++;
		}

	}

	function update_proj_label($lbl,&$err_str){

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];
                $pqid=$this->project_info[$pq_fn];
		$proj_id=$this->project_info[$proj_id_fn];

		if(!is_rel_lbl($lbl)){
			$err_str="Please use a correct label name.";
			return FALSE;
		}

		$tlbl = escape_data($lbl);
		$query = "SELECT $proj_id_fn FROM $proj_tbl WHERE $pq_fn=$pqid AND $proj_id_fn!=$proj_id AND label='$tlbl' LIMIT 1";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl table.");

		if(mysql_num_rows($result)){
                     	$err_str="A release with the same label already exists.";
			return FALSE;
		}

		if($lbl!=$this->project_info['label']){
			$query = "UPDATE $proj_tbl SET label='$tlbl' WHERE $proj_id_fn=$proj_id";
			//echo "$query <br>"; die();
			if(!(@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__.' Could not update projects table.');

			$this->history_buf.="\nLabel changed: ".$this->project_info['label']." -> $lbl\n";
			$this->project_info['label']=$lbl;
		}
		return TRUE;
	}

	function update_proj_deadline($ddate,&$err_str){

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];
                $pqid=$this->project_info[$pq_fn];
		$proj_id=$this->project_info[$proj_id_fn];

                if(!is_date($ddate)){
			$err_str="Please enter correct deadline date.";
			return FALSE;
		}
		$sdate=date_norm2sql($ddate);
                $nrld=escape_data($sdate);
                if(!$nrld){
			$err_str="Please enter correct deadline date.";
			return FALSE;
		}

		if($sdate!=$this->project_info[$proj_deadline_fn]){
			$query = "UPDATE $proj_tbl SET $proj_deadline_fn='$nrld' WHERE $proj_id_fn=$proj_id";
			//echo "$query <br>";
			if(!(@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__.' Could not update projects table.');

			$this->history_buf.="\nDeadline changed: ".date_sql2norm($this->project_info[$proj_deadline_fn])." -> $ddate\n";
			$this->project_info[$proj_deadline_fn]=$nrld;
		}
		return TRUE;
	}

	 function update_member_vote($rid,$upri){
		global $uid;

		$pnets_tbl=$this->object_params['pnets_tbl'];
		$pnet_type_pri=$this->object_params['pnet_type_pri'];

		$this->init_proj_members();
		if(!$upri)
			return;
		if(isset($this->mem_pri[$uid][$rid]) && ($this->mem_pri[$uid][$rid]==$upri))
			return;

		$query = "DELETE FROM $pnets_tbl WHERE rid=$rid AND (uid=$uid OR uid=0) AND type=$pnet_type_pri";
                if(!@mysql_query($query))
                      	_die2(__FILE__.':'.__LINE__." Could not delete from $pnets_tbl table.");

                $query = "INSERT INTO $pnets_tbl (rid, uid, xval, type) VALUES ('$rid', '$uid', '$upri', '$pnet_type_pri')";
                //echo "$query <br>";
                if(!@mysql_query($query))
                      	_die2(__FILE__.':'.__LINE__." Could not update $pnets_tbl table.");

	}

	function update_plan_pri(&$rid2order){

		$req_tbl=$this->object_params['requests_tbl'];
		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$req_proj_pri_fn=$this->object_params['req_proj_pri_fn'];
		$pnets_tbl=$this->object_params['pnets_tbl'];
		$pnet_type_pri=$this->object_params['pnet_type_pri'];

		foreach($rid2order as $r => $o){
			$query = "UPDATE $req_tbl SET plan_pri='$o' WHERE rid=$r";
			//echo "$query <br>";
			if(!(@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
		}
		if(count($rid2order))
			return;

		$this->init_proj_members(FALSE);

                $query = "SELECT r.rid, SUM(e.xval) AS mvote FROM $req_tbl AS r LEFT OUTER JOIN $pnets_tbl AS e ON r.rid=e.rid AND e.type=$pnet_type_pri WHERE r.$req_in_proj_fn='plan' AND r.$req_proj_id_fn=".$this->proj_id." AND e.uid IN (".implode(',',$this->proj_members).")  GROUP BY r.rid ORDER BY mvote DESC, r.$req_proj_pri_fn DESC";
                //echo "$query <br>";
                if(!($result=@mysql_query($query)))
                        _die2(__FILE__.':'.__LINE__." Can not select from $req_tbl table.");
                for($i=1;$row=mysql_fetch_array($result, MYSQL_NUM);$i++){
                        //echo "rid=$row[0] sum_pri=$row[1] <br>";
                        $query = "UPDATE $req_tbl SET plan_pri=$i WHERE rid=$row[0]";
                        if(!(@mysql_query($query)))
                                _die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
                }
	}

	function fixer_with_max_atime($qid){
		$max_at=0;
		$uid=FALSE;
		echo "fixer_with_max_atime: $qid <br>\n";
		$fixers=&$this->phase_pq_resources(1,$qid);
		foreach ($fixers as $k => $u) {
			$at=$this->wcal->total_whs($u)-$this->fixers_whp[$u];
			//echo "fixer_with_max_atime: fixer=$u total=".$this->wcal->total_whs($u)." at=$at <br>";
		   if($at>=$max_at){
		   	$max_at=$at;
		   	$uid=$u;
		   }
		}
		return $uid;
	}

	function fixer_with_min_chp_atime($qid, $req_hrs=FALSE, $waste_adjust=TRUE){

		$min_chp=FALSE;
		$uid=FALSE;
		if(!$req_hrs)
			$req_hrs=$this->project_info['dflt_fix_time'];

		//echo "fixer_with_min_chp_atime: $qid, $req_hrs <br>\n";
		$fixers=&$this->phase_pq_resources(1,$qid);
		foreach ($fixers as $u) {
			if($this->ts_user_share($u)<=0)
				continue;

			$scaler=$this->ts_user_scaler($u);
			$rh=$req_hrs*$scaler;

			$waste=$waste_adjust?$this->fixer_wasted_hrs($u):0;

			$chp=$this->fixer_ch($u)-$waste;
			$at=$this->wcal->total_whs($u)-$this->fixer_wh($u)+$waste;

			if(!$min_chp){
				$min_chp=$chp;
				$uid=$u;
			}elseif(($chp<$min_chp) && ($at>=$rh)){
				$min_chp=$chp;
				$uid=$u;
			}
		}
		return $uid;
	}

	function tester_with_max_atime($qid){
		$max_at=0;
		$uid=FALSE;
		$max_fhp=$this->get_max_fhp();
		$testers=&$this->phase_pq_resources(2,$qid);
		foreach ($testers as $k => $u) {
			$end_of_fix_whp=$this->wcal->ch2whp($u, $max_fhp);
			//echo "wd=",count($uid_wrk_days_2d[$k])," end_of_fix_whp=$end_of_fix_whp, rt={$this->testers_required_hrs[$k]}<br>";
			$rt=$this->wcal->total_whs($u)-$end_of_fix_whp-$this->testers_required_hrs[$u];
			//echo "tester=$v at=$rt <br>";
		   if($rt>=$max_at){
		   	$max_at=$rt;
		   	$uid=$k;
		   }
		}
		return $uid;
	}

	function tester_with_min_chp_atime($qid, $req_hrs=FALSE){

		$min_chp=FALSE;
		$uid=FALSE;
		if(!$req_hrs)
			$req_hrs=$this->project_info['dflt_test_time'];

		$max_fhp=$this->get_max_fhp();
		$testers=&$this->phase_pq_resources(2,$qid);
		foreach ($testers as $u) {
			if($this->ts_user_share($u)<=0)
				continue;

			$scaler=$this->ts_user_scaler($u);
			$rh=$req_hrs*$scaler;

			$chp=$this->tester_ch($u);
			$end_of_fix_whp=$this->wcal->ch2whp($u, $max_fhp,TRUE,__LINE__);
			if($end_of_fix_whp<0)
				continue;

			$at=$this->wcal->total_whs($u)-$end_of_fix_whp-$this->tester_wh($u);

			if(!$min_chp){
				$min_chp=$chp;
				$uid=$u;
			}elseif(($chp<$min_chp) && ($at>=$rh)){
				$min_chp=$chp;
				$uid=$u;
			}
		}
		return $uid;
	}

	var $fixer_rid_et=array();
	var $tester_rid_et=array();
	function init_ets(){
		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$pnets_tbl=$this->object_params['pnets_tbl'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$ph1_fixer_fn=$this->object_params['pm_phases'][0]['implementer_fn'];
		$ph1_fixer_spec=$this->object_params['pm_phases'][0]['implementer_spec'];
		$ph1_dflt_fix_fn=$this->object_params['pm_phases'][0]['dflt_fix_fn'];
		$ph1_pnet_type=$this->object_params['pm_phases'][0]['pnet_type'];
		$ph1_done_sm=$this->object_params['pm_phases'][0]['done_state_mask'];
		$ph1_start_sm=$this->object_params['pm_phases'][0]['start_state_mask'];
		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$req_proj_pri_fn=$this->object_params['req_proj_pri_fn'];

		if(($fix_and_test=(count($this->object_params['pm_phases'])>1)))
			$ph2_fixer_fn=$this->object_params['pm_phases'][1]['implementer_fn'];

		if(count($this->rid2qid_2d)==0)
			$this->init_fixers();

		$query = "SELECT r.rid, u.uid, IF(e.xval IS NULL,-1,e.xval), r.state, r.class, r.complete_pct FROM $req_tbl AS r INNER JOIN users AS u ON  (u.user_is&".USER_IS_CUSTOMER.")=0 LEFT JOIN $pnets_tbl AS e ON r.rid=e.rid AND u.uid=e.uid AND e.type=$ph1_pnet_type WHERE r.$req_in_proj_fn='".$this->inproj."' AND r.$req_proj_id_fn=".$this->proj_id." AND ".$this->sql_states_str;
		//echo "$query<br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Can not select ph1 from $pnets_tbl.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "init_ets: rid={$row[0]} fixer={$row[1]} <br>";
			if($row[3]&$ph1_done_sm[$row[4]]){
				//echo "init_ets: et=0 <br>";
				$this->fixer_rid_et[$row[1]][$row[0]]=0;
			}elseif($row[2]>=0){
				$pct=100;
				if(($row[3]&$ph1_start_sm[$row[4]]) && $row[5])
					$pct=(100-$row[5]);
				//echo "init_ets: rid=$row[0] et={$row[2]} pct=$pct <br>\n";
				//$this->fixer_rid_et[$row[1]][$row[0]]=$row[2];
				$this->fixer_rid_et[$row[1]][$row[0]]=round($row[2]*$pct/100);
			}
		}

		if(!$fix_and_test)
			return;
		$this->init_testers();

		$ph2_fixer_fn=$this->object_params['pm_phases'][1]['implementer_fn'];
		$ph2_fixer_spec=$this->object_params['pm_phases'][1]['implementer_spec'];
		$ph2_dflt_fix_fn=$this->object_params['pm_phases'][1]['dflt_fix_fn'];
		$ph2_pnet_type=$this->object_params['pm_phases'][1]['pnet_type'];
		$ph2_done_sm=$this->object_params['pm_phases'][1]['done_state_mask'];
		$ph2_start_sm=$this->object_params['pm_phases'][1]['start_state_mask'];

		$query = "SELECT r.rid, u.uid, IF(e.xval IS NULL,-1,e.xval), r.state, r.class FROM $req_tbl AS r INNER JOIN users AS u ON  (u.user_is&".USER_IS_CUSTOMER.")=0 LEFT JOIN $pnets_tbl AS e ON r.rid=e.rid AND u.uid=e.uid AND e.type=$ph2_pnet_type WHERE r.$req_in_proj_fn='".$this->inproj."' AND r.$req_proj_id_fn=".$this->proj_id." AND ".$this->sql_states_str;

		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Can not select from $pnets_tbl table.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "init_ets: rid={$row[0]} tester={$row[1]} et={$row[2]}<br>";
			if($row[3]&$ph2_done_sm[$row[4]]){
				//echo "init_ets: et=0 <br>";
				$this->tester_rid_et[$row[1]][$row[0]]=0;
			}elseif($row[2]>=0){
				//echo "init_ets: eT={$row[2]} <br>";
				$this->tester_rid_et[$row[1]][$row[0]]=$row[2];
			}
		}
	}

	function &rids_wihtout_any_estimate($phase,$uids=FALSE){
		$uid2rids=array();
		$ref=&$uid2rids;

		if($phase==1){
			$this->init_fixers(TRUE);
			$implementers=array_keys($this->all_fixers);
		}elseif($phase==2){
			$this->init_testers();
			$implementers=array_keys($this->all_testers);
		}else
			return $ref;

		if(!is_array($uids)){
			foreach($this->proj_members as $u)
				arr_push($implementers,$u);

			$uids=&$implementers;
		}

		foreach(array_keys($this->rid2qid_2d) as $r){
			if($phase==1){
				$uns=$this->users_for_rid_without_fet($r);
			}elseif($phase==2){
				$uns=$this->users_for_rid_without_tet($r);
			}else
				return $ref;
			//echo "rid=$r $uns <br>";
			foreach ($uns as $u){
				if(!in_array($u,$uids)) continue;
				$uid2rids[$u][$r]=$r;
			}
		}
		return $ref;
	}

	function &users_wihtout_rid_estimate($phase,$uids=FALSE){
		$uid2rids=array();
		$ref=&$uid2rids;
		if($phase==1){
			$this->init_fixers(TRUE);
			$implementers=array_keys($this->all_fixers);
		}elseif($phase==2){
			$this->init_testers();
			$implementers=array_keys($this->all_testers);
		}else
			return $ref;

		if(!is_array($uids))
			$uids=&$implementers;

		foreach ($uids as $u) {
			if($phase==1)
				$prs=$this->rids_without_fet_by_user($u);
			else
				$prs=$this->rids_without_tet_by_user($u);

			if(count($prs))
				$uid2rids[$u]=&$prs;
		}
		return $ref;
	}

	function rids_without_fet_by_user($fixer){

		if(!isset($this->fixer_rid_et[$fixer]))
			return array_keys($this->rid2qid_2d);

		$pr_list=array();
		foreach (array_keys($this->rid2qid_2d) as $rid) {
			if(!$this->fixer_has_et($rid,$fixer))
				array_push($pr_list,$rid);
		}
		return $pr_list;
	}

	function rids_without_tet_by_user($tester){

		if(!isset($this->tester_rid_et[$tester]))
			return array_keys($this->rid2qid_2d);

		$pr_list=array();
		foreach (array_keys($this->rid2qid_2d) as $rid) {
			if(!$this->tester_has_et($rid,$tester))
				array_push($pr_list,$rid);
		}
		return $pr_list;
	}

	function users_for_rid_without_fet($rid){
		$uns=array();
		foreach ($this->rid2qid_2d[$rid] as $c => $v) {
			//echo "users_for_rid_without_fet: component=$c <br>";
			if(!isset($this->qid2fixer_2d[$c]))
				continue;

			$uids=array_merge($this->qid2fixer_2d[$c],$this->proj_members);
			foreach ($uids as $u) {
				//echo "rid=$rid qid=$c uid=$u <br>";
				if($this->fixer_has_et($rid,$u))
					return array();
				else
					$uns[$u]=$u;
			}
		}
		return $uns;
	}

	function users_for_rid_without_tet($rid){
		$uns=array();

		$uids=array_merge($this->all_testers,$this->proj_members);
		foreach ($uids as $u) {
			//echo "users_for_rid_without_tet: rid=$rid tester=$u et=".$this->tester_et($rid,$u)."<br>";
			if($this->tester_has_et($rid,$u))
				return array();
			else
				$uns[$u]=$u;
		}
		return $uns;
	}

	function emails_for_rid_without_fet($rid){
		$uns = array();
		foreach ($this->users_for_rid_without_fet($rid) as $u)
			$uns[$u]=YxlUsers::full_email($u);
		return $uns;
	}

	function emails_for_rid_without_tet($rid){
		$uns=array();
		foreach ($this->users_for_rid_without_tet($rid) as $u)
			$uns[$u]=YxlUsers::full_email($u);
		return $uns;
	}

	function user_has_et($phase,$rid,$fixer){
		if($phase==1)
			return $this->fixer_has_et($rid,$fixer);
		elseif($phase==2)
			return $this->tester_has_et($rid,$fixer);
		else
			return FALSE;
	}

	function fixer_et($rid, $fixer, $scale=TRUE){
		$et=-1;
		$scaler=$this->ts_user_scaler($fixer);
		$uids=array_merge(array_keys($this->fixer_rid_et),$this->proj_members);
		foreach($uids as $u){
			$ets=&$this->fixer_rid_et[$u];
			if($fixer && ($fixer==$u) && isset($ets[$rid]))
				return $scale?round($ets[$rid]*$scaler):$ets[$rid];

			if(isset($ets[$rid]) && ($ets[$rid]>$et))
				$et=$ets[$rid];
		}
		if($et<0)
			$et=$this->project_info['dflt_fix_time'];
		return $scale?round($et*$scaler):$et;
	}

	function fixer_has_et($rid, $fixer){
		return isset($this->fixer_rid_et[$fixer][$rid]) && ($this->fixer_rid_et[$fixer][$rid]>=0)?TRUE:FALSE;
	}


	function tester_et($rid, $tester, $scale=TRUE){
		$et=-1;
		$scaler=$this->ts_user_scaler($tester);
		$uids=array_merge(array_keys($this->tester_rid_et),$this->proj_members);
		foreach($uids as $u){
			$ets=&$this->tester_rid_et[$u];
			if($tester && ($tester==$u) && isset($ets[$rid]))
				return $scale?round($ets[$rid]*$scaler):$ets[$rid];

			if(isset($ets[$rid]) && ($ets[$rid]>$et))
				$et=$ets[$rid];
		}
		if($et<0)
			$et=$this->project_info['dflt_test_time'];
		return $scale?round($et*$scaler):$et;
	}

	function tester_has_et($rid, $tester){
		return isset($this->tester_rid_et[$tester][$rid]) && ($this->tester_rid_et[$tester][$rid]>=0)?TRUE:FALSE;
	}

	function email_proj_members($subject,$body){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];

		$tmp=array();
		if(count($this->proj_members)==0)
			$this->init_proj_members(FALSE);

		foreach($this->proj_members as $i => $uid){
			$tmp[$uid]= YxlUsers::full_email($uid);
		}
		$cc_str=implode(', ',$tmp);

		$added_by=simple_query("SELECT added_by FROM $proj_tbl WHERE $proj_id_fn={$this->proj_id}");
		if(!$added_by)
			_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl table.");
		$to=YxlUsers::full_email($added_by);

		$headers=$this->mail_headers."\nCC: $cc_str";
		yxl_mail($to,$subject,$body,$headers);
	}

	function email_fixers($subject,$body){
		$ul=array();
		foreach ($this->rid2qid_2d as $r => $qids){
			$emails=$this->emails_for_rid_without_fet($r);
			//echo "email_fixers: rid=$r "; print_r($emails); echo "<br>\n";
			foreach ($emails as $i => $email)
				$ul[$email]=$email;
		}
		if(count($ul)){
			$to_fixers=implode(', ',$ul);
			//echo "To: $to_fixers <br>";
			yxl_mail($to_fixers,$subject,$body,$this->mail_headers);
		}
	}

	function email_testers($subject,$body){
		if(count($this->object_params['pm_phases'])<2)
			return;

		$ul=array();
		foreach ($this->rid2qid_2d as $r => $qids){
			$emails=$this->emails_for_rid_without_tet($r);
			//echo "email_testers: rid=$r "; print_r($emails); echo "<br>\n";
			foreach ($emails as $i => $email)
				$ul[$email]=$email;
		}
		if(count($ul)){
			$to_testers=implode(', ',$ul);
			//echo "To: $to_testers <br>";
			yxl_mail($to_testers,$subject,$body,$this->mail_headers);
		}
	}

	function email($puid, $subject, $synop, $cc_uids){
		$to_puid=YxlUsers::full_email($puid);
		$cc_list=array();
		while($u=array_pop($cc_uids))
			array_push($cc_list,YxlUsers::full_email($u));

		if(count($cc_list))
			$headers=$this->mail_headers."\nCC: ".implode(', ',$cc_list);
		else
			$headers=$this->mail_headers;

		//echo "email_fixer: ".html_safe_text($to_puid."\n".$headers)."<br>\n";
		// echo "$to_puid <br>";
		yxl_mail($to_puid,$subject,$synop,$headers);
	}

	function format2str($format,$date){
		$darr= explode('/',$date);
		$dtp=mktime(0, 0, 0, $darr[0], $darr[1],$darr[2]);
		$q=intval(($darr[0]-1)/3)+1;
		$s=date($format,$dtp-$this->wcal->tz_shift);
		return eregi_replace('q',"$q",$s);
	}

	function delete_project(){

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];
		$proj_mem_type=$this->object_params['proj_mem_type']; // like MEMB_TYPE_ITPROJ
		$hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_dc=$this->object_params['history_proj_decommit_type']; // like HISTORY_OPT_ITPM
		$hist_proj_plan=$this->object_params['history_proj_plan_type'];


		$query="DELETE FROM $req_tbl WHERE $req_proj_id_fn=".$this->proj_id." AND options&".REQ_OPT_RPSTMP;
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not delete from $req_tbl table.");

		$query="UPDATE $req_tbl SET $req_in_proj_fn='maybe', $req_proj_id_fn=0 WHERE $req_proj_id_fn=".$this->proj_id;
		if(!@mysql_query($query))
			_die(__FILE__.':'.__LINE__." Could not update $req_tbl table.");

		$query="DELETE FROM $proj_tbl WHERE $proj_id_fn=".$this->proj_id." LIMIT 1";
		if(!@mysql_query($query))
			_die(__FILE__.':'.__LINE__." Could not delete from $proj_id_fn table.");

		$query="DELETE FROM members WHERE xid=".$this->proj_id." AND type=$proj_mem_type";
		if(!@mysql_query($query))
			_die(__FILE__.':'.__LINE__.' Could not delete from members table.');

		$query="DELETE FROM $hist_tbl WHERE xid=".$this->proj_id." AND options=$hist_proj_dc";
		if(!@mysql_query($query))
			_die(__FILE__.':'.__LINE__." Could not delete from $hist_tbl table.");

		$query="DELETE FROM $hist_tbl WHERE xid=".$this->proj_id." AND options=$hist_proj_plan";
		if(!@mysql_query($query))
			_die(__FILE__.':'.__LINE__." Could not delete from $hist_tbl table.");

		if($this->rts->have_extra())
			$this->rts->x_remove_proj_templates($this->proj_id);
		$filters=new UserFilters($this->rts->wfid());
		$filters->remove_tagged_filters('_proj_tag',$this->proj_id);
	}

	var $ordered_plan=array();
	var $title_row=array();
	function plan_order(){
		$req_tbl=$this->object_params['requests_tbl'];
		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];
                $req_proj_pri_fn=$this->object_params['req_proj_pri_fn'];

		$t=array();
		$op=array();
		$tp=array();
		$is_multiphase=array();
		$proj_id=$this->proj_id;

		$sql_sel=array();
		$i=0;
		foreach($this->object_params['pm_phases'] as $phase){
			$i++;
			$ph_fixer_fn=$phase['implementer_fn'];
			$ph_plan_start=$phase['plan_start_fn'];
			$ph_plan_done=$phase['plan_done_fn'];
			$ph_start=$phase['start_fn'];
			$ph_done=$phase['done_fn'];
			$sql_sel[]="r.$ph_fixer_fn";
			if($i>1)
				$sql_sel[]="r.test_wrk_hrs AS ph{$i}_wrk_hrs";
			else
				$sql_sel[]="r.fix_wrk_hrs AS ph{$i}_wrk_hrs";

			if($this->inproj=='plan'){
				$sql_sel[]="r.plan_$ph_fixer_fn AS ph{$i}_fixer";
				$sql_sel[]="r.$ph_fixer_fn AS ph{$i}_ofixer";
			}else
				$sql_sel[]="r.$ph_fixer_fn AS ph{$i}_fixer";

			$sql_sel[]="UNIX_TIMESTAMP(r.$ph_plan_start) AS ph{$i}_plan_start";
			$sql_sel[]="UNIX_TIMESTAMP(r.$ph_plan_done) AS ph{$i}_plan_done";
			if($phase['start_fn'])
				$sql_sel[]="UNIX_TIMESTAMP(r.$ph_start) AS ph{$i}_start";
			$sql_sel[]="UNIX_TIMESTAMP(r.$ph_done) AS ph{$i}_done";
		}
		$sql_str=implode(', ',$sql_sel);

		$query = "SELECT r.rid, r.state, r.$pq_fn, CONCAT('[',IF(p.parent,p.name,'General'),'] ',r.synopsis) AS synopsis, r.class, r.options, r.plan_pri, r.depends_on, r.phase_of_rid, r.complete_pct, r.filed_by, r.customer_id, r.customer_pri, $sql_str, COUNT(r2.rid) AS ph_num FROM $req_tbl AS r LEFT JOIN $req_tbl AS r2 ON r.rid=r2.phase_of_rid, $pq_tbl AS p WHERE r.$pq_fn=p.$pq_fn AND r.$req_in_proj_fn='".$this->inproj."' AND r.$req_proj_id_fn=$proj_id AND ".$this->sql_states_str." GROUP BY r.rid ORDER BY r.plan_pri, r.$req_proj_pri_fn ASC LIMIT 1000";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
		$num_of_rids=mysql_num_rows($result);

		while($row=mysql_fetch_array($result, MYSQL_ASSOC)){
			// Consitancy check
			for($i=1;$i<=count($this->object_params['pm_phases']);$i++){
				$ph_fixer_fn=$this->object_params['pm_phases'][$i-1]['implementer_fn'];
				if($this->inproj=='plan'){
					$of=$row["ph{$i}_ofixer"];
					if(!$of || !YxlUsers::exists($of)){
						$u=$this->pick_safe_owner($i,$row[$pq_fn]);
						$row["ph{$i}_ofixer"]=$u;
						$row[$ph_fixer_fn]=$u;
						_warn_yoxel("Inconsistent DB: unknown $ph_fixer_fn=$of for rid={$row['rid']}.");
					}
					$pf=$row["ph{$i}_fixer"];
					if(!$pf || !YxlUsers::exists($pf)){
						$u=$row["ph{$i}_ofixer"];
						$row["ph{$i}_fixer"]=$u;
						$row['plan_'.$ph_fixer_fn]=$u;
						_warn_yoxel("Inconsistent DB: unknown plan_$ph_fixer_fn=$pf for rid={$row['rid']}.");
					}
				}else{
					$of=$row["ph{$i}_fixer"];
					if(!$of || !YxlUsers::exists($of)){
						$u=$this->pick_safe_owner($i,$row[$pq_fn]);
						$row["ph{$i}_fixer"]=$u;
						$row[$ph_fixer_fn]=$u;
						$query="UPDATE $req_tbl SET $ph_fixer_fn=$u WHERE rid=".$row['rid'];
						//echo "$query <br>\n";
						if(!@mysql_query($query))
							_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
						_warn_yoxel("Corrected inconsistent DB: unknown $ph_fixer_fn=$of for rid={$row['rid']} -> $u.");
					}
				}
			}
			$this->rid2row[$row['rid']]=$row;
			//echo "<span style='color:green;'>plan_order: rid=".$row['rid'].", pri=".$row['plan_pri']." </span><br>\n";
			$depends_on[$row['rid']]=array();
			if($row['ph_num'])
				$is_multiphase[$row['rid']]=TRUE;

			if($row['phase_of_rid']){
				//echo "Checking ".$row['phase_of_rid'];
				// Check if the parent request exists indeed
				if(!simple_query("SELECT COUNT(rid) FROM $req_tbl WHERE rid=".$row['phase_of_rid'])){
					_warning("Parent request #".$row['phase_of_rid']." does not exist. Correcting record for request #".$row['rid'].".");
					$SQL="UPDATE $req_tbl SET phase_of_rid=0 WHERE rid=".$row['rid'];
					//echo "$SQL <br>\n";
					if(!@mysql_query($SQL))
						_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
					$row['phase_of_rid']=0;
				}else
					$is_multiphase[$row['phase_of_rid']]=TRUE;
			}
			if($row['depends_on'])
				array_push($tp,$row['rid']);
			else
				array_push($t,$row['rid']);
		}

		// Multiphase requests are just info titles so lets exclude
		// them from the project. But keep them in a titles array
		if(count($is_multiphase)){
			$sql_str='rid='.implode(' OR rid=',array_keys($is_multiphase));
			$query="UPDATE $req_tbl SET $req_in_proj_fn='no' WHERE $sql_str";
			//echo "$query <br>\n";
			if(!@mysql_query($query))
				_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
			$query="SELECT r.rid, r.state, r.$pq_fn, r.synopsis, r.class FROM $req_tbl AS r, $pq_tbl AS p WHERE r.$pq_fn=p.$pq_fn AND ($sql_str)";
			//echo "$query <br>\n";
			if(!($result=@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
			while($row=mysql_fetch_array($result, MYSQL_ASSOC))
				$this->title_row[$row['rid']]=$row;
		}

		// Remove title requests from initial list
		while($r=array_shift($t))
			if(!isset($is_multiphase[$r]))
				array_push($op,$r);
		//echo "<br>\nblocked:<br>\n"; print_r($tp); echo "<br>\nop:<br>\n"; print_r($op);

		$blockers=array();
		// For all requests that depend on others
		while(count($tp)){
			$t=array();
			// Select 1st level blockers, that do not depend on any other requests
			while($rid=array_shift($tp)){
				// Does $tp[i] depend on any others in $tp?
				$depends_rid=FALSE;
				for($j=0;$j<count($tp);$j++){
					if(($this->rid2row[$rid]['depends_on']==$tp[$j]) || ($this->rid2row[$rid]['depends_on']==$this->rid2row[$tp[$j]]['phase_of_rid'])){
						$depends_rid=$tp[$j];
						break;
					}
				}
				if($depends_rid){
					//echo "$rid depends on $depends_rid <br>\n";
					array_push($t,$rid);
				}else
					array_push($blockers,$rid);
			}

			// Leave the n-level blocker for the next interation
			$tp=$t;
			//echo "<br>\nblockers:<br>\n";print_r($blockers);echo "<br>\ntp:<br>\n";print_r($tp);echo "<br>\n";

			// Insert the blockers into the release list
			while($rid=array_shift($blockers)){
				//echo "<br>\nrid=$rid<br>\n";
				$t=array();
				while($k=array_pop($op)){
					// If blocker depends on $op[k]
					if(($this->rid2row[$rid]['depends_on'] == $k) || ($this->rid2row[$rid]['depends_on'] == $this->rid2row[$k]['phase_of_rid'])){
						//echo "Found blocker $k for $rid <br>\n";
						array_push($op,$k);
						break;
					}
					array_unshift($t,$k);
				}
				//echo "<br>\n op: ";print_r($op);echo "<br>\n t: ";print_r($t);
				$found=FALSE;
				while($k=array_shift($t)){
					// If blocker has lower priority (higher order) than $t[k]
					//echo "$rid=".$this->rid2row[$rid]['plan_pri']." "." $k=".$this->rid2row[$k]['plan_pri'];
					if(($this->rid2row[$rid]['plan_pri'] < $this->rid2row[$k]['plan_pri'])){
						array_unshift($t,$k);
						$found=TRUE;
						//echo "Found pri order $k for $rid <br>\n";
						break;
					}
					array_push($op,$k);
				}

				//echo "<br>\n op: ";print_r($op); echo "<br>\n t: ";print_r($t);echo "<br>\n";
				if($found){
					array_push($op,$rid);
					$op=array_merge($op,$t);
				}else
					array_push($op,$rid);

				//echo "<br>\n op: "; print_r($op); echo "<br>\n";
			}
			//print_r($tp);
			//echo "<br>\n";
		}
		//print_r($op);echo "<br>\n";
		while($rid=current($op)){
			// figure out $dependency list here
			$depends_on=$this->get_dependency_list($rid);
			//echo "<br>\n*** $rid depends on:"; print_r($depends_on); echo "<br>\n";
			if(!isset($is_multiphase[$rid]))
				array_push($this->ordered_plan,array(&$this->rid2row[$rid],$depends_on));
			//else
			//	echo "multiphase: $rid <br>\n";
			next($op);
		}
		return($num_of_rids);
	}

	function get_max_end_chp($rids, &$rid){
		//echo "get_max_end_chp: "; print_r($rids);echo "<br>\n";
		$max_chp=$this->wcal->start_ch();
		$rid=0;
		for($i=0;$i<count($this->ordered_plan);$i++){
			$row = & $this->ordered_plan[$i][0];
			if(count($this->ordered_plan[$i])<3)
			// Times are not computed yet
				continue;

			if(!in_array($row['rid'],$rids))
			// Consider only the list or rids passed as the parameter
				continue;

			$a = &$this->ordered_plan[$i][2];

			// If a blocker is not in release then neither dependent can be
			if($a['is_conflict'])
				$ch=$this->wcal->end_ch();
			else
				$ch=$a['fix_end_ch'];

			//echo "get_max_end_chp: {$row['rid']}, fixer=".$a['fixer']." ".$this->wcal->ch2date("Y/m/j H:00",$ch)."<br>\n";

			if($ch>$max_chp){
				$max_chp=$ch;
				$rid=$row['rid'];
			}
		}
		return $max_chp;
	}

	function ph_start_end_adjust(&$row,$phase,$proj_phase_start_ch,&$ph_start_ch,&$ph_end_ch){

		$ph_start_sm=&$this->object_params['pm_phases'][$phase-1]['start_state_mask'];
                $ph_done_sm=&$this->object_params['pm_phases'][$phase-1]['done_state_mask'];

		$ph_start_tp_fn='ph'.$phase.'_start';
		$ph_end_tp_fn='ph'.$phase.'_done';

		if(($row['state']&$ph_start_sm[$row['class']]) && isset($row[$ph_start_tp_fn]) && $row[$ph_start_tp_fn])
                 	$ph_start_ch=$this->wcal->tp2ch($row[$ph_start_tp_fn],FALSE);

		if(($row['state']&$ph_done_sm[$row['class']]) && $row[$ph_end_tp_fn]){
                 	$ph_end_ch=$this->wcal->tp2ch($row[$ph_end_tp_fn],FALSE);

                        if($row[$ph_start_tp_fn])
                        	$ph_start_ch=$this->wcal->tp2ch($row[$ph_start_tp_fn],FALSE);
                        else
                        	$ph_start_ch=$this->wcal->tp2ch($row[$ph_end_tp_fn],FALSE);

		}elseif($row['state']&$ph_done_sm[$row['class']] && $row[$ph_end_tp_fn]){
                        if($row[$ph_start_tp_fn])
                        	$ph_start_ch=$ph_end_ch=$this->wcal->tp2ch($row[$ph_start_tp_fn],FALSE);
			elseif($proj_phase_start_ch)
				$ph_start_ch=$ph_end_ch=$proj_phase_start_ch;
		}
	}

	function compute_plan_1(){ // ph1 fixing only


		$ph1_dflt_fix_fn=$this->object_params['pm_phases'][0]['dflt_fix_fn'];
		$pq_fn=$this->object_params['pq_fn'];
		$ph1_done_sm=$this->object_params['pm_phases'][0]['done_state_mask'];


		$proj_id=$this->proj_id;
		if(!count($this->ordered_plan))
			$this->plan_order();

		$num_of_rids=0;
		$still_fixing=0;
		$max_fix_tp=0;
		for($i=0;$i<count($this->ordered_plan);$i++){
			$row = &$this->ordered_plan[$i][0];
			$depends_on = $this->ordered_plan[$i][1];
			$auto=$row['options']&REQ_OPT_RPSAUTO;
			$num_of_rids++;

			$conflict=FALSE;
			$ft=$fixer=FALSE;
			$no_fet=FALSE;
			$fix_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;
			$wh_shift_reason=0;
			$move_mode=WH_WASTE_USE;

			array_push($this->log,"compute_plan: rid={$row['rid']} <br>\n");
			//echo "<span style='color:green;'>compute_plan: rid={$row['rid']}, qid={$row['qid']} </span><br>\n";
			$fixer=(($row['ph1_fixer'] && !$auto)?$row['ph1_fixer']:$this->fixer_with_min_chp_atime($row[$pq_fn]));
			if(!$fixer)
				$fixer=$row['ph1_fixer'];

			$ft=$this->fixer_et($row['rid'],$fixer);
			$no_fet=!$this->fixer_has_et($row['rid'],$fixer);

			// Check for dependencies here and adjust whp if necessary
			//
			$ok_start=TRUE;
			//print_r($depends_on);
			if(count($depends_on)){
			// Request depends on others, shift start time
				//echo "{$row['rid']} depends on ";print_r($depends_on);echo "<br>\n";

				// When do the show-stoppers end?
				$end_ch=$this->get_max_end_chp($depends_on, $wh_shift_reason);
				//echo "rid: {$row['rid']}, max_end_chp: ".$this->wcal->ch2date("Y/m/j H:00",$end_ch)."<br>\n"; echo "latest dependency is rid: $wh_shift_reason<br>\n";

				// Translate it back into work hour
				if(($end_wh=$this->wcal->ch2whp($fixer,$end_ch))>=0){
					//echo "new shifted wh: $end_wh <br>\n";

					$dwh=$end_wh-$this->fixer_wh($fixer);
					if($dwh>0){
					// Never decrease work hours pointer !
						//echo "adjusting fixer: $fixer from ",$this->fixer_wh($fixer)," to ",$this->wcal->ch2whp($fixer,$end_ch),"<br>\n";

						// Also register $dwh hrs of waste
						$ra=$this->move_fixer_hp($fixer,$dwh,FALSE,WH_WASTE_REGISTER);
					}
					$move_mode=WH_WASTE_IGNORE;
				}else
					$ok_start=FALSE;
			}

			$fix_start_ch=$plan_ph1_start_ch=$fix_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;

			if($ok_start){
			// If no dependency or dependecy shift went ok
				//$this->wh_debug(); echo "Moving fixer=".YxlUsers::username($fixer)."($fixer) by $ft hours. <br>";
				$ra=$this->move_fixer_hp($fixer,$ft,FALSE,$move_mode);
				$wrk_hrs_exceed=$ra[0];
				$fix_start_ch=$plan_ph1_start_ch=$ra[1];
				$fix_end_ch=$ra[2];

				$this->ph_start_end_adjust($row,1,FALSE,$fix_start_ch,$fix_end_ch);

				//echo "fix_start: ".$this->wcal->ch2date("m/j/Y H:00",$fix_start_ch)," fix_end: ".$this->wcal->ch2date("m/j/Y H:00",$fix_end_ch),"<br>\n";
				if($wrk_hrs_exceed>0)
					$conflict=2;
			}else{
				$conflict=1;
				$wrk_hrs_exceed=$ft;
			}

			//echo $fix_start_date,"<br>\n";
			//echo $dbg_test_start_date,"<br>\n";
			$max_fhp=$this->get_max_fhp();
			$fix_ends_on=$this->wcal->ch2date("m/j/Y H:00", $max_fhp);

			$a=array();
			$a['is_conflict']=$conflict;
			$a['fixer']=$fixer;
			$a['fix_time']=$ft;
			$a['no_fix_estimate']=$no_fet;
			$a['fix_start_ch']=$fix_start_ch;
			$a['plan_ph1_start_ch']=$plan_ph1_start_ch;
			$a['fix_end_ch']=$fix_end_ch;
			$a['wh_shift_reason']=$wh_shift_reason;
			$a['exceed_by_wrk_hrs']=$wrk_hrs_exceed;
			$a['max_fhp']=$max_fhp;
			array_push($this->ordered_plan[$i],$a);
			//print_r($row);
			//echo "<br>\n";

			if(!($row['state']&$ph1_done_sm[$row['class']]))
				$still_fixing++;
			elseif($row['ph1_done']>$max_fix_tp)
				$max_fix_tp=$row['ph1_done'];
		}

		if(!$this->project_info['estimated_fix_end'] && isset($max_fhp))
			$this->project_info['estimated_fix_end']=$this->wcal->ch2date("Y-m-j H:00", $max_fhp);
		//foreach($this->work_trace as $u => $t)
		//	echo "$u: "; print_r($t); echo "<br>\n";
		$this->project_info['gantts_limit']=$this->wcal->ch2date("Y-m-j",$this->project_info['proj_deadline_ch']+7*HOURS_IN_DAY);
		return($num_of_rids);
	}

	function compute_plan_1t(){
		$ph1_dflt_fix_fn=$this->object_params['pm_phases'][0]['dflt_fix_fn'];
		$pq_fn=$this->object_params['pq_fn'];
		$ph1_start_sm=$this->object_params['pm_phases'][0]['start_state_mask'];
		$ph1_done_sm=$this->object_params['pm_phases'][0]['done_state_mask'];

		$proj_id=$this->proj_id;
		if(!count($this->ordered_plan))
			$this->plan_order();

		$num_of_rids=0;
		$still_fixing=0;
		$max_fix_tp=0;

		$proj_impl_start_ch=$this->wcal->date2ch(date_sql2norm($this->project_info['implement_start_date']));
		$max_fhp=$this->wcal->date2ch();
		$rid2ch=array();
		for($i=0;$i<count($this->ordered_plan);$i++){
			$row = &$this->ordered_plan[$i][0];
			//echo "row: ";print_r($row); echo "<br>\n";
			$depends_on = $this->ordered_plan[$i][1];
			$auto=$row['options']&REQ_OPT_RPSAUTO;
			$num_of_rids++;
			$conflict=FALSE;

			$ft=FALSE;
			$fixer=FALSE;
			$no_fet=FALSE;
			$fix_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;
			$wh_shift_reason=0;
			$move_mode=WH_WASTE_USE;

			array_push($this->log,"compute_plan_1t: rid={$row['rid']} <br>\n");
			//echo "<span style='color:green;'>compute_plan: rid={$row['rid']}, pid={$row[$pq_fn]} </span><br>\n";
			$fixer=$row['ph1_fixer'];

			if($this->ts_user_share($fixer)<=0){
				_warning("Warning: ".YxlUsers::name($fixer)." has zero time allocated for this project. Time esitmation for rid ".$row['rid']." is incorrect!");
				$ft=$this->fixer_et($row['rid'],$fixer,FALSE);
			}else
				$ft=$this->fixer_et($row['rid'],$fixer);

			$no_fet=!$this->fixer_has_et($row['rid'],$fixer);

			// Check for dependencies here and adjust whp if necessary
			if(count($depends_on)){
			// Request depends on others, shift start time
				//echo "{$row['rid']} depends on ";print_r($depends_on);echo "<br>\n";

				// When do the show-stoppers end?
				$end_ch=$this->get_max_end_chp($depends_on, $wh_shift_reason);
				//echo "rid: {$row['rid']}, max_end_chp: ".$this->wcal->ch2date("Y/m/j H:00",$end_ch)."<br>\n"; echo "latest dependency is rid: $wh_shift_reason<br>\n";

				if($this->wcal->total_whs($fixer)==0){
					$start_ch=$this->wcal->start_ch();
					$this->wcal->extend($fixer,$start_ch);
				}

				// Translate it back into work hour
				if(($end_wh=$this->wcal->ch2whp($fixer,$end_ch,FALSE))>=0){
					//echo "new shifted wh: $end_wh <br>\n";

					$dwh=$end_wh-$this->fixer_wh($fixer);
					if($dwh>0){
					// Never decrease work hours pointer !
						//echo "adjusting fixer: $fixer from ",$this->fixer_wh($fixer)," to ",$this->wcal->ch2whp($fixer,$end_ch),"<br>\n";

						// Also register $dwh hrs of waste
						$ra=$this->move_fixer_hp($fixer,$dwh,FALSE,WH_WASTE_REGISTER);
					}
					$move_mode=WH_WASTE_IGNORE;
				}else
					_die(__FILE__.':'.__LINE__.' Not supposed to happen: exceeded the calendar.');
			}

			$fix_start_ch=$plan_ph1_start_ch=$fix_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;

			//$this->wh_debug(); echo "Moving fixer=".YxlUsers::username[$fixer]."($fixer) by $ft hours. <br>";
			$ra=$this->move_fixer_hp($fixer,$ft,TRUE,$move_mode);
			//echo "ra: ";print_r($ra); echo "<br>\n";
			if($ra[0]>0)
				_die(__FILE__.':'.__LINE__.' Not supposed to happen: exceeded the calendar.');

			//print_r($row); echo "<br>\n";
			if(!($row['state']&$ph1_done_sm[$row['class']]))
				$still_fixing++;
			// Catching latest fixed_date
			elseif($row['ph1_done']>$max_fix_tp)
				$max_fix_tp=$row['ph1_done'];

			//echo "fix_start_pre ({$row['rid']}): ".$this->wcal->ch2date("m/j/Y H:00",$ra[1])," fix_end: ".$this->wcal->ch2date("m/j/Y H:00",$ra[2]),"<br>\n";

			$fix_start_ch=$plan_ph1_start_ch=$ra[1];
			$fix_end_ch=$ra[2];
			$this->ph_start_end_adjust($row,1,$proj_impl_start_ch,$fix_start_ch,$fix_end_ch);

			//echo "fix_start_post ($row[0]): ".$this->wcal->ch2date("m/j/Y H:00",$fix_start_ch)," fix_end: ".$this->wcal->ch2date("m/j/Y H:00",$fix_end_ch),"<br>\n";

			//echo $fix_start_date,"<br>\n";
			$max_fhp=$this->get_max_fhp();
			$fix_ends_on=$this->wcal->ch2date("m/j/Y H:00", $max_fhp);
			//$this->wh_debug(); echo "{$row[0]} +++fixer=$fixer ft=$ft cmax_fhp=$max_fhp ($fix_ends_on) +++tester=$tester tt=$tt cmax_thp=$max_thp ($test_ends_on)<br>\n";

			$a=array();
			$a['is_conflict']=FALSE;
			$a['fixer']=$fixer;
			$a['fix_time']=$ft;
			$a['no_fix_estimate']=$no_fet;
			$a['fix_start_ch']=$fix_start_ch;
			$a['plan_ph1_start_ch']=$plan_ph1_start_ch;
			$a['fix_end_ch']=$fix_end_ch;
			$a['wh_shift_reason']=$wh_shift_reason;
			$a['exceed_by_wrk_hrs']=0;
			$a['max_fhp']=$max_fhp;
			array_push($this->ordered_plan[$i],$a);
		}

		$this->project_info['gantts_limit']=$this->project_info['proj_deadline_ch']>$max_fhp?$this->wcal->ch2date("Y-m-j",$this->project_info['proj_deadline_ch']+7*HOURS_IN_DAY,FALSE):$this->wcal->ch2date("Y-m-j",$max_fhp+7*HOURS_IN_DAY,FALSE);
		//echo $this->project_info['gantts_limit']," <br>\n";
		if(!$still_fixing)
			$actual_max_fhp=$this->wcal->tp2ch($max_fix_tp);

		//echo "max_fhp(nfix=$still_fixing): ".$this->wcal->ch2date("Y-m-j H:00",$max_fhp)." <br>\n";
		return($num_of_rids);
	}

	function compute_plan_2($gtrace=FALSE, $mode=0){
		$pq_fn=$this->object_params['pq_fn'];
		$ph1_dflt_fix_fn=$this->object_params['pm_phases'][0]['dflt_fix_fn'];
		$ph1_done_sm=$this->object_params['pm_phases'][0]['done_state_mask'];
		$ph2_dflt_fix_fn=$this->object_params['pm_phases'][1]['dflt_fix_fn'];
		$ph2_done_sm=$this->object_params['pm_phases'][1]['done_state_mask'];

		$rel_id=$this->rel_id;
		if(!count($this->ordered_plan))
			$this->plan_order();

		$this->work_trace=array();
		$num_of_rids=$still_fixing=$max_fix_tp=0;
		for($i=0;$i<count($this->ordered_plan);$i++){
			$row = &$this->ordered_plan[$i][0];
			$depends_on = $this->ordered_plan[$i][1];
			$auto=$row['options']&REQ_OPT_RPSAUTO;
			$num_of_rids++;

			$conflict=FALSE;
			$tt=$ft=FALSE;
			$tester=$fixer=FALSE;
			$no_fet=$no_tet=FALSE;
			$fix_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;
			$wh_shift_reason=0;
			$move_mode=WH_WASTE_USE;

			array_push($this->log,"compute_plan: rid={$row['rid']} <br>\n");
			//print_r($row);_warning("compute_plan: rid={$row['rid']}, pid={$row['pid']}");

			$fixer=(($row['ph1_fixer'] && !$auto)?$row['ph1_fixer']:$this->fixer_with_min_chp_atime($row[$pq_fn]));
			if(!$fixer)
				$fixer=$row['ph1_fixer'];

			$ft=$this->fixer_et($row['rid'],$fixer);
			$no_fet=!$this->fixer_has_et($row['rid'],$fixer);

			// Check for dependencies here and adjust whp if necessary
			//
			$ok_start=TRUE;
			if(count($depends_on)){
			// Request depends on others, shift start time
				//echo "{$row['rid']} depends on ";print_r($depends_on);echo "<br>\n";

				// When do the show-stoppers end?
				$end_ch=$this->get_max_end_chp($depends_on, $wh_shift_reason);
				//echo "rid: {$row['rid']}, max_end_chp: ".$this->wcal->ch2date("Y/m/j H:00",$end_ch)."<br>\n"; echo "latest dependency is rid: $wh_shift_reason<br>\n";

				// Translate it back into work hour
				if(($end_wh=$this->wcal->ch2whp($fixer,$end_ch))>=0){
					$dwh=$end_wh-$this->fixer_wh($fixer);
					//echo "new shifted wh: $end_wh dwh=$dwh<br>\n";

					if($dwh>0){
					// Never decrease work hours pointer !
						//echo "adjusting fixer: $fixer from ",$this->fixer_wh($fixer)," to ",$this->wcal->ch2whp($fixer,$end_ch),"<br>\n";
						// Also register $dwh hrs of waste!
						$ra=$this->move_fixer_hp($fixer,$dwh,FALSE,WH_WASTE_REGISTER);
					}
					$move_mode=WH_WASTE_IGNORE;
				}else
					$ok_start=FALSE;
			}

			$fix_start_ch=$plan_ph1_start_ch=$fix_end_ch=FALSE;
			$dbg_test_start_ch=FALSE;
			$dbg_test_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;

			if($ok_start){
			// If no dependency or dependecy shift went ok
			//$this->wh_debug(); echo "Moving fixer=".YxlUsers::username[$fixer]."($fixer) by $ft hours. <br>";
				$ra=$this->move_fixer_hp($fixer,$ft,FALSE,$move_mode);
				$wrk_hrs_exceed=$ra[0];
				$fix_start_ch=$plan_ph1_start_ch=$ra[1];
				$fix_end_ch=$ra[2];
				$mvm=$ra[3];

                        	$this->ph_start_end_adjust($row,1,FALSE,$fix_start_ch,$fix_end_ch);

				//echo "fix_start: ".$this->wcal->ch2date("m/j/Y H:00",$fix_start_ch)," fix_end: ".$this->wcal->ch2date("m/j/Y H:00",$fix_end_ch)," whe=$wrk_hrs_exceed, mvm=$mvm<br>\n";
				if($wrk_hrs_exceed<=0){
					$tester=(($row['ph2_fixer'] && !$auto)?$row['ph2_fixer']:$this->tester_with_min_chp_atime($row[$pq_fn]));
					if(!$tester)
						$tester=$row['ph2_fixer'];
					$tt=$this->tester_et($row['rid'],$tester);
					$no_tet=!$this->tester_has_et($row['rid'],$tester);
					//$this->wh_debug(); echo "Moving tester=".YxlUsers::username($tester)."($tester) by $tt hours (rid={$row['rid']}). <br>";
					$ra=$this->move_tester_hp($tester,$tt);
					$wrk_hrs_exceed=$ra[0];
					$dbg_test_start_ch=$ra[1];
					$dbg_test_end_ch=$ra[2];
					//echo "rmthp: $wrk_hrs_exceed <br>\n";
					if($wrk_hrs_exceed>0){
						$conflict=3;
						$this->move_fixer_hp($fixer,-$ft,FALSE,$mvm);
					}
				}else
					$conflict=2;
			}else{
				$conflict=1;
				$wrk_hrs_exceed=$ft;
			}
			//echo $fix_start_date,"<br>\n";
			//echo $dbg_test_start_date,"<br>\n";
			//$this->wh_debug(); echo $row['rid']."===<br>\n";

			$max_fhp=$this->get_max_fhp();
			$fix_ends_on=$this->wcal->ch2date("m/j/Y H:00", $max_fhp,TRUE,__LINE__);
			//echo "+++++ fixer=$fixer ft=$ft cmax_fhp=$max_fhp ($fix_ends_on)<br>\n";

			$max_thp=$this->get_max_thp();
			$test_ends_on=$this->wcal->ch2date("m/j/Y H:00", $max_thp,TRUE,__LINE__);
			//echo "+++tester=$tester tt=$tt cmax_thp=$max_thp ($test_ends_on)<br>\n";

			$a=array();
			$a['is_conflict']=$conflict;
			$a['tester']=$tester;
			$a['fixer']=$fixer;
			$a['test_time']=$tt;
			$a['fix_time']=$ft;
			$a['no_test_estimate']=$no_tet;
			$a['no_fix_estimate']=$no_fet;
			$a['fix_start_ch']=$fix_start_ch;
			$a['plan_ph1_start_ch']=$plan_ph1_start_ch;
			$a['dbg_test_start_ch']=$dbg_test_start_ch;
			$a['fix_end_ch']=$fix_end_ch;
			$a['dbg_test_end_ch']=$dbg_test_end_ch;
			$a['wh_shift_reason']=$wh_shift_reason;
			$a['exceed_by_wrk_hrs']=$wrk_hrs_exceed;
			$a['max_fhp']=$max_fhp;
			$a['max_thp']=$max_thp;
			array_push($this->ordered_plan[$i],$a);
			//print_r($row);
			//echo "<br>\n";

			if(!($row['state']&$ph1_done_sm[$row['class']]))
				$still_fixing++;
			elseif($row['ph1_done']>$max_fix_tp)
				$max_fix_tp=$row['ph1_done'];
		}

		if(!$this->project_info['estimated_fix_end'] && isset($max_fhp))
			$this->project_info['estimated_fix_end']=$this->wcal->ch2date("Y-m-j H:00", $max_fhp);
		if(!$this->project_info['estimated_test_end'] && isset($max_thp))
			$this->project_info['estimated_test_end']=$this->wcal->ch2date("Y-m-j H:00", $max_thp);
		//foreach($this->work_trace as $u => $t)
		//	echo "$u: "; print_r($t); echo "<br>\n";
		$this->project_info['gantts_limit']=$this->wcal->ch2date("Y-m-j",$this->project_info['proj_deadline_ch']+7*HOURS_IN_DAY);

		// Array of current work hour pointer for all testers
		$twp=array();
		for($i=0;$i<count($this->ordered_plan);$i++){
			$row=&$this->ordered_plan[$i][0];
			$a = &$this->ordered_plan[$i][2];
			$conflict=$a['is_conflict'];
			if($conflict)
				continue;

			$tester=$a['tester'];
			$tt=$a['test_time'];
			if(!isset($twp[$tester])){
				$twp[$tester]=$this->wcal->ch2whp($tester,$max_fhp);
				//echo "$tester starts at ".$this->wcal->wh2date("Y/m/j(H)",$tester,$twp[$tester])."<br>\n";
			}

			$start_ch=$this->wcal->whp2ch($tester,$twp[$tester]);
			$twp[$tester]+=$tt;
			//echo "!".$twp[$tester]." ".$this->testers_required_hrs[$tester]."<br>\n";
			$end_ch=$this->wcal->whp2ch($tester,$twp[$tester]);
			$a['test_start_ch']=$start_ch;
			$a['test_end_ch']=$end_ch;
			//echo "tester=$tester, rid=$row[0], sch=$start_ch, ech=$end_ch <br>\n".$this->wcal->ch2date("Y-m-j H:00",$start_ch).",".$this->wcal->ch2date("Y-m-j H:00",$end_ch);
		}
		return($num_of_rids);
	}

	function compute_plan_2t(){
		$req_tbl=$this->object_params['requests_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$ph1_dflt_fix_fn=$this->object_params['pm_phases'][0]['dflt_fix_fn'];
		$ph1_start_sm=$this->object_params['pm_phases'][0]['start_state_mask'];
		$ph1_done_sm=$this->object_params['pm_phases'][0]['done_state_mask'];
		$ph2_dflt_fix_fn=$this->object_params['pm_phases'][1]['dflt_fix_fn'];
		$ph2_done_sm=$this->object_params['pm_phases'][1]['done_state_mask'];

		$rel_id=$this->proj_id;
		if(!count($this->ordered_plan))
			$this->plan_order();
		$this->work_trace=array();
		$num_of_rids=0;$still_fixing=0;
		$max_fix_tp=0;
		//echo date("m/j/Y H:00", $max_fix_tp)." $still_fixing <br>\n";
		$proj_impl_start_ch=$this->wcal->date2ch(date_sql2norm($this->project_info['implement_start_date']));
		$max_fhp=$this->wcal->date2ch();
		$max_thp=$this->wcal->date2ch();
		$rid2ch=array();
		for($i=0;$i<count($this->ordered_plan);$i++){
			$row = & $this->ordered_plan[$i][0];
			//echo "row: ";print_r($row); echo "<br>\n";
			$depends_on = $this->ordered_plan[$i][1];
			$auto=$row['options']&REQ_OPT_RPSAUTO;
			$num_of_rids++;

			$conflict=FALSE;
			$tt=$ft=FALSE;
			$tester=$fixer=FALSE;
			$no_fet=$no_tet=FALSE;
			$fix_end_ch=FALSE;
			$test_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;
			$wh_shift_reason=0;
			$move_mode=WH_WASTE_USE;

			array_push($this->log,"compute_plan: rid=".$row['rid']." <br>\n");
			//echo "<span style='color:green;'>compute_plan: rid=".$row['rid'].", pid=".$row[$pq_fn]." </span><br>\n";
			$fixer=$row['ph1_fixer'];

			if($this->ts_user_share($fixer)<=0){
				_warning("Warning: ".YxlUsers::name($fixer)." has zero time allocated for this project. Time esitmation for rid ".$row['rid']." is incorrect!");
				$ft=$this->fixer_et($row['rid'],$fixer,FALSE);
			}else
				$ft=$this->fixer_et($row['rid'],$fixer);

			$no_fet=!$this->fixer_has_et($row['rid'],$fixer);

			// Check for dependencies here and adjust whp if necessary
			//
			if(count($depends_on)){
			// Request depends on others, shift start time
				//echo "{$row['rid']} depends on ";print_r($depends_on);echo "<br>\n";

				// When do the show-stoppers end?
				$end_ch=$this->get_max_end_chp($depends_on, $wh_shift_reason);
				//echo "rid: {$row['rid']}, max_end_chp: ".$this->wcal->ch2date("Y/m/j H:00",$end_ch)."<br>\n"; echo "latest dependency is rid: $wh_shift_reason<br>\n";

				if($this->wcal->total_whs($fixer)==0){
					$start_ch=$this->wcal->start_ch();
					$this->wcal->extend($fixer,$start_ch);
				}

				// Translate it back into work hour
				$end_wh=$this->wcal->ch2whp($fixer,$end_ch);
				//echo "new shifted wh: $end_wh fixer=$fixer <br>\n";

				if($end_wh>=0){
					$dwh=$end_wh-$this->fixer_wh($fixer);

					if($dwh>0){
					// Never decrease work hours pointer !
						//echo "adjusting fixer: $fixer from ",$this->fixer_wh($fixer)," to ",$this->wcal->ch2whp($fixer,$end_ch),"<br>\n";

						// Also register $dwh hrs of waste!
						$ra=$this->move_fixer_hp($fixer,$dwh,FALSE,WH_WASTE_REGISTER);
					}
					$move_mode=WH_WASTE_IGNORE;
				}else
					_die(__FILE__.':'.__LINE__.' Not supposed to happen: exceeded the calendar.');
			}

			$fix_start_ch=$plan_ph1_start_ch=$fix_end_ch=FALSE;
			$test_start_ch=FALSE;
			$test_end_ch=FALSE;
			$wrk_hrs_exceed=FALSE;

			//$this->wh_debug(); echo "Moving fixer=".YxlUsers::username[$fixer]."($fixer) by $ft hours. <br>";
			$ra=$this->move_fixer_hp($fixer,$ft,TRUE,$move_mode);

			if($ra[0]>0)
				_die(__FILE__.':'.__LINE__.' Not supposed to happen: exceeded the calendar.');

			// Need to know if there are still any requests not done with ph1
			if(!($row['state']&$ph1_done_sm[$row['class']]))
				$still_fixing++;
			// Catching the latest actual ph1 fixed/done date
			elseif($row['ph1_done']>$max_fix_tp)
				$max_fix_tp=$row['ph1_done'];

			//echo "fix_start: ".$this->wcal->ch2date("m/j/Y H:00",$ra[1])," fix_end: ".$this->wcal->ch2date("m/j/Y H:00",$ra[2]),"<br>\n";

			$fix_start_ch=$plan_ph1_start_ch=$ra[1];
			$fix_end_ch=$ra[2];
			$this->ph_start_end_adjust($row,1,$proj_impl_start_ch,$fix_start_ch,$fix_end_ch);

			//echo "fix_start_post ({$row['rid']}): ".$this->wcal->ch2date("m/j/Y H:00",$fix_start_ch)," fix_end: ".$this->wcal->ch2date("m/j/Y H:00",$fix_end_ch),"<br>\n";
			$tester=$row['ph2_fixer'];

			if($this->ts_user_share($tester)<=0){
				_warning("Warning: ".YxlUsers::name($tester)." has zero time allocated for this project. Time esitmation for rid ".$row['rid']." is incorrect!");
				$tt=$this->tester_et($row['rid'],$tester,FALSE);
			}else
				$tt=$this->tester_et($row['rid'],$tester);

			$no_tet=!$this->tester_has_et($row['rid'],$tester);
			//$this->wh_debug(); echo "Moving tester=".YxlUsers::username[$tester]."($tester) by $tt hours (rid={$row['rid']}). <br>";
			$ra=$this->move_tester_hp($tester,$tt,TRUE);
			if($ra[0]>0)
				_die(__FILE__.':'.__LINE__.' Not supposed to happen: exceeded the calendar.');

			//echo $fix_start_date,"<br>\n";
			//echo $test_start_date,"<br>\n";
			$max_fhp=$this->get_max_fhp();
			$fix_ends_on=$this->wcal->ch2date("m/j/Y H:00", $max_fhp);
			$max_thp=$this->get_max_thp();
			$test_ends_on=$this->wcal->ch2date("m/j/Y H:00", $max_thp);
			//$this->wh_debug(); echo "{$row[0]} +++fixer=$fixer ft=$ft cmax_fhp=$max_fhp ($fix_ends_on) +++tester=$tester tt=$tt cmax_thp=$max_thp ($test_ends_on)<br>\n";

			$a=array();
			$a['is_conflict']=FALSE;
			$a['fixer']=$fixer;
			$a['tester']=$tester;
			$a['fix_time']=$ft;
			$a['test_time']=$tt;
			$a['no_fix_estimate']=$no_fet;
			$a['no_test_estimate']=$no_tet;
			$a['fix_start_ch']=$fix_start_ch;
			$a['plan_ph1_start_ch']=$plan_ph1_start_ch;
			$a['fix_end_ch']=$fix_end_ch;
			$a['wh_shift_reason']=$wh_shift_reason;
			$a['exceed_by_wrk_hrs']=0;
			$a['max_fhp']=$max_fhp;
			$a['max_thp']=$max_thp;
			array_push($this->ordered_plan[$i],$a);

			// If verified on a certain ph2 done date, save the date
			if($row['state']&$ph2_done_sm[$row['class']] && $row['ph2_done'])
				$rid2ch[$tester][$row['rid']]=$this->wcal->tp2ch($row['ph2_done'],FALSE);
		}

		// Sort by date per user
		foreach(array_keys($rid2ch) as $u)
			asort($rid2ch[$u]);
		//echo "rid2ch: ";print_r($rid2ch); echo "<br>\n";

		$this->project_info['gantts_limit']=$this->project_info['proj_deadline_ch']>$max_thp?$this->wcal->ch2date("Y-m-j",$this->project_info['proj_deadline_ch']+7*HOURS_IN_DAY):$this->wcal->ch2date("Y-m-j",$max_thp+7*HOURS_IN_DAY);
		//echo $this->project_info['gantts_limit']," <br>\n";
		if(!$still_fixing)
			$actual_max_fhp=$this->wcal->tp2ch($max_fix_tp);

		//echo "max_fhp(nfix=$still_fixing): ".$this->wcal->ch2date("Y-m-j H:00",$max_fhp)." <br>\n";
		$twp=array();
		$still_testing=0;$max_test_tp=0;
		$rel_test_start_ch=$this->project_info['state']=='test'?$this->wcal->date2ch(date_sql2norm($this->project_info['test_start_date'])):$max_fhp;
		for($i=0;$i<count($this->ordered_plan);$i++){
			$row=&$this->ordered_plan[$i][0];
			$a = &$this->ordered_plan[$i][2];

			// Need to know if there are still any requests not done with ph2
			if(!($row['state']&$ph2_done_sm[$row['class']]))
				$still_testing++;
			// Catching latest ph2 done date
			elseif($row['ph2_done']>$max_test_tp)
				$max_test_tp=$row['ph2_done'];

			$tester=$a['tester'];
			$tt=$a['test_time'];
			//echo "rid={$row['rid']} tester=$tester tt=$tt<br>\n";

			if(!isset($twp[$tester])){
			// Initialize tester work hour pointer for the tester
			// At the end of the release ph1 date
				$twp[$tester]=$this->wcal->ch2whp($tester,$max_fhp,FALSE);
				//echo "rid={$row['rid']} $tester starts at ".$this->wcal->wh2date("Y-m-j H:00",$tester,$twp[$tester]).$this->wcal->ch2date("Y-m-j H:00",$max_fhp)."<br>\n";
			}


			$start_ch=$this->wcal->whp2ch($tester,$twp[$tester]);
			$twp[$tester]+=$tt;
			//echo "start: rid={$row['rid']} ".$twp[$tester]." requires: ".$this->testers_required_hrs[$tester]."<br>\n";
			$end_ch=$this->wcal->whp2ch($tester,$twp[$tester]);
			$a['test_start_ch']=$start_ch;
			$a['test_end_ch']=$end_ch;
			//echo $this->wcal->ch2date("m/j/Y H:00", $max_thp)." $still_testing <br>\n";
			//echo "pre: rid={$row['rid']} tester=$tester, rid=$row[0], sch=".$this->wcal->ch2date("Y-m-j H:00",$start_ch).", ech=".$this->wcal->ch2date("Y-m-j H:00",$end_ch)." $row[16] <br>\n";


			if($row['state']&$ph2_done_sm[$row['class']] && $row['ph2_done']){
			// For a request that is done with ph2 and has the date for it
			// The issue with ph2 is that there probably no ph2 start state
			// in our workflow so lets make the date of that ourselves

				//echo "verified: $row[0] <br>\n";

				// Lets set ph2_start_ch to ph2_end_ch of the previous rid
				// So lets find that previous rid first
				// Previous based on the order of our release requests
				$prev_k=0;
				$fech=$this->wcal->tp2ch($row['ph1_done'],FALSE);
				$tech=$this->wcal->tp2ch($row['ph2_done'],FALSE);
				//echo "tech rid={$row['rid']}: ".$this->wcal->ch2date("Y-m-j H:00",$tech)." rel_test_start: ".$this->wcal->ch2date("Y-m-j H:00",$rel_test_start_ch)." <br>\n";

				$prev_test_end=$tech>=$rel_test_start_ch?$rel_test_start_ch:$fech;
				$a['test_end_ch']=$this->wcal->tp2ch($row['ph2_done'],FALSE);
				foreach(array_keys($rid2ch[$tester]) as $k){
					//echo "$k -> ".$this->wcal->ch2date("m/j/Y H:00", $rid2ch[$tester][$k])." <br>\n";
					if(($k==$row['rid']) && isset($rid2ch[$tester][$prev_k])){
						$prev_test_end=$rid2ch[$tester][$prev_k];
						break;
					}
					$prev_k=$k;
				}

				if(($tech>=$rel_test_start_ch) && ($prev_test_end<$rel_test_start_ch))
				// Adjust if previous request ph2 done date is earlier
				// then release ph2 start
					$prev_test_end=$rel_test_start_ch;

				$a['test_start_ch']=$prev_test_end;
			}elseif($row['state']&$ph2_done_sm[$row['class']] && !$row['ph2_done'])
			// For a request that is done with ph2 but date is not knows
				$a['test_start_ch']=$a['test_end_ch']=$rel_test_start_ch;

			if(!($row['state']&$ph2_done_sm[$row['class']])){
			// For requests that are not done with ph2
			// block calendar time slots currently estimated
			// for implementation and testing (ph1 & ph2)
			// This will tell other releases in planning stage to
			// avoid these slots.
				$query="UPDATE $req_tbl SET cal_fix_start='".$this->wcal->ch2date("Y-m-j H:00",$a['fix_start_ch'],FALSE)."', cal_fix_end='".$this->wcal->ch2date("Y-m-j H:00",$a['fix_end_ch'],FALSE)."', cal_test_start='".$this->wcal->ch2date("Y-m-j H:00",$a['test_start_ch'],FALSE)."', cal_test_end='".$this->wcal->ch2date("Y-m-j H:00",$a['test_end_ch'],FALSE)."' WHERE rid=".$row['rid'];
				//echo "$query <br>\n";
				if(!@mysql_query($query))
					_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
			}
			//echo "post: tester=$tester, rid=$row[0], sch=".$this->wcal->ch2date("Y-m-j H:00",$a['test_start_ch']).", ech=".$this->wcal->ch2date("Y-m-j H:00",$a['test_end_ch'])." <br>\n";
		}

		if(!$still_testing)
			$actual_max_thp=$this->wcal->tp2ch($max_test_tp);

		return($num_of_rids);
	}

	function build_title_info($title_rid, &$title_info, $start_tp, $end_tp, $title_synop){
		if($title_rid && !isset($title_info[$title_rid])){
			//echo "build_title_info: $title_rid, ".date("Y-m-j H:00",$start_tp).", ".date("Y-m-j H:00",$end_tp).", $title_synop <br>\n";
			$title_info[$title_rid]=array("$title_rid: $title_synop",date("Y-m-j H:00",$start_tp),date("Y-m-j H:00",$end_tp));

			// Just save ch here
			$this->title_ft_ch[$title_rid]=array($start_tp,$end_tp);
		}elseif($title_rid){
			if($start_tp<$this->title_ft_ch[$title_rid][0]){
				$this->title_ft_ch[$title_rid][0]=$start_tp;
				$title_info[$title_rid][1]=date("Y-m-j H:00",$start_tp);
			}
			if($end_tp>$this->title_ft_ch[$title_rid][1]){
				$this->title_ft_ch[$title_rid][1]=$end_tp;
				$title_info[$title_rid][2]=date("Y-m-j H:00",$end_tp);
			}
			//echo "build_title_info2: "; print_r($title_info[$title_rid]); echo "<br>\n";
		}
	}

	function phN_done_requests($phN){

		if(!isset($this->sql_done_states_str[$phN]))
                        _die(__FILE__.':'.__LINE__." phase$phN is not defined.");

		$proj_id=$this->proj_id;
		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$ft_nrid=simple_query("SELECT COUNT(r.rid) FROM $req_tbl AS r WHERE r.$req_in_proj_fn='yes' AND r.$req_proj_id_fn=$proj_id AND ".$this->sql_done_states_str[$phN]);
		//echo $this->sql_done_states_str[$phN];

		return($ft_nrid);
	}

	function phN_start_state_mask($phN,$class){
		if(!isset($this->object_params['pm_phases'][$phN]))
                        _die(__FILE__.':'.__LINE__." phase$phN is not defined.");
		return $this->object_params['pm_phases'][$phN]['start_state_mask'][$class];
	}

	function phN_done_state_mask($phN,$class){
		if(!isset($this->object_params['pm_phases'][$phN]))
                        _die(__FILE__.':'.__LINE__." phase$phN is not defined.");
		return $this->object_params['pm_phases'][$phN]['done_state_mask'][$class];
	}

	function construct_sql_states_str($class2sm){
		if(!$class2sm || !count($class2sm))
			return $this->sql_states_str;

		$proj_cm=$this->object_params['proj_class_mask'];

		$sql=array(); $sql_str='';
		foreach(array_keys($this->rts->classes['options']) as $c){
			if(!($c&$proj_cm)) continue;
                        array_push($sql,"(r.class=$c AND r.state&".$class2sm[$c].")");
		}

                if(count($sql))
			$sql_str="(".implode(' OR ',$sql).")";

		return $sql_str;
	}

	function construct_sql_success_states_str($class2sm){
		if(!$class2sm || !count($class2sm))
			return $this->sql_success_states_str;
		$sql=array(); $sql_str='';
		foreach(array_keys($this->rts->classes['options']) as $c)
                        array_push($sql,"(r.class=$c AND r.state&".$this->last_state_mask[$c]." AND r.state&".$class2sm[$c].")");
                if(count($sql))
			$sql_str="(".implode(' OR ',$sql).")";
		return $sql_str;
	}

	function is_last_state($class,$state){
		$req_last_sm=$this->object_params['req_last_state_mask'];
		return($state&$req_last_sm[$class]);
	}

	function valid_proj_ids_str($pqid=FALSE){
		$pq_fn=$this->object_params['pq_fn'];
		if(!$pqid)
			$pqid=$this->project_info[$pq_fn];

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];

                $query="SELECT $proj_id_fn FROM $proj_tbl WHERE $pq_fn=$pqid";
		if(!($result=@mysql_query($query)))
                	_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl table.");

		$proj_ids_str='';
                while($row=mysql_fetch_array($result, MYSQL_NUM))
                	$proj_ids_str.=','.$row[0];

		return substr($proj_ids_str,1);
	}

	function unscheduled_project_requests($phN=FALSE){
		if(!$phN)
                	$ph=$this->current_phase();

                if(!$phN)
			$phN=count($this->object_params['pm_phases']);

		if(!isset($this->object_params['pm_phases'][$phN-1]))
                        _die(__FILE__.':'.__LINE__." phase$phN is not defined.");

		$sql_state_str=$this->construct_sql_states_str($this->object_params['pm_phases'][$phN-1]['done_state_mask']);

		$pq_fn=$this->object_params['pq_fn'];
		$pqid=$this->project_info[$pq_fn];
		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->proj_id;

		$req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$proj_ids_str=$this->valid_proj_ids_str();
		$query=$this->rts->request_query_sqlstr(FALSE);
		$query.=" AND (r.$pq_fn=$pqid OR c.parent=$pqid) AND $sql_state_str AND r.$req_in_proj_fn!='no' AND (!(r.$req_proj_id_fn IN ($proj_ids_str)) OR r.$req_proj_id_fn IS NULL) ORDER BY r.rid";
		//echo "$query <br>\n"; die();
		if(!($result=@mysql_query($query)))
                	_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		$rids='';
                while($row=mysql_fetch_array($result, MYSQL_ASSOC))
			$rids.=','.$row['rid'];

		if(!$rids)
			return;

		// 03.11.08: bug578 reverting in_proj from 'maybe' back to 'yes'
		$query="UPDATE $req_tbl SET $req_proj_id_fn=$proj_id, $req_in_proj_fn='yes' WHERE rid IN(".substr($rids,1).")";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
                	_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");
	}

	var $ts_start_date;

	// % of time a user spends on critical requests
	var $ts_ucts=array();

	// % of time a user has for non critical requests
	var $ts_ushare=array();

	// List of products for which release is being planned, implemented
	var $ts_pqids_in_work=array();

	// Users considered for time sharing
	var $ts_users=array();

	// List of products a user is potentially engaged in
	var $ts_user_in_pqids=array();

	var $ts_mgr_in_pqs=array();

	function init_timeshare($overlap_mode){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
		$proj_rts_opt=$this->rts->get_proj_rts_opt();
		$pq_fn=$this->object_params['pq_fn'];
		$pnets_tbl=$this->object_params['pnets_tbl'];
		$pnet_type_ts=$this->object_params['pnet_type_ts'];
		$ph2=count($this->object_params['pm_phases'])>1;

		$ts_specs=array();
		$ph1_fixer_spec=$this->object_params['pm_phases'][0]['implementer_spec'];
		arr_push($ts_specs,$ph1_fixer_spec);
		if(count($this->object_params['pm_phases'])>1){
			$ph2_fixer_spec=$this->object_params['pm_phases'][1]['implementer_spec'];
			arr_push($ts_specs,$ph2_fixer_spec);
		}

		$pqid=$this->project_info[$pq_fn];

		if($overlap_mode){
			// This is the mode for cross product time sharing assignment
			// Multiple pqids are examined for resource conflicts
			// Main rel_id should be in planning mode, others may have not even started yet
			if($this->project_info['state']!='plan')
				_quit_error('The project is not in planning stage.');

			$start_ch=$this->wcal->date2ch();
			$rld=simple_query("SELECT $proj_deadline_fn FROM $proj_tbl WHERE $pq_fn=$pqid AND state!='plan' AND state!='done' AND (options&$proj_rts_opt)=0 ORDER BY  $proj_id_fn DESC, state ASC LIMIT 1");
			if($rld){
				$rld_norm=date_sql2norm($rld);
				$rld_ch=$this->wcal->date2ch($rld_norm);
				if($rld_ch>$start_ch)
					$start_ch=$rld_ch;
			}

			$this->ts_start_date=$this->wcal->ch2date("m/j/Y",$start_ch);
			$sql_start_date=date_norm2sql($this->ts_start_date);
			$this->ts_resource_conflict($sql_start_date);
		}
		// The purpose of non-overlap mode is to fill out ts_share array
		// with values from DB only for current pqid,
		// proj_id defined at this point is used in actual project time estimation

		// Construct array of products a user is likely to work on
		// This is based on which products a manager is likely to work on
		// which is figured out by ts_resource_conflict above
		$mgr2pqids=&$this->ts_mgr_in_pqs;

		$this->init_fixers(TRUE);
		foreach($this->all_fixers as $u){
			$m=YxlUsers::report_to($u);
			//if(!isset($mgr2pqids[$m]) && !isset($mgr2pqids[$u])) continue;
			arr_push($this->ts_users,$u);
			if(isset($mgr2pqids[$m]))
				$this->ts_user_in_pqids[$u]=$mgr2pqids[$m];
			if(isset($mgr2pqids[$u]))
				$this->ts_user_in_pqids[$u]=$mgr2pqids[$u];
		}

		if($ph2){
			if(count($this->all_testers)==0)
				$this->init_testers();
			foreach($this->all_testers as $u){
				$m=YxlUsers::report_to($u);
				//if(!isset($mgr2pqids[$m]) && !isset($mgr2pqids[$u])) continue;
				arr_push($this->ts_users,$u);
				if(isset($mgr2pqids[$m]))
					$this->ts_user_in_pqids[$u]=$mgr2pqids[$m];
				if(isset($mgr2pqids[$u]))
					$this->ts_user_in_pqids[$u]=$mgr2pqids[$u];
			}
		}


		// Add current pid to $pqids and $usr2pids
		//$impl=$this->project_info['state']=='plan'?FALSE:TRUE;
		if($overlap_mode)
			$this->ts_pqids_in_work[$pqid]=FALSE;
		else
		// This will triger correct initialisation of ts_ushare later on
			$this->ts_pqids_in_work[$pqid]=TRUE;

		foreach($this->ts_users as $u)
			$this->ts_user_in_pqids[$u][$pqid]=$pqid;

		// Check if we missed any non-conflicting pids with user time shares set.
		if($overlap_mode && count($this->ts_users)){
			$sql_u='(uid='.implode(' OR uid=',$this->ts_users).')';
			$query="SELECT rid FROM $pnets_tbl WHERE type=".$pnet_type_ts." AND rid!=0 AND $sql_u GROUP BY rid";
			//echo "$query <br>\n";
			if(!($result=@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__.' Could not select from $pnets_tbl table.');

			$pqids=&$this->ts_pqids_in_work;
			while($row=mysql_fetch_array($result, MYSQL_NUM)){
				if(!isset($this->pqs_2d['name'][$row[0]]))
					continue;

				if(!isset($pqids[$row[0]]))
					$pqids[$row[0]]=FALSE;
			}
		}

		if($overlap_mode){
			$pqid_ct_pct=simple_query("SELECT ct_pct FROM $proj_tbl WHERE $pq_fn=".$this->project_info[$pq_fn]." ORDER BY state ASC, $proj_id_fn DESC LIMIT 1");
			//echo "$pqid_ct_pct <br>\n";
			if(!$pqid_ct_pct)
				$pqid_ct_pct=0;
		}else
			$pqid_ct_pct=$this->project_info['ct_pct']?$this->project_info['ct_pct']:0;

		$query="SELECT u.uid, IF(p.xval IS NULL,$pqid_ct_pct,p.xval) FROM users AS u LEFT JOIN $pnets_tbl AS p ON p.rid=0 AND u.uid=p.uid AND p.type=".$pnet_type_ts;
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $pnets_tbl table.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "uct: $row[0]=$row[1] <br>\n";
			$this->ts_ucts[$row[0]]=$row[1];
			foreach($this->ts_pqids_in_work as $pid => $impl)
				if($impl)
					$this->ts_ushare[$pid][$row[0]]=100-$row[1];
		}

		$query="SELECT pn.rid, u.uid, pn.xval, IF(pn.xval IS NULL,0,1) FROM users AS u LEFT JOIN $pnets_tbl AS pn ON u.uid=pn.uid AND pn.type=".$pnet_type_ts." WHERE pn.rid IN (".implode(',',array_keys($this->ts_pqids_in_work)).")";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__.' Could not select from users table.');

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "pid=$row[0], uid=$row[1], pct=".$row[2].", null=".$row[3]."<br>\n";
			if($row[3])
				$this->ts_ushare[$row[0]][$row[1]]=$row[2];
		}

		foreach($this->all_fixers as $u)
			if(YxlUsers::user_is($u)&USER_IS_LOCKED)
				$this->ts_ushare[$pqid][$u]=0;

		if($ph2){
			foreach($this->all_testers as $u)
				if(YxlUsers::user_is($u)&USER_IS_LOCKED)
					$this->ts_ushare[$pqid][$u]=0;
		}
	}

	function _ts_resource_conflict(){

		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];

		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$ph1_mgr_fn=$this->object_params['pm_phases'][0]['mgr_fn'];
		$ph2_mgr_fn=$this->object_params['pm_phases'][0]['mgr_fn'];

		$two_ph=FALSE;
		$est_proj_end_fn="estimated_fix_end";
		if(count($this->object_params['pm_phases'])>1){
			$two_ph=TRUE;
			$ph2_mgr_fn=$this->object_params['pm_phases'][1]['mgr_fn'];
			$est_proj_end_fn="estimated_test_end";
		}

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];

		$pqid=$this->project_info[$pq_fn];

		$query = "SELECT IF(parent,parent,$pq_fn), $ph1_mgr_fn".($two_ph?", $ph2_mgr_fn":'')." FROM $pq_tbl";
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $pq_tbl table.");
		$rnd_mrgs=array();
		$qa_mgrs=array();
		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			$this->ts_mgr_in_pqs[$row[1]][$row[0]]=$row[0];
			if($two_ph)
				$this->ts_mgr_in_pqs[$row[2]][$row[0]]=$row[0];
		}

		$plans=array();
		$impls=array();
		$query = "SELECT p.$pq_fn, IF(r.state='implement' OR r.state='test',1,0) FROM $pq_tbl AS p LEFT JOIN $proj_tbl AS r ON p.$pq_fn=r.$pq_fn WHERE !p.parent";
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $pq_tbl.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			if(!isset($this->ts_pqids_in_work[$row[0]])){
				$this->ts_pqids_in_work[$row[0]]=$row[1];
				continue;
			}
			if($row[1])
				$this->ts_pqids_in_work[$row[0]]=$row[1];
		}
	}

	function ts_resource_conflict($sql_start_date=FALSE){

		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];

		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$ph1_mgr_fn=$this->object_params['pm_phases'][0]['mgr_fn'];
		$ph2_mgr_fn=$this->object_params['pm_phases'][0]['mgr_fn'];

		$two_ph=FALSE;
		$est_proj_end_fn="estimated_fix_end";
		if(count($this->object_params['pm_phases'])>1){
			$two_ph=TRUE;
			$ph2_mgr_fn=$this->object_params['pm_phases'][1]['mgr_fn'];
			$est_proj_end_fn="estimated_test_end";
		}

		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];

		$pqid=$this->project_info[$pq_fn];

		$query = "SELECT r.rid, p.$ph1_mgr_fn, p.$ph2_mgr_fn FROM $req_tbl AS r, $pq_tbl AS p WHERE r.$pq_fn=p.$pq_fn AND r.$req_in_proj_fn!='yes' AND (p.parent=$pqid OR r.$pq_fn=$pqid)";
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
		$rnd_mrgs=array();
		$qa_mgrs=array();
		while($row=mysql_fetch_array($result, MYSQL_NUM)){
		// echo "resource_conflict: rnd=$row[1] qa=$row[2] <br>";
			if($row[1]) $rnd_mrgs[$row[1]]=TRUE;
			if($two_ph && $row[2]) $qa_mgrs[$row[2]]=TRUE;
		}

		$plans=array();
		$impls=array();
		$query = "SELECT $proj_id_fn, state FROM $proj_tbl WHERE (state='plan' OR ((state='implement' OR state='test') AND $est_proj_end_fn>".($sql_start_date?"'$sql_start_date'":'NOW()').")) AND $pq_fn!=$pqid";
		//echo "$query <br>";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $proj-tbl.");

		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			//echo "$row[0] $row[1] <br>\n";
			if($row[1]=='plan')
				array_push($plans,$row[0]);
			else
				array_push($impls,$row[0]);
		}

		if(count($plans)){
			$sql_s1="(r.$req_proj_id_fn=".implode(" OR r.$req_proj_id_fn=",$plans).')';
			$query = "SELECT p.$ph1_mgr_fn, p.$ph2_mgr_fn, IF(p.parent,p.parent,p.$pq_fn), r.rid FROM $req_tbl AS r, $pq_tbl AS p WHERE r.$pq_fn=p.$pq_fn AND $sql_s1 AND r.$req_in_proj_fn='plan'";
			//echo "$query <br>";
			if(!($result=@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
			$overlap=FALSE;
			while($row=mysql_fetch_array($result, MYSQL_NUM)){
				if(isset($rnd_mrgs[$row[0]])){
					//echo "resource_conflict, pid=$row[2]: rnd=$row[0] <br>";
					//if(!isset($this->ts_mgr_in_pqs[$row[0]]))
					//	$this->ts_mgr_in_pqs[$row[0]]=array();
					$this->ts_mgr_in_pqs[$row[0]][$row[2]]=$row[2];
				}
				if(isset($qa_mrgs[$row[1]])){
					//echo "resource_conflict, pid=$row[2]: qa=$row[1] <br>";
					//if(!isset($this->ts_mgr_in_pqs[$row[1]]))
					//	$this->ts_mgr_in_pqs[$row[1]]=array();
					$this->ts_mgr_in_pqs[$row[1]][$row[2]]=$row[2];
				}
				$this->ts_pqids_in_work[$row[2]]=FALSE;
			}
		}

		if(count($impls)){
			$sql_s1="(r.$req_proj_id_fn=".implode(" OR r.$req_proj_id_fn=",$impls).')';
			$query = "SELECT p.$ph1_mgr_fn, p.$ph2_mgr_fn, IF(p.parent,p.parent,p.$pq_fn) FROM $req_tbl AS r, $pq_tbl AS p, users AS u WHERE r.$pq_fn=p.$pq_fn AND $sql_s1 AND $req_in_proj_fn='yes'";
			if(!($result=@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
			$overlap=FALSE;
			while($row=mysql_fetch_array($result, MYSQL_NUM)){
				if(isset($rnd_mrgs[$row[0]])){
					//echo "resource_conflict, pid=$row[2]: rnd=$row[0] <br>";
					//if(!isset($this->ts_mgr_in_pqs[$row[0]]))
					//	$this->ts_mgr_in_pqs[$row[0]]=array();
					$this->ts_mgr_in_pqs[$row[0]][$row[2]]=$row[2];
					$this->ts_pqids_in_work[$row[2]]=TRUE;
				}
				if(isset($qa_mrgs[$row[1]])){
					//echo "resource_conflict, pid=$row[2]: qa=$row[1] <br>";
					//if(!isset($this->ts_mgr_in_pqs[$row[1]]))
					//	$this->ts_mgr_in_pqs[$row[1]]=array();
					$this->ts_mgr_in_pqs[$row[1]][$row[2]]=$row[2];
					$this->ts_pqids_in_work[$row[2]]=TRUE;
				}
			}
		}
	}

	function ts_users_num(){
		return count($this->ts_users);
	}

	function ts_conflict_exists(){
		if(count($this->ts_mgr_in_pqs)>1 || count($this->ts_pqids_in_work)>1)
			return TRUE;
		return FALSE;
	}

	function ts_user_critical_share($uid){
		return isset($this->ts_ucts[$uid])?$this->ts_ucts[$uid]:0;
	}

	function ts_user_has_share($uid, $pqid=FALSE){
		$pq_fn=$this->object_params['pq_fn'];
		if(!$pqid)
			$pqid=$this->project_info[$pq_fn];
		return isset($this->ts_ushare[$pqid][$uid]);
	}

	function ts_user_share($uid, $pqid=FALSE){
		$pq_fn=$this->object_params['pq_fn'];
		if(!$pqid)
			$pqid=$this->project_info[$pq_fn];

		//print_r($this->ts_ushare);
		if($this->ts_user_has_share($uid, $pqid))
			return $this->ts_ushare[$pqid][$uid];
		else
			return 100-$this->ts_user_critical_share($uid);
	}

	function ts_user_scaler($uid){
		$ts=$this->ts_user_share($uid);
		if($ts>0)
			$scaler=100/$ts;
		else
			$scaler=999;
		return $scaler;
	}

	function ts_user_in_pqids($uid){
		return isset($this->ts_user_in_pqids[$uid])?$this->ts_user_in_pqids[$uid]:array();
	}

	function ts_update_ushares(&$pct,&$avail){
		$pnets_tbl=$this->object_params['pnets_tbl'];
		$pnet_type_ts=$this->object_params['pnet_type_ts'];

		if(!count($avail) || !count($pct))
                        _die('No data to process.');

                $sql_del=array();
                $sql_insert=array();
                foreach($pct as $u => $upcts){
                        //echo "User: $u <br>\n";
                        $x=0;
                        array_push($sql_del,"(uid='$u' AND (rid=".implode(' OR rid=',array_keys($upcts))."))");
                        foreach($upcts as $p => $n){
                                //echo "prod=$p, pct=$n <br>\n";
                                $x+=$n;
                                array_push($sql_insert,"($p, $u, $pnet_type_ts, $n)");
                        }
                        if($x!=$avail[$u])
                                _quit_error('Total pct='.$x.' != avail='.$avail[$u]);
                }
                if(count($sql_insert)){
                        if(!@mysql_query ("LOCK TABLES $pnets_tbl WRITE"))
                                _die2(__FILE__.':'.__LINE__." Could not lock $pnets_tbl table. ");
                        if(count($sql_del)){
                                $query="DELETE FROM $pnets_tbl WHERE type=$pnet_type_ts AND (".implode(' OR ',$sql_del).")";
                                //echo "$query <br>\n";
                                if(!@mysql_query($query))
                                        _die2(__FILE__.':'.__LINE__." Could not delete from $pnets_tbl table.");
                        }
                        $query="INSERT INTO $pnets_tbl (rid, uid, type, xval) VALUES ".implode(', ',$sql_insert);
                        //echo "$query <br>\n";
                        if(!@mysql_query($query))
                                _die2(__FILE__.':'.__LINE__." Could not insert into $pnets_tbl table.");
                        if(!@mysql_query ("UNLOCK TABLES"))
                                _die2(__FILE__.':'.__LINE__." Could not unlock $pnets_tbl table. ");

                }
	}

	function current_phase(){
		$pq_fn=$this->object_params['pq_fn'];
		if(!isset($this->project_info[$pq_fn]))
			_die('Project is not initialized.');

		$state=$this->project_info['state'];
		for($i=0;$i<count($this->object_params['pm_phases']);$i++)
			if($this->object_params['pm_phases'][$i]['proj_state']==$state)
				return $i+1;
		return FALSE;
	}

	function get_project_info($tag,$ph=FALSE){
		$pq_fn=$this->object_params['pq_fn'];
		if(!isset($this->project_info[$pq_fn]))
			_die('Project is not initialized.');

		$proj_id_fn=$this->object_params['proj_id_fn'];

                $pqid=$this->project_info[$pq_fn];
		$state=$this->project_info['state'];

		if((substr($tag,0,6)=='phase_')){
			if(!$ph)
				$ph=$this->current_phase();
                        if(!$ph)
                        	_die("Can not use tag $tag in this state (not an implementation phase).");
		}

		switch($tag){
			case 'pqid':
				return $pqid;
			case 'state':
				return $state;
			case 'label':
				return $this->project_info['label'];
			case 'plan_start_date':
				return $this->project_info[$tag];
			case 'implement_start_date':
				return $this->project_info[$tag];
			case 'deadline':
                		$proj_deadline_fn=$this->object_params['proj_deadline_fn'];
				return $this->project_info[$proj_deadline_fn];
			case 'phase_fix_rate':
				return $ph==1?$this->project_info['good_fix_rate']:$this->project_info['good_test_rate'];
			case 'phase_planned_end':
                                return $ph==1?$this->project_info['estimated_fix_end']:$this->project_info['estimated_test_end'];
			case 'phase_start':
				return $ph==1?$this->project_info['implement_start_date']:$this->project_info['test_start_date'];
			case 'phase_end':
				if($this->get_phase_num()>1)
					return $ph==1?$this->project_info['test_start_date']:$this->project_info['done_date'];
				return $this->project_info['done_date'];
			default:
				_die("Unknown tag $tag.");
		}
	}


	function plan_req_info($rn,$tag,$ph=FALSE){

		if(!isset($this->ordered_plan[0][0]['rid']))
			_die(__FILE__.':'.__LINE__.' Project plan is not initialized.');

		if(!isset($this->ordered_plan[$rn][0]))
                        _die(__FILE__.':'.__LINE__.' Request is not in the project plan.');


		if((substr($tag,0,6)=='phase_')){
			if(!$ph)
				$ph=$this->current_phase();
                        if(!$ph)
                        	_die("Can not use tag $tag in this state (not an implementation phase).");

			$ph_ssm=$this->object_params['pm_phases'][$ph-1]['start_state_mask'];
			$ph_dsm=$this->object_params['pm_phases'][$ph-1]['done_state_mask'];
		}


		switch($tag){
			case 'is_conflict':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
				return $this->ordered_plan[$rn][2]['is_conflict'];
			case 'phase_start_sm':
				return $ph_ssm;

			case 'phase_done_sm':
				return $ph_dsm;

			case 'start_shift_reason':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
				return $this->ordered_plan[$rn][2]['wh_shift_reason'];

			case 'exceed_by_wrk_hrs':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
				return $this->ordered_plan[$rn][2]['exceed_by_wrk_hrs'];

			case 'phase_time':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
				return $this->ordered_plan[$rn][2][$ph==1?'fix_time':'test_time'];

			case 'phase_without_estimate':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
				return $this->ordered_plan[$rn][2][$ph==1?'no_fix_estimate':'no_test_estimate'];

			case 'phase_planned_start':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
				return $this->ordered_plan[$rn][0]['ph'.$ph.'_plan_start'];

			case 'phase_planned_end':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
				return $this->ordered_plan[$rn][0]['ph'.$ph.'_plan_done'];

			case 'phase_pre_adjust_start':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');

                        	$key = 'plan_ph'.$ph.'_start_ch';
				$ch=isset($this->ordered_plan[$rn][2][$key])?$this->ordered_plan[$rn][2][$key]:FALSE;
				//echo "$key -->$ch <br><br>\n";
				return $ch?$this->wcal->ch2tp($ch):FALSE;

			case 'phase_start':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');

                        	$key = $ph==1?'fix_start_ch':'test_start_ch';
				$ch=isset($this->ordered_plan[$rn][2][$key])?$this->ordered_plan[$rn][2][$key]:FALSE;
				//echo "$key -->$ch <br><br>\n";
				return $ch?$this->wcal->ch2tp($ch,FALSE):FALSE;

			case 'phase_end':
				if(!isset($this->ordered_plan[$rn][2]))
                        		_die('The project plan is not computed.');
                        	$key = $ph==1?'fix_end_ch':'test_end_ch';
				$ch=isset($this->ordered_plan[$rn][2][$key])?$this->ordered_plan[$rn][2][$key]:FALSE;
				return $ch?$this->wcal->ch2tp($ch,FALSE):FALSE;

			case 'phase_complete_pct':
				$RInfo=&$this->ordered_plan[$rn][0];
				$ism=$this->rts->inwork_state_mask[$RInfo['class']];
				$fsm=$this->rts->fixed_state_mask[$RInfo['class']];
				$vsm=$this->rts->verified_state_mask[$RInfo['class']];
				if($ph==1){
					if($RInfo['state']&$ism)
						return $this->ordered_plan[$rn][0]['complete_pct'];
					elseif($RInfo['state']&$fsm)
						return 100;
					else
						return 0;
				}else{
					if($RInfo['state']&$vsm)
						return 100;
					else
						return 0;
				}

			case 'phase_fixer':
				return $this->ordered_plan[$rn][2][$ph==1?'fixer':'tester'];

			default:
				_die("Unknown tag $tag.");
		}

	}

	var $rdetection_mode=0;
	function &phase_pq_resources($ph,$pqid){
		$a=array(); $ref=&$a;
		if($ph==1){
			if($this->rdetection_mode==1)
				$ref=&$this->qid2fixer_2d[$pqid];
			else
				$ref=&$this->all_fixers;
		}elseif($ph==2)
			$ref=&$this->all_testers;

		return $ref;
	}

	function &plan_pq_fixers($rn,$ph){
		if(!isset($this->ordered_plan[0][0]['rid']))
			_die(__FILE__.':'.__LINE__.' Project plan is not initialized.');

		if(!isset($this->ordered_plan[$rn][0]))
                        _die(__FILE__.':'.__LINE__.' Request is not in the project plan.');

		if(!$ph)
			$ph=$this->current_phase();
		if(!$ph)
                        _die("Can not use plan_pq_fixers in this state (not an implementation phase).");
		$pq_fn=$this->object_params['pq_fn'];
		$row=&$this->ordered_plan[$rn][0];

		return $this->phase_pq_resources($ph,$row[$pq_fn]);

	}

	function &project_all_fixers($ph){
		$pq_fn=$this->object_params['pq_fn'];
		if(!isset($this->project_info[$pq_fn]))
			_die('Project is not initialized.');

		if(!$ph)
			$ph=$this->current_phase();
		if(!$ph)
                        _die("Can not use project_all_fixers in this state (not an implementation phase).");

		$a=array();
		$ref=&$a;
		if($ph==1)
			$ref=&$this->all_fixers;
		elseif($ph==2)
			$ref=&$this->all_testers;

		return $ref;
	}

	function project_fixer_hrs($u,$ph,$tag){
		$pq_fn=$this->object_params['pq_fn'];
		if(!isset($this->project_info[$pq_fn]))
			_die('Project is not initialized.');

		if(!$ph)
			$ph=$this->current_phase();
		if(!$ph)
                        _die("Can not use project_fixer_hrs in this state (not an implementation phase).");

		$max_fhp=$this->get_max_fhp();
		$total_work_hrs=$this->wcal->ch2whp($u,$max_fhp);
		switch($tag){
			case 'total':
				if($ph==1)
					return $total_work_hrs;
				elseif($ph==2)
					return $this->wcal->total_whs($u)-$total_work_hrs;
			case 'used':
				if($ph==1)
					return $this->fixer_wh($u)-$this->fixer_wasted_hrs($u);
				elseif($ph==2)
					return $this->tester_wh($u)?$this->tester_wh($u)-$total_work_hrs:0;
			default:
				_die("Unknown tag $tag.");
		}
	}

	function phase_done_requests_num(){
		$current_phase=$this->current_phase();
		$n=0;
		for($i=0;$i<count($this->ordered_plan);$i++){
                	$row=&$this->ordered_plan[$i][0];
                	$ph_done_sm=$this->plan_req_info($i,'phase_done_sm',$current_phase);
			if($row['state']&$ph_done_sm[$row['class']])
				$n++;
		}
		return $n;
	}

	function in_last_phase(){
		if(!($ph=$this->current_phase()))
                	_die(__FILE__.':'.__LINE__."Can not use the function in this project state (not an implementation phase).");
		return $ph==count($this->object_params['pm_phases']);
	}

	function start_next_phase(){

		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

                $req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
                $proj_deadline_fn=$this->object_params['proj_deadline_fn'];

		$proj_id=$this->project_info[$proj_id_fn];
		$current_phase=$this->current_phase();

		$isd=$this->project_info['implement_start_date'];
		$ph_fix_wrk_hrs_fn=($current_phase==1)?'fix_wrk_hrs':'test_wrk_hrs';

		$ph_fixer_fn=$this->object_params['pm_phases'][$current_phase-1]['implementer_fn'];
		$ph_inwork_fn=$this->object_params['pm_phases'][$current_phase-1]['start_fn'];
		if(!$ph_inwork_fn && isset($this->object_params['pm_phases'][$current_phase-2]))
			$ph_inwork_fn=$this->object_params['pm_phases'][$current_phase-2]['done_fn'];
		$ph_fixed_fn=$this->object_params['pm_phases'][$current_phase-1]['done_fn'];


		$query="SELECT AVG($ph_fix_wrk_hrs_fn), NOW() FROM $req_tbl WHERE $req_in_proj_fn='yes' AND $req_proj_id_fn=$proj_id";
		//echo $query." <br>\n";
		$avg_ft=simple_query($query);
                if(!$avg_ft)
                	_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

                $tsd=$avg_ft[1];
                $avg_ft=$avg_ft[0];

		if($this->in_last_phase()){
			$next_state='done';
			$state_date_fn="done_date";
		}else{
			$next_state=$this->object_params['pm_phases'][$current_phase]['proj_state'];
			$state_date_fn=$next_state."_start_date";
		}

		if($ph_inwork_fn){
	                if($current_phase==1){
				$this->init_fixers(TRUE);
				$fixers=&$this->all_fixers;
			}else{
				$this->init_testers();
				$fixers=&$this->all_testers;
			}

	                $isd_norm=date_sql2norm($isd); $isd_ch=$this->wcal->date2ch($isd_norm);
	                $tsd_norm=date_sql2norm($tsd); $tsd_ch=$this->wcal->date2ch($tsd_norm);
			$this->wcal->init($isd_ch,$tsd_ch,$fixers);

	                $sql_str='(u.uid IN ('.implode(',',$fixers).'))';
	                $query="SELECT SUM(r.$ph_fix_wrk_hrs_fn), u.uid FROM users AS u, $req_tbl AS r WHERE u.uid=r.$ph_fixer_fn AND $sql_str AND !(r.$req_in_proj_fn='yes') AND !(r.$req_proj_id_fn=$proj_id) AND !(r.$ph_inwork_fn IS NULL) AND r.$ph_inwork_fn>'$isd' AND !(r.$ph_fixed_fn IS NULL) AND r.$ph_fixed_fn<'$tsd' AND r.customer_pri=".REQ_PRI_CRITICAL." AND !(r.$ph_fix_wrk_hrs_fn IS NULL) GROUP BY u.uid";
	                //echo "$query <br>\n";
	                if(!($result=@mysql_query($query)))
	                	_die2(__FILE__.':'.__LINE__.' Could not select from requests table.');
	                $avg_ct_pct=0;
	                $n=0;
	                while($row=mysql_fetch_array($result, MYSQL_NUM)){
	                	//echo "fixer=$row[1], crit_wrk_hrs=$row[0] <br>\n";
	                        if($this->wcal->total_whs($row[1])){
	                        	$avg_ct_pct+=$row[0]/$this->wcal->total_whs($row[1]);
	                        	$n++;
	                        }
	                }
	                if($n)
	                	$avg_ct_pct=round(100*$avg_ct_pct/$n);
	                //echo "phase: $current_phase, avg_ft: $avg_ft, avg_ct_pct: $avg_ct_pct <br>\n";

		}

		$query = "UPDATE $proj_tbl SET state='$next_state', $state_date_fn=NOW()";

		if(($current_phase==1) && $ph_inwork_fn)
			$query .= ", avg_fix_hrs=$avg_ft";
		elseif(($current_phase==2) && $ph_inwork_fn)
			$query .= ", avg_test_hrs=$avg_ft";

		if((($current_phase==1) || !$this->project_info['ct_pct']) && $ph_inwork_fn)
			$query .= ", ct_pct=$avg_ct_pct";

		$query .= " WHERE $proj_id_fn=$proj_id";

                //echo "$query <br>\n";
                if(!@mysql_query($query))
                	_die2(__FILE__.':'.__LINE__." Could not update $proj_tbl.");
	}

	function  decommit_requests(&$dcrs){
		global $uid;

		if(!count($dcrs))
			return;

		$req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_dc=$this->object_params['history_proj_decommit_type'];

		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$dcrs_str=implode(',',$dcrs);

		$query="SELECT rid, state, fixer_uid, tester_uid FROM $req_tbl WHERE rid IN ($dcrs_str)";
		//echo "$query<br>\n"; die();
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		$hist_insert=array();
		while($row=mysql_fetch_array($result)){
			array_push($hist_insert,"('".$row[0]."', '$uid', '".$row[1]."', NOW(), '$hist_proj_dc', '$proj_id', '".$row[2]."', '".$row[3]."')");
		}

		$query = "UPDATE $req_tbl SET $req_in_proj_fn='no' WHERE rid IN ($dcrs_str)";
		//echo "$query<br>\n"; die();
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");

		$query = "DELETE FROM $hist_tbl WHERE xid=$proj_id AND options=$hist_proj_dc AND rid IN ($dcrs_str)";
		//echo "$query <br>";
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not delete from $hist_tbl table.");

		if(count($hist_insert)){
			$hist_insert_str=implode(', ',$hist_insert);
			$query = "INSERT INTO $hist_tbl (rid, uid, state, date, options, xid, xid2, xid3) VALUES $hist_insert_str";
			//echo "$query <br>";die();
			if(!@mysql_query($query))
				_die2(__FILE__.':'.__LINE__." Could not insert into $hist_tbl table.");
		}

		$this->history_buf.= "\nDe-committing requests:\n".
			"<rid>$dcrs_str</rid>";
	}

	function  recommit_requests(&$rcrs){
		if(!count($rcrs))
			return;

		$req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_dc=$this->object_params['history_proj_decommit_type'];

		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$rcrs_str=implode(',',$rcrs);
		$query ="DELETE FROM $hist_tbl WHERE rid IN ($rcrs_str) AND options=$hist_proj_dc AND xid=$proj_id";
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not delete from $hist_tbl table.");

		$query = "UPDATE $req_tbl SET $req_proj_id_fn=$proj_id, $req_in_proj_fn='yes' WHERE rid IN ($rcrs_str)";
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");

		$this->history_buf.= "\nRe-committing requests:\n".
			"<rid>$rcrs_str</rid>";
	}

	function start_first_phase(&$rids,&$in_proj,&$phfixer,&$ph_sch,&$ph_ech,&$ph_ech_max){
		global $uid;
		global $https_path;

		if(!count($rids))
			return;

		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

                $req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];
                $req_proj_pri_fn=$this->object_params['req_proj_pri_fn'];

                $hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_plan=$this->object_params['history_proj_plan_type'];

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];

		$proj_id=$this->project_info[$proj_id_fn];
		$pqid=$this->project_info[$pq_fn];
		$lbl=$this->project_info['label'];

		$sql_str='';
		for($ph=1;$ph<=$this->get_phase_num();$ph++){
			$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
			$sql_str.=", $ph_fixer_fn AS ph{$ph}_fixer";
		}

		$query = "SELECT rid, class, state, phase_of_rid, ".substr($sql_str,1)." FROM $req_tbl WHERE $req_proj_id_fn=$proj_id AND $req_in_proj_fn='plan'";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		$r2of=array();$rcs=array();$r2p=array();
		while($row=mysql_fetch_array($result, MYSQL_ASSOC)){
			for($ph=1;$ph<=$this->get_phase_num();$ph++)
				$r2of[$ph][$row['rid']]=$row['ph'.$ph.'_fixer'];
			$rcs[$row['rid']]=array(
				'class' => $row['class'],
				'state' => $row['state']);
			if($row['phase_of_rid'])
				$r2p[$row['rid']]=$row['phase_of_rid'];
		}

		$f2r_2d=array();
		$parents=array();
		$hist_insert=array();
		foreach($rids as $i => $r){
			if(!$in_proj[$i]) continue;
			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$class2dsm=$this->get_phase_info($ph,'done_state_mask');
				$dsm=$class2dsm[$rcs[$r]['class']];
				if(!isset($f2r_2d[$ph]))
					$f2r_2d[$ph]=array();
				if(!($dsm&$rcs[$r]['state']))
					$f2r_2d[$ph][$phfixer[$ph][$i]][$r]=$r;
			}

			$query = "UPDATE $req_tbl SET $req_proj_id_fn=$proj_id, $req_in_proj_fn='yes'";
			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
				$ph_plan_sfn=$this->get_phase_info($ph,'plan_start_fn');
				$ph_plan_efn=$this->get_phase_info($ph,'plan_done_fn');
				$ph_cal_sfn=str_replace('plan', 'cal', $ph_plan_sfn);
				$ph_cal_efn=str_replace('plan', 'cal', $ph_plan_efn);

				$query.=", $ph_fixer_fn=".$phfixer[$ph][$i];
				$query.=", plan_$ph_fixer_fn=".$phfixer[$ph][$i];
				$query.=", $ph_plan_sfn='".$ph_sch[$ph][$i]."'";
				$query.=", $ph_plan_efn='".$ph_ech[$ph][$i]."'";
				$query.=", $ph_cal_sfn='".$ph_sch[$ph][$i]."'";
				$query.=", $ph_cal_efn='".$ph_ech[$ph][$i]."'";
			}
			$query.=", options=options&~".REQ_OPT_RPSTMP.", $req_proj_pri_fn='".($i+1)."' WHERE rid=$r";
			//echo "$query <br>\n";
			if(!@mysql_query($query))
				_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");

			if(isset($r2p[$r]) && $r2p[$r])
				arr_push($parents,$r2p[$r]);

			array_push($hist_insert,"('".$r."', '$uid', NOW(), '$hist_proj_plan', '$proj_id')");
		}

		// Delete all TMP requests previousely removed from the plan,
		// 	except involved parents
		if(count($parents)){
			$query="UPDATE $req_tbl SET options=options&~".REQ_OPT_RPSTMP." WHERE rid IN (".implode(',',$parents).")";
			if(!@mysql_query($query))
				_die2(__FILE__.':'.__LINE__." Could not update $req_tbl  table.");
		}

		$query="DELETE FROM $req_tbl WHERE $req_proj_id_fn=$proj_id AND $req_in_proj_fn='no' AND options&".REQ_OPT_RPSTMP;
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not delete from requests table.");

		$fisrt_proj_state=$this->object_params['pm_phases'][0]['proj_state'];

		$query = "UPDATE $proj_tbl SET state='$fisrt_proj_state', implement_start_date=NOW()";
		for($ph=1;$ph<=$this->get_phase_num();$ph++){
			$ph_plan_efn=$this->get_phase_info($ph,'plan_done_fn');
			$ph_plan_max_efn=str_replace('plan', 'estimated', $ph_plan_efn);
			$query.=", $ph_plan_max_efn='".$ph_ech_max[$ph]."'";
		}
		$query.=" WHERE $proj_id_fn=$proj_id";
		if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not update $proj_tbl table.");

		if(count($hist_insert)){
			$hist_insert_str=implode(', ',$hist_insert);
			$query = "INSERT INTO $hist_tbl (rid, uid, date, options, xid) VALUES $hist_insert_str";
			//echo "$query <br>";
			if(!@mysql_query($query))
				_die2(__FILE__.':'.__LINE__." Could not insert into $hist_tbl table.");
		}

		$pqname=simple_query("SELECT name FROM $pq_tbl WHERE $pq_fn=$pqid");
		if(!$pqname)
			_die2(__FILE__.':'.__LINE__." Could not select from $pq_tbl table.");

		// Send some kind of notification email to all new fixers !!!!
		foreach($f2r_2d[1] as $u => $rs){
			if(count($rs)<1) continue;
			$rid_str=implode(', ',$rs);
			$cc_uids=array();
			foreach($rs as $r){
				$ou=$r2of[1][$r];
				if($ou==$u) continue;
				$cc_uids[$ou]=$ou;
			}
			//echo YxlUsers::name($u]." -> $rid_str <br>\n";
			$msg  = "  Hello {YxlUsers::name($u)},\n\n";
			$msg .= " The '$pqname $lbl' release members have agreed on a plan and initiated the implementation phase.\n\n";
			$msg .= " You are reponsible for fixing/implementing the following requests:\n\n";
			$msg .="\t$rid_str\n\n";
			$msg .=" The deadline for the implementation phase is ".date_sql2norm($ph_ech_max[1],TRUE).". To see the status of the implementation phase go to:\n\n";
			$msg.= "\t$https_path/rps_track.php?proj_id=".$proj_id."\n\n";
			$msg .=$this->proj_m_sign;
			//$rps->email($u, 'Start implementation for the next release.', format_text($msg), $cc_uids);
		}

		// Similar with testers !!!!
		foreach($f2r_2d[2] as $u => $rs){
			if(count($rs)<1) continue;
			$rid_str=implode(', ',$rs);
			$cc_uids=array();
			foreach($rs as $r){
				$ou=$r2of[2][$r];
				if($ou==$u) continue;
				$cc_uids[$ou]=$ou;
			}
			//echo YxlUsers::name($u]." -> $rid_str <br>\n";
			$msg  = "  Hello {YxlUsers::name($u)},\n\n";
			$msg .= " The '$pqname $lbl' release members have agreed on a plan and initiated the implementation phase.\n\n";
			$msg .= " You will be reponsible for testing the following requests when they have been fixed/implemented:\n\n";
			$msg .="\t$rid_str\n\n";
			$msg .=" The deadline for the implementation phase is ".date_sql2norm($ph_ech_max[2],TRUE).". To see the status of the implementation phase go to:\n\n";
			$msg.= "\t$https_path/rps_track.php?proj_id=".$proj_id."\n\n";
			$msg .=$this->proj_m_sign;
			//$rps->email($u, 'Testing info for the next release.', format_text($msg), $cc_uids);
		}

	}

	function get_plan_addable_requests($sql_where='',$start=0){

		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

                $req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$pqid=$this->project_info[$pq_fn];
		$proj_ids_str=$this->valid_proj_ids_str();

		$query="SELECT req.rid, req.syn, req.class FROM
        		(SELECT r.rid, CONCAT(r.rid, ': [', IF(p.parent,p.name,'General'), '] ', r.synopsis) AS syn, r.class FROM $req_tbl AS r INNER JOIN $pq_tbl AS p ON r.$pq_fn=p.$pq_fn WHERE (r.$pq_fn=$pqid OR p.parent=$pqid) AND (!(r.$req_in_proj_fn IN ('yes','plan')) OR !(r.$req_proj_id_fn IN ($proj_ids_str))) AND ".$this->sql_states_str." AND r.customer_pri!=".REQ_PRI_CRITICAL." $sql_where ORDER BY r.customer_pri DESC, r.$pq_fn LIMIT ".$start.",100) AS req
        	LEFT OUTER JOIN
        		(SELECT COUNT(rid) AS phn, phase_of_rid FROM $req_tbl GROUP BY phase_of_rid) AS ch ON req.rid=ch.phase_of_rid WHERE ch.phn IS NULL";

		//die("$query <br>\n");
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
		return $result;
	}

	function plan_update(&$rids,&$phfixer,&$upri){
		global $uid;

		$req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$rm_rids=array();

		foreach($rids as $i => $r){
			$rm=FALSE;
			$rm_id='rm_'.$r;
			if(isset($_POST[$rm_id]))
				$rm=TRUE;
			$auto=FALSE;
			$auto_id='auto_'.$r;
			if(isset($_POST[$auto_id]))
				$auto=TRUE;

			$sql_str='';
			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
				$sql_str.=", plan_".$ph_fixer_fn;
			}
			//echo $sql_str;

			$RInfo=simple_query("SELECT options $sql_str FROM $req_tbl WHERE rid=$r");
			if(!$RInfo)
				_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");
			//print_r($RInfo);

			$opt=$RInfo[0];
			//echo "ORIG: rid=$r fixer=$RInfo[1] tester=$RInfo[2]<br>";
			//echo "NEW: rid=$r fixer={$phfixer[1][$i]} tester={$phfixer[2][$i]} upri=$upri[$i] rm=$rm $auto_id=$auto<br>";
			$sql_update=array();

			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
				if(!$auto && ($RInfo[$ph]!=$phfixer[$ph][$i]))
					array_push($sql_update,"plan_$ph_fixer_fn=".$phfixer[$ph][$i]);
			}

			if($rm){
				// Remove only if all blocker rids are being removed too
				$ok2rm=TRUE;
				//echo "rid=$r, blockers: ";print_r($this->get_blocking_list($r));echo "<br>\n"; die();
				foreach($this->get_blocking_list($r) as $b)
					if(in_array($b,$rid) && !isset($_POST['rm_'.$b])){
						error_alert("Can not remove request $r because request $b which depends on it is still in the plan.");
						$ok2rm=FALSE;
						break;
					}
				if($ok2rm){
					array_push($sql_update,"$req_in_proj_fn='no'");
					array_push($sql_update,"$req_proj_id_fn=NULL");
					$opt=$RInfo[0]&~REQ_OPT_RPSAUTO;
					array_push($rm_rids,$r);
				}
			}else{
				if($auto)
					$opt=$RInfo[0]|REQ_OPT_RPSAUTO;
				else
					$opt=$RInfo[0]&~REQ_OPT_RPSAUTO;
			}
			if($opt!=$RInfo[0])
				array_push($sql_update,"options=$opt");

			if(count($sql_update)){
				$sql_update_str=implode(', ',$sql_update);
				$query = "UPDATE $req_tbl SET $sql_update_str WHERE rid=$r";
				//echo "$query <br>";
				if(!@mysql_query($query))
					_die2(__FILE__.':'.__LINE__." Could not UPDATE $req_tbl table.");
			}
			$this->update_member_vote($r,$upri[$i]);
		}

		if(count($rm_rids))
			$this->history_buf.= "\nRemoving requests from the plan\n".
				"<rid>".implode(',',$rm_rids)."</rid>";
	}

	function plan_add_requests($rs){
		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

		$req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

                $pqid=$this->project_info[$pq_fn];

                $proj_id_fn=$this->object_params['proj_id_fn'];
                $proj_id=$this->project_info[$proj_id_fn];

		$proj_ids_str=$this->valid_proj_ids_str();

		$sql_rids='';
		for($i=0;$i<count($rs);$i++){
			if(!is_numeric($rs[$i]) || ($rs[$i]<1))
				continue;
			$sql_rids.=','.$rs[$i];
		}

		if(!$sql_rids)
			return FALSE;

		$sql_rids=substr($sql_rids,1);

		$add_rids_=array();
		$query="SELECT rid FROM $req_tbl WHERE (rid IN ($sql_rids) OR phase_of_rid In ($sql_rids)) AND !($req_proj_id_fn IN ($proj_ids_str) AND $req_in_proj_fn IN ('yes','plan'))";
		//echo "$query<br>\n";
		if (!($SQL = mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		while($Child = mysql_fetch_row($SQL)){
			$add_rids_[$Child[0]]=$Child[0];
			foreach($this->get_dependency_list($Child[0]) as $ChildDep)
				$add_rids_[$ChildDep]=$ChildDep;
		}

		//echo "add_rids_: ";print_r($add_rids_);echo "<br>\n";
		$add_rids=array();
		while($r=array_pop($add_rids_)){
			$row=simple_query("SELECT r.$pq_fn, p.parent, r.state, r.customer_pri, r.class FROM $req_tbl AS r INNER JOIN $pq_tbl AS p ON r.$pq_fn=p.$pq_fn WHERE r.rid='$r'");
			if(!$row)
				_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl.");

			if(($row[1]!=$pqid) && ($row[0]!=$pqid))
				_quit_error('Could not add this request to the current plan. Request '.$rid.' depends on a request '.$r.' from a different product/queue.');

			if($this->is_last_state($row[4],$row[2]))
				continue;

			if($row[3]==REQ_PRI_CRITICAL)
				continue;

			array_push($add_rids,$r);
		}

		//echo "add_rids: ";print_r($add_rids);echo "<br>\n";
		if(count($add_rids)){
			$sql_str='';
			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
				$sql_str.=", plan_".$ph_fixer_fn."=".$ph_fixer_fn;
			}
			$query = "UPDATE $req_tbl SET $req_in_proj_fn='plan', $req_proj_id_fn=$proj_id, options=options|".REQ_OPT_RPSAUTO." $sql_str WHERE rid IN (".implode(',',$add_rids).")";
			//echo "$query <br>";
			if(!(@mysql_query($query)))
				_die2(__FILE__.':'.__LINE__." Could not update $req_tbl table.");

			$this->history_buf.="\nAdding requests to the plan\n".
				"<rid>".implode(',',$add_rids)."</rid>";
		}
	}

	function track_select_decommits(){
		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

		$req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

                $hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_dc=$this->object_params['history_proj_decommit_type'];

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];

		$sql_str='';
		for($ph=1;$ph<=$this->get_phase_num();$ph++){
			$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
			$sql_str.=", $ph_fixer_fn AS ph".$ph."_fixer";
		}

		$query="SELECT r.rid, CONCAT('[',c.name,'] ',r.synopsis) AS syn, r.state, r.class, r.synopsis, r.phase_of_rid, rl.label $sql_str FROM $hist_tbl AS h INNER JOIN $req_tbl AS r ON h.rid=r.rid INNER JOIN $proj_tbl AS rl ON r.$req_proj_id_fn=rl.$proj_id_fn INNER JOIN $pq_tbl AS c ON r.$pq_fn=c.$pq_fn WHERE h.options=$hist_proj_dc AND (rl.state!='plan' AND rl.state!='done') AND r.$req_in_proj_fn='no' GROUP BY r.rid ORDER BY h.xid DESC";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $hist_tbl table.");

		return $result;
	}

	function track_update_gfr($gfr){
		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
                $proj_id=$this->project_info[$proj_id_fn];

		$query = "UPDATE $proj_tbl SET good_fix_rate=$gfr WHERE $proj_id_fn=$proj_id";
		if(!(@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not update $proj_tbl table.");

		$this->history_buf.= "\nAcceptable(good) fix rate: ".$this->project_info['good_fix_rate']."-> $gfr\n";
	}

	function &get_planned_reqnum_by_fixer(&$req_list){

		$hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_plan=$this->object_params['history_proj_plan_type'];

		$req_tbl=$this->object_params['requests_tbl'];
		$req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$sql_str='';
		for($ph=1;$ph<=$this->get_phase_num();$ph++){
			$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
			$sql_str.=", plan_$ph_fixer_fn AS ph".$ph."_fixer";
		}
		$query = "SELECT ".substr($sql_str,1).", r.rid FROM $hist_tbl AS h INNER JOIN $req_tbl AS r ON h.rid=r.rid WHERE h.options=$hist_proj_plan AND h.xid=$proj_id";
        	//echo $query;
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $hist_tbl table.");

		$ph_planned_prs_by_fixer=array();
		for($ph=1;$ph<=$this->get_phase_num();$ph++)
			$ph_planned_prs_by_fixer[$ph]=array();

		$ref=&$ph_planned_prs_by_fixer;
		while($row=mysql_fetch_array($result, MYSQL_ASSOC)){
			arr_push($req_list,$row['rid']);
			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$tag='ph'.$ph.'_fixer';
				if(isset($ph_planned_prs_by_fixer[$ph][$row[$tag]]))
                        		$ph_planned_prs_by_fixer[$ph][$row[$tag]]++;
                		else
                        		$ph_planned_prs_by_fixer[$ph][$row[$tag]]=1;
			}
		}
		return $ref;
	}

	function done_select_decommits($limit_list=FALSE){
		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

		$req_tbl=$this->object_params['requests_tbl'];
                $req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];
                $req_proj_pri_fn=$this->object_params['req_proj_pri_fn'];

                $hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_dc=$this->object_params['history_proj_decommit_type'];

                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$limit_sql_str='';
		if(is_array($limit_list)){
			if(count($limit_list))
				$limit_sql_str=" AND r.rid IN (".implode(',',$limit_list).")";
			else
				$limit_sql_str=" AND 0";
		}

		$sql_str='';
		for($ph=1;$ph<=$this->get_phase_num();$ph++){
			$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
			$sql_str.=", $ph_fixer_fn AS ph".$ph."_fixer";
		}

		$query = "SELECT r.rid, CONCAT('[',IF(c.parent,c.name,'General'),']',r.synopsis) AS syn, r.state, r.phase_of_rid, h.state, h.xid2, h.xid3, r.class $sql_str FROM $hist_tbl AS h INNER JOIN $req_tbl AS r ON h.rid=r.rid INNER JOIN $pq_tbl AS c ON r.$pq_fn=c.$pq_fn LEFT JOIN $pq_tbl AS p ON c.parent=p.$pq_fn WHERE h.options=$hist_proj_dc AND h.xid=$proj_id $limit_sql_str ORDER BY r.phase_of_rid, r.$req_proj_pri_fn DESC";
        	//echo $query;
        	if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $hist_tbl table.");

		return $result;
	}

	function &get_decommit_requests($SQL,$exclude_list=FALSE){
		if(!$exclude_list)
			$exclude_list=array();

		$decommits=array();
		$ref=&$decommits;

		while($row=mysql_fetch_array($SQL, MYSQL_ASSOC)){
			if(in_array($row['rid'],$exclude_list))
				continue;
			arr_push($decommits,$row['rid']);
		}

		return $ref;

	}

	function &plan_sorted_by_phase(&$exclude_list){

		$new_order=array();
		$ref=&$new_order;
        	$tmp_added=array();
        	// Add multiphase requests first
        	for($i=0;$i<count($this->ordered_plan);$i++){
                	$row=&$this->ordered_plan[$i][0];
			if(in_array($row['rid'],$exclude_list))
				continue;
                	$parent=$row['phase_of_rid'];
                	if(isset($tmp_added[$i]))
                        	continue;
                	if(!$parent)
                        	continue;

                	// Add all subrequests
                	for($j=0;$j<count($this->ordered_plan);$j++){
                        	$r=&$this->ordered_plan[$j][0];
                        	if($r['phase_of_rid']!=$parent)
                                	continue;
                        	array_push($new_order,$j);
                        	$tmp_added[$j]=TRUE;
                	}
        	}

        	// Add regular requests now
        	for($i=0;$i<count($this->ordered_plan);$i++){
                	$row=&$this->ordered_plan[$i][0];
			if(in_array($row['rid'],$exclude_list))
				continue;
                	$parent=$row['phase_of_rid'];
                	if($parent)
                        	continue;
                	array_push($new_order,$i);
        	}

		return $ref;
	}

	function get_request_title($rid){
		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

		$req_tbl=$this->object_params['requests_tbl'];

		$title_syn=simple_query("SELECT CONCAT('[',IF(p.parent,p.name,'General'),'] ',r.synopsis) FROM $req_tbl AS r INNER JOIN $pq_tbl AS p ON r.$pq_fn=p.$pq_fn WHERE r.rid=$rid");
		if(!$title_syn)
			$title_syn='Unknown request (please check database consistancy)';
		return $title_syn;
	}

	function done_select_in_proj($exclude_reqs){
		$pq_tbl=$this->object_params['pq_tbl'];
                $pq_fn=$this->object_params['pq_fn'];

		$req_tbl=$this->object_params['requests_tbl'];
		$req_in_proj_fn=$this->object_params['req_in_proj_fn'];
                $req_proj_id_fn=$this->object_params['req_proj_id_fn'];

		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$sql_str='';
		for($ph=1;$ph<=$this->get_phase_num();$ph++){
			$ph_fixer_fn=$this->get_phase_info($ph,'implementer_fn');
			$sql_str.=", $ph_fixer_fn AS ph".$ph."_fixer";
		}

		// Output PRs that have been closed in parallel to the formal plan
		// 03.11.08: bug578 reverting in_proj from 'maybe' back to 'yes'
		// 03.11.08: bug578 but also need to exclude scheduled requests
		$excl_sql_str=count($exclude_reqs)?' AND !(r.rid IN ('.implode(',',$exclude_reqs).'))':'';

        	$query = "SELECT r.rid, CONCAT('[',IF(c.parent,c.name,'General'),'] ',r.synopsis) AS syn, r.state, r.class, r.phase_of_rid $sql_str FROM $req_tbl AS r INNER JOIN $pq_tbl AS c ON r.$pq_fn=c.$pq_fn WHERE r.$req_proj_id_fn=$proj_id AND r.$req_in_proj_fn='yes' $excl_sql_str ORDER BY r.rid DESC";
        	//echo $query;
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $req_tbl table.");

		return $result;
	}

	function done_print_table_plan(&$req_list,&$Project,$tbl_title='',&$ph_done_prs_by_fixer,&$ph_work_stats_by_fixer){

		if(!$req_list)
			$req_list=array();

		$state2str=&$this->rts->state2str;
        	$wfid=$this->rts->wfid();

		$num_of_rids=count($req_list);
                $cspn=3+2*$this->get_phase_num();

                // Would be nice to add the same statistics per fixer !!!
                if($num_of_rids){
			echo '
<table width="100%" border=0 cellpadding=0 cellspacing=1 class="dgray">
<tr bgcolor="white">
        <th colspan='.$cspn.'>'.$tbl_title.'</th>
</tr>
<tr class=hgray>
        <th width="5%">ID</th>
        <th width="50%">[Component] Synopsis</th>
        <th>Sate</th>';

        		for($ph=1;$ph<=$this->get_phase_num();$ph++)
                        	echo '
        <th>'.$Project['implementer_title'][$ph].'</th>
        <th>'.$Project['phase_title'][$ph].' e/a hrs</th>';

        		echo '
</tr>
<tr class="separator">
        <td colspan='.$cspn.' height="1"></td>
</tr>';
		}


		$prev_task_rid=0;
		foreach($req_list as $i){
			$row=&$this->ordered_plan[$i][0];

			$etat=array();
			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$dsm=~0;
				if($ph>1)
					$dsm=$this->phN_done_state_mask($ph-1,$row['class']);

				if(!($dsm&$row['state']))
					continue;

				$tag='ph'.$ph.'_fixer';
				$htag='ph'.$ph.'_wrk_hrs';
				$et=$this->raw_ph_fixer_et($ph,$row['rid'],$row[$tag]);

				$at=$row[$htag];
				$etat[$ph]=($et===FALSE?'-':$et).'/'.($at===FALSE?'-':$at);

				if(isset($ph_done_prs_by_fixer[$ph][$row[$tag]])){
                        		$ph_done_prs_by_fixer[$ph][$row[$tag]]++;
					if($et!==FALSE){
						$ph_work_stats_by_fixer[$ph][$row[$tag]]['et_num']++;
						$ph_work_stats_by_fixer[$ph][$row[$tag]]['et']+=$et;
					}
					if($at!==FALSE){
						$ph_work_stats_by_fixer[$ph][$row[$tag]]['at_num']++;
						$ph_work_stats_by_fixer[$ph][$row[$tag]]['at']+=$at;
					}
                		}else{
                        		$ph_done_prs_by_fixer[$ph][$row[$tag]]=1;
					$ph_work_stats_by_fixer[$ph][$row[$tag]]=array(
						'et_num'=>($et===FALSE?0:1),
						'et'=>($et===FALSE?0:$et),
						'at_num'=>($at===FALSE?0:1),
						'at'=>($at===FALSE?0:$at));
				}
			}

			if($row['phase_of_rid'] && ($row['phase_of_rid']!=$prev_task_rid)){
			// This is a subrequest
				$title_syn=$this->get_request_title($row['phase_of_rid']);
				echo '
<tr class="lgray">
	<td></td>
	<th align="left" style="color: #66789C;" colspan="'.($cspn-1).'">'.$title_syn.'</th>
</tr>';
				$prev_task_rid=$row['phase_of_rid'];
			}elseif(!$row['phase_of_rid'])
				$prev_task_rid=0;

			$tdth='th';
                	if($this->rts->bug_class==$row['class'])
				$tdth='td';

                	echo '
<tr class="lgray" onmouseover="this.className=\'lgs\';" onmouseout="this.className=\'lgray\';" onClick="'.$wfid.'_view_win_open('.$row['rid'].')">
	<'.$tdth.' align="left">'.$row['rid'].'</'.$tdth.'>
	<'.$tdth.' align="left">'.($row['phase_of_rid']?'&#8658;&nbsp;':'').html_safe_text($row['synopsis']).'</'.$tdth.'>
	<td align="center">'.$state2str[$row['state']].'</td>';

		for($ph=1;$ph<=$this->get_phase_num();$ph++)
                			echo '
	<td align="center">'.YxlUsers::name($row['ph'.$ph.'_fixer']).'</td>
	<td align="center">'.$etat[$ph].'</td>';
		echo '
</tr>';
        	}
	echo '
</table><br>';
	}

	function done_print_table_sql($SQL,&$Project,$tbl_title=''){
        	$wfid=$this->rts->wfid();
		$num_of_rids=mysql_num_rows($SQL);
		$cspn=3+$this->get_phase_num();

		// Would be nice to add the same statistics per fixer !!!
		if($num_of_rids){
			echo '
<table width="100%" border=0 cellpadding=0 cellspacing=1 class="dgray">
<tr bgcolor="white">
	<th colspan='.$cspn.'>'.$tbl_title.'</th>
</tr>
<tr class=hgray>
	<th width="5%">ID</th>
	<th width="50%">[Component] Synopsis</th>
	<th>Sate</th>';

			for($ph=1;$ph<=$this->get_phase_num();$ph++)
				echo '
	<th>'.$Project['implementer_title'][$ph].'</th>';
			echo '
</tr>
<tr class="separator">
	<td height="1" colspan='.$cspn.'></td>
</tr>';
		}

		$prev_task_rid=0;
		while($row=mysql_fetch_array($SQL, MYSQL_ASSOC)){
			if($row['phase_of_rid'] && ($row['phase_of_rid']!=$prev_task_rid)){
		// This is a subrequest
				$title_syn=$this->get_request_title($row['phase_of_rid']);
			echo '
<tr class="lgray">
	<td></td>
	<th align="left" style="color: #66789C;" colspan="'.($cspn-1).'">'.$title_syn.'</th>
</tr>';
				$prev_task_rid=$row['phase_of_rid'];
			}elseif(!$row['phase_of_rid'])
				$prev_task_rid=0;

			$tdth='th';
                	if($this->rts->bug_class==$row['class'])
				$tdth='td';

               	 echo '
<tr class="lgray" onmouseover="this.className=\'lgs\';" onmouseout="this.className=\'lgray\';" onClick="'.$wfid.'_view_win_open('.$row['rid'].')">
	<'.$tdth.' align="left">'.$row['rid'].'</'.$tdth.'>
	<'.$tdth.' align="left">'.($row['phase_of_rid']?'&#8658;&nbsp;':'').html_safe_text($row['syn']).'</'.$tdth.'>
	<td align="center">'.$this->rts->state2str[$row['state']].'</td>';

			for($ph=1;$ph<=$this->get_phase_num();$ph++){
				$fixer_uid=$row['ph'.$ph.'_fixer'];
				$usr=YxlUsers::exists($fixer_uid)?YxlUsers::name($fixer_uid):'N/A';
                		echo '
	<td align="center">'.$usr.'</td>';
			}
			echo '
</tr>';
        	}

		if($num_of_rids)
			echo '
</table><br>';
	}

	function set_manual_ordering_mode(){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

                $query="UPDATE $proj_tbl SET options=options|".PROJ_OPT_MANUAL." WHERE $proj_id_fn=$proj_id";
                //echo "$query <br>\n";
                if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl table.");

		$this->project_info['options']|=PROJ_OPT_MANUAL;

		$this->history_buf.="\nSwitching to manual ordering mode\n";
	}

	function set_auto_ordering_mode(){
		$proj_tbl=$this->object_params['proj_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

                $query="UPDATE $proj_tbl SET options=options&~".PROJ_OPT_MANUAL." WHERE $proj_id_fn=$proj_id";
                //echo "$query <br>\n";
                if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl table.");

		$this->project_info['options']&=~PROJ_OPT_MANUAL;

		$this->history_buf.="\nSwitching to vote based, automatic ordering mode\n";
	}

	function in_manual_ordering_mode(){
		$proj_opt=$this->project_info['options'];
		return $proj_opt&PROJ_OPT_MANUAL;
	}

	function &fill_pnets($rids=FALSE,$user=FALSE){
		global $uid;

		if(!$user)
			$user=$uid;

		$pnets_tbl=$this->object_params['pnets_tbl'];
		$phnum=$this->get_phase_num();

		if(!$rids)
			$rids=array();

		$ph_pnets_types=array();
		foreach($this->object_params['pm_phases'] as $phase)
			array_push($ph_pnets_types,$phase['pnet_type']);

                $query = "SELECT rid, type, xval FROM $pnets_tbl WHERE uid=$user AND type IN (".implode(',',$ph_pnets_types).") AND rid IN (".implode(',',$rids).")";
		//echo "$query <br>\n";
                if(!($result=@mysql_query($query)))
                        _die(__FILE__.':'.__LINE__." Could not select from $pnets_tbl table.");

		$pnets=array();
		$ref=&$pnets;
                while($row=mysql_fetch_array($result, MYSQL_NUM)){
                        //echo "$row[0] $row[1] <br>";
			if(($ph=array_search($row[1],$ph_pnets_types))===FALSE)
				continue;

			$pnets[$ph][$row[0]]=$row[2];
                }

		return $ref;
	}

	function raw_ph_fixer_et($ph,$rid,$fixer){
		$fpnets=&$this->fill_pnets(array($rid),$fixer);
		return isset($fpnets[$ph-1][$rid])?$fpnets[$ph-1][$rid]:FALSE;
	}

	function save_pnets(&$pnets){
		global $uid;

		$pnets_tbl=$this->object_params['pnets_tbl'];

		for($ph=0;$ph<count($this->object_params['pm_phases']);$ph++){

			if(!isset($pnets[$ph]) || !count($pnets[$ph]))
				continue;

			$pnets_type=$this->object_params['pm_phases'][$ph]['pnet_type'];
			$query="DELETE FROM $pnets_tbl WHERE type=$pnets_type AND uid=$uid AND rid IN (".implode(',',array_keys($pnets[$ph])).")";
		//echo "$query <br>\n";
			 if(!@mysql_query($query))
				_die(__FILE__.':'.__LINE__." Could not delete from $pnets_tbl table.");

			$sql_ins_str='';
			foreach($pnets[$ph] as $rid => $hrs)
				$sql_ins_str.=", ($rid, $uid, $hrs, $pnets_type)";

			$query="INSERT INTO $pnets_tbl (rid, uid, xval, type) VALUES ".substr($sql_ins_str,1);
			//echo "$query <br>\n";
			if(!@mysql_query($query))
				_die2(__FILE__.':'.__LINE__." Could not insert into $pnets_tbl table.");
		}
	}

	function view_menu($proj_menu,&$Project,$show_del=FALSE,$show_next=FALSE){
          global $dir_name;
          global $is_customer;

		$cph=$this->current_phase();
	  	$proj_id_fn=$this->object_params['proj_id_fn'];
          	$proj_id=$this->project_info[$proj_id_fn];
	  	$proj_state=$stag=$this->project_info['state'];
	  	if($cph)
			$stag='track';

		$s1=' class="wbtn" style="border-style:inset;" disabled';
		$s2=' class="gbtn"';

          	$main_here	= ($proj_menu == PROJ_MENU_MAIN)    ?$s1:$s2;
          	$hist_here	= ($proj_menu == PROJ_MENU_HISTORY) ?$s1:$s2;
          	$est_here	= ($proj_menu == PROJ_MENU_EST)     ?$s1:$s2;
          	$ts_here	= ($proj_menu == PROJ_MENU_TS)      ?$s1:$s2;

		echo '
<table id="proj_menu" border=0 cellpadding=0 cellspacing=0 width="100%" align="center">
<tr>
        <td>
        <span class="proj_nav"><input type="button" name="" value="'.$proj_state.'"'.$main_here.' onClick="document.location=\''.$dir_name.$Project[$stag.'_script'].'?proj_id='.$proj_id.'\'">';

		if($proj_state=='plan')
			echo '
        <input type="button" name="" value="estimate"'.$est_here.' onClick="document.location=\''.$dir_name.$Project['estimate_script'].'?proj_id='.$proj_id.'\'">
        <input type="button" name="" value="time sharing"'.$ts_here.' onClick="document.location=\''.$dir_name.$Project['ts_script'].'?proj_id='.$proj_id.'\'">';

		if(isset($Project['history_script']))
			echo '
        <input type="button" name="" value="history"'.$hist_here.' onClick="document.location=\''.$dir_name.$Project['history_script'].'?proj_id='.$proj_id.'\'">';

		echo '</span>';
		if($this->rts->have_extra() && ($xtid=$this->rts->x_exists_proj_template($proj_id,$stag)))
			echo '
	<a href="'.$dir_name.'/report/x_reports.php?xtid='.$xtid.'">&nbsp;&gt;&gt;&nbsp;generate custom reports&nbsp;&lt;&lt;</a>';

		echo '
	</td>
	<td align="right">';
		if($show_del || $show_next)
			echo '
	<span class="proj_nav">';

		if($show_del)
			echo '<input type="button" class="gbtn" style="background:red" name="" value="delete" onClick="go(\'cancel\')">';

		if($show_next){
			if($cph){
				if($this->in_last_phase())
					$next_str='done';
				else
                        		$next_str=$this->object_params['pm_phases'][$cph]['proj_state'];
			}else
				$next_str=$this->object_params['pm_phases'][0]['proj_state'];
                	echo '<input type="button" class="gbtn" name="" value="'.$next_str.' >>" onClick="go(\'next\')">';
		}
		if($show_del || $show_next)
			echo '</span>';
        	echo '
        </td>
</tr>
</table><br>';
        }

        function select_history(){

		$phist_tbl=$this->object_params['proj_history_tbl'];

                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$query="SELECT uid, state, UNIX_TIMESTAMP(date) AS date, descr FROM $phist_tbl WHERE proj_id=$proj_id AND type=".PROJ_HISTORY_MAIN." ORDER BY date DESC";
        	//echo $query;
        	if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $phist_tbl table.");

		return $result;
	}

	function add_comment(&$Project, $comment, $cc=FALSE){
		global $uid;
		global $https_path;

		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];
		$lbl=$this->project_info['label'];
		$pqs_2d=&$this->rts->pqs_2d;
		$pqid=$this->get_project_info('pqid');
		$pqname=$pqs_2d['name'][$pqid];

		$pqfn=$this->object_params['pq_fn'];
		$view_script=$Project['history_script'];

		$hcomment = "Action: Comment\n".$comment;
		$this->log_history($uid,$hcomment);

		$mail = '  '.YxlUsers::name($uid,TRUE)." has added the following comment for the project $pqname '$lbl':\n\nComment:";
		$mail .= str_replace("\n", "\n>", format_text(str_replace("\r", "", $comment), 80));
		$mail .= "\n\n  For details please use the following link:\n\n".$https_path.$view_script."?proj_id=$proj_id\n\n";

		// Read CC list
		if (!$cc)
			$cc=$this->proj_members;

		arr_unshift($cc,$this->project_info['added_by']);
		$this->rts->split_cc_list($cc,$internal_cc,$junk);

		$Header = $this->mail_headers;
		$subject=strtoupper($Project['tag']).': '.$Project['proj_title'].' comment added';
		$CC='';
		if(count($internal_cc)){
			$To=array_shift($internal_cc);
			$a=explode('"',$To);
			$int_mail  = '  Hello '.$a[1].",\n\n".$mail;

			if(count($internal_cc))
				$CC=implode(', ',$internal_cc);
			if($CC)
				$Header.= "\nCC: $CC";

			yxl_mail($To, $subject, format_text($int_mail), $Header);
		}
	}

	function log_history($uid,$descr,$check_buf=FALSE){

		$phtbl=$this->object_params['proj_history_tbl'];
		$proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];
		$state=$this->project_info['state'];

		if($check_buf && !$this->history_buf)
			return;

		$query = "INSERT INTO $phtbl (proj_id, uid, state, date, descr, type) VALUES ($proj_id, $uid, '$state', NOW(), '".escape_data($descr.$this->history_buf)."', ".PROJ_HISTORY_MAIN.")";
		//echo "$query<br>\n";
		if (!mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not insert into $phtbl table.");
	}

	function get_hlinks(){

		$phist_tbl=$this->object_params['proj_history_tbl'];

                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$query="SELECT hid, descr, uid, date FROM $phist_tbl WHERE proj_id=$proj_id AND type=".PROJ_HISTORY_HLINK." ORDER BY date";
        	//echo $query;
        	if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $phist_tbl table.");

		$hlinks=array();
		while($row=mysql_fetch_array($result,MYSQL_ASSOC))
			$hlinks[$row['hid']]=$row;

		return $hlinks;
	}

	function add_hlink($hlink){
		global $uid;

		$phist_tbl=$this->object_params['proj_history_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$query="INSERT INTO $phist_tbl (proj_id, uid, date, descr, type) VALUES ($proj_id, $uid, NOW(), '".escape_data($hlink)."', ".PROJ_HISTORY_HLINK.")";
        	//echo $query;
        	if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not insert into $phist_tbl table.");
	}

	function remove_hlink($hid){

		$phist_tbl=$this->object_params['proj_history_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_id=$this->project_info[$proj_id_fn];

		$query="DELETE FROM $phist_tbl WHERE type=".PROJ_HISTORY_HLINK." AND hid=$hid";
        	//echo $query;
        	if(!@mysql_query($query))
			_die2(__FILE__.':'.__LINE__." Could not delete from $phist_tbl table.");
	}


	function &recent_decommits(){
                $pq_fn=$this->object_params['pq_fn'];

                $hist_tbl=$this->object_params['history_tbl'];
		$hist_proj_dc=$this->object_params['history_proj_decommit_type'];

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
		$proj_pq=$this->project_info[$pq_fn];

		$drs=array();
		$ref=&$drs;

		$query="SELECT $proj_id_fn FROM $proj_tbl WHERE state='done' AND $pq_fn=$proj_pq ORDER BY $proj_id_fn DESC LIMIT 1";
        	//echo $query;
		$ldp=simple_query($query);

		$query = "SELECT h.rid FROM $hist_tbl AS h INNER JOIN $proj_tbl AS proj ON h.xid=proj.$proj_id_fn AND proj.$pq_fn=$proj_pq WHERE h.options=$hist_proj_dc AND (!(proj.state IN ('plan','done'))".($ldp?" OR proj.$proj_id_fn=$ldp":'').")";
        	//echo $query;
        	if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $hist_tbl table.");

		while($row=mysql_fetch_row($result))
			arr_push($drs,$row[0]);

		return $ref;
	}

	var $alerts=array();
	function add_alert($type,$title,$comment,$details=FALSE){
		array_push($this->alerts,array($type,$title,$comment,$details));
	}

	function show_alert($first_only=TRUE){
		global $dir_name;
		global $alert2css;

		if(count($this->alerts)<1)
			return;

		$a=array_shift($this->alerts);
		echo '
   <div width="100%" align="center">
        <div class="'.$alert2css[$a[0]].'" align="center">'.$a[1].'<hr width="30%">'.$a[2];
		if($a[3])
        		echo '<br>
		[<span class="shhd" id="alert_details_show" onClick="ShowHideToggle(1,\'alert_details\')">show details</span><span class="shhd" id="alert_details_hide" style="display: none;" onClick="ShowHideToggle(0,\'alert_details\')">hide details</span>]';

		echo '
	</div>
	<div class="alert_details" id="alert_details0" style="display:none" align="center">'.$a[3].'</div>
   </div>';
	}

	function _dashboard_active_projects($did,$uid,$info=FALSE){
		global $is_customer;
		global $dir_name;

		if($is_customer)
			return FALSE;

		if(prereq_check('pps_no_projects'))
			return FALSE;

		if($info){
			return array('header' => 'Active projects',
				'default_size' => 1,
				'type' => 'table/text',
				'edit' => array('header' => array('title' => 'Dashlet Header'),
						'rlimit' => array(
							'title' => '#projects to show',
							'options' => array(5=>5,10=>10,25=>25)),
					)
				);
		}

		$lim=5;
		if($did){
			$dp=&new DashletParams($did);
			$lim=$dp->pget('rlimit');
		}
		if(!$lim) $lim=5;

		$ctag=$this->object_params['class_tag'];

		$proj_tbl=$this->object_params['proj_tbl'];
                $proj_id_fn=$this->object_params['proj_id_fn'];
                $proj_sn=$this->object_params['proj_title'];

		$pq_tbl=$this->object_params['pq_tbl'];
		$pq_fn=$this->object_params['pq_fn'];
		$pq_sn=$this->rts->object_params['pq_title'];

		$tbl_str='';

		$query="SELECT proj.label, proj.state, p.name, proj.$proj_id_fn FROM $proj_tbl AS proj INNER JOIN $pq_tbl AS p ON proj.$pq_fn=p.$pq_fn AND !p.parent WHERE  proj.state!='done' ORDER BY proj.$proj_id_fn, proj.state LIMIT $lim";
		//echo "$query <br>\n";
		if(!($result=@mysql_query($query)))
			_die2(__FILE__.':'.__LINE__." Could not select from $proj_tbl  table.");

		if(mysql_num_rows($result)<1)
			return $tbl_str;
			$tbl_str.='
	<tr style="height: 1" class="separator">
			<td colspan=3></td>
	</tr>
	<tr class=lgray>
		<th>'.$proj_sn.' Label</th>
		<th>'.$pq_sn.'</th>
		<th>Sate</th>
	</tr>';
		while($row=mysql_fetch_array($result, MYSQL_NUM)){
			$url=$ctag.'/'.$ctag.'_track.php?proj_id='.$row[3];
			if($row[1]=='plan')
				$url=$ctag.'/'.$ctag.'_plan.php?proj_id='.$row[3];

			$tbl_str.= '
	<tr class="lgray" onmouseover="this.className=\'lgs\';" onmouseout="this.className=\'lgray\';" onClick="document.location=\''.$dir_name.'/'.$url.'\'">
		<td align="center">'.$row[0].'</td>
		<td align="center">'.$row[2].'</td>
		<td align="center">'.$row[1].'</td>
	</tr>';
		}

		if($tbl_str)
			$tbl_str= '
		<table width="100%" align="center" valign=top border="0" cellpadding="0" cellspacing="1" class="dgray">'.$tbl_str.'
		</table>';
		return $tbl_str;
	}
}

  function validate_project_params(&$Project){

	if(!isset($Project['tag']))
		_die('Parameter tag is not set.');

	if(!isset($Project['page_title']))
		$Project['page_title']='Project Management';

	if(!isset($Project['proj_title']))
		$Project['proj_title']='Project';

	if(!isset($Project['pq_title']))
		$Project['pq_title']='Product';

	if(!isset($Project['sub_pq_title']))
		$Project['sub_pq_title']='Sub-Product';

	$prefix='/'.$Project['tag'].'/'.$Project['tag'];
	if(!isset($Project['plan_script']))
		$Project['plan_script']=$prefix.'_plan.php';

	if(!isset($Project['track_script']))
		$Project['track_script']=$prefix.'_track.php';

	if(!isset($Project['done_script']))
		$Project['done_script']=$prefix.'_done.php';

	if(!isset($Project['summary_script']))
		$Project['summary_script']=$prefix.'_projects.php';

	if(!isset($Project['estimate_script']))
		$Project['estimate_script']=$prefix.'_estimate.php';

	if(!isset($Project['ts_script']))
		$Project['ts_script']=$prefix.'_ts.php';

	if(!isset($Project['history_script']))
		$Project['history_script']=$prefix.'_history.php';

	if(!isset($Project['remind_script']))
		$Project['remind_script']=$prefix.'_remind.php';

	if(!isset($Project['vote_script']))
		$Project['vote_script']=$prefix.'_mem_pri.php';

	if(!isset($Project['gantts_script']))
		$Project['gantts_script']=$prefix.'_gantts_wrapper.php';

	if(!isset($Project['bdc_script']))
		$Project['bdc_script']=$prefix.'_bdc_wrapper.php';

	if(!isset($Project['class_name']))
		_die('Parameter class_name is not set.');

	if(!isset($Project['unhide_menu']))
		$Project['unhide_menu']=array();
  }
?>
