#!/usr/local/bin/php
<?php
// $Header: /home/CVS/rps/utils/parse_mail.php,v 1.11 2008/02/16 00:16:52 yoxel Exp $
/*
 * ----START-LICENCE----
 * Copyright 2002-2008 Yoxel Systems, Santa Clara, CA, USA (info@yoxel.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * -----END-LICENCE-----
 */

// This is a parser of email comments directed to Yoxel subsystems:
// - RequestTracker (bts)
// - TickerTracker (tts)
//
// This includes users' responses/comments to emails sent by Yoxel and
// CVS/Subversion commit info/comments. The latter is the integration
// mechanism CVS/Subversion -> Yoxel.
//
// The script expects the email in a To:/CC:/BCC: field in the following form:
//	<prefix><string>@<$MyDomain> where prefix is an identifier of
// a Yoxel subsystem: 'bts', 'rts', 'tts', ...
// For example:
//	bts123@mytracker.com
//	rts_yoxel@mydomain.net
//
// The From: field is supposed to be an email that can be found in
//  Yoxel database. Although you can disable the check and accept
// emails from any address.
//
//The Subject: field tells the script which request it is related to.
// First it check for a Bugzilla style subject [Bug id1 id2 id3 ...]
// Then it simply looks for 1st numeric value preceding ':' (Yoxel style)
// In the future the script may be aware of other styles to support
// emails comments sent by snvnotify or git post-revceive hook
//
// PREREQUISITE: PEAR module Mail/mimeDecode.php
//
// INSTALLATION:
// You probably want to create a Unix account and create the following
// .forward file in its home directory:
// |<yoxel_root_dir>/utils/parse_mail.php <yoxel_root_dir>
//
// Then configure your mail system/aliases to make sure that all
// Yoxel related emails rts*@..., bts*@..., tts*@...
// are directed to this Unix account and eventually passed to the script.

// Some useful links describing CVS/Subversion Bugzilla integration:
// http://sean-janus.optionpc.com/me/software/bugtraq/
// http://www.einval.com/~steve/software/cvs-bugzilla/
//
// The same instructions can be applied to Yoxel if you use the
// parse_mail.php script instead of bugzilla_email_append.pl one.

  if($_SERVER['argc']<2)
	local_exit('<yoxel_root_dir> parameter is required.');

  $rps_dir=$_SERVER['argv'][1];
  if(!is_dir($rps_dir))
	local_exit('Directory specified by <yoxel_root_dir> does not exist.');

  // To translate email 'To:' field prefix to Yoxel subsystem name.
  $prefix2type=array('bts'=>'bts',
  		'rts'=>'bts', 'tts'=>'tts', 'kbs'=>'kbs');

  // Truncate comments longer than
  $climit=102400;

  $hide_header=TRUE;
  $RequireLogin=FALSE;
  chdir($rps_dir);
  require_once('./includes/db.inc');
  require_once 'Mail/mimeDecode.php';

  // To translate domain name of email 'From:' field to domain name
  // that Yoxel knows your email by. To accept email comments from
  // domains other than that registered in your Yoxel account.
  $translate_domain=array('example.com'=>$MyDomain,
  		'sub.example.com'=>$MyDomain);

  // In some cases a user name of a 'From:' field' needs to be translated first before
  // being chacked agains registered YoxelIDs, but this is not implemented at the moment.
  // You can disable email/YoxelID checking all together by specifying FALSE below.
  $check_from=TRUE;

  $fp = fopen("php://stdin","r");
  $input = '';
  while (!feof($fp))
	$input .= fread($fp, 8192);
  fclose($fp);

  $params=array('include_bodies'=>TRUE,
		'decode_bodies'=>TRUE,
		'decode_headers'=>TRUE);

  $decoder = new Mail_mimeDecode($input);
  $structure = $decoder->decode($params);

  $my_domain=implode('\.',explode('.',$MyDomain));
  $prefix_re=implode('|',array_keys($prefix2type));

  // Parse Header
  $sig=$un=$edn=$ct=$hide_comment=FALSE;
  $rids=array();
  foreach($structure->headers as $hk => $hdr){
	if(($hk=="to") || ($hk=="cc") || ($hk=="bcc")){
		if(preg_match("/(".$prefix_re.")(\S*)@".$my_domain."/i",$hdr,$matches)){
			$rts_type=$prefix2type[strtolower($matches[1])];
			$sig=$matches[2];
		}
	}

	if($hk=="from"){
		if(preg_match("/<\s*(\S+)@(\S+)\s*>/i",$hdr,$matches)){
			$un=$matches[1];
			$edn=$matches[2];
		}elseif(preg_match("/(\S+)@(\S+)\s*/i",$hdr,$matches)){
			$un=$matches[1];
			$edn=$matches[2];
		}
	}

	if($hk=="subject"){
		if(preg_match("/^\s*\[Bug\s+([\d\s]+)\s*\]/i",$hdr,$matches)){
                // Looks like a bugzilla-watcher style email
                        $rids=explode(' ',$matches[1]);
			$hide_comment=TRUE;
                }elseif(preg_match("/[^\d]*(\d+):/i",$hdr,$matches)){
                // Assume this is a respond/comment to Yoxel email
                        $rids=array($matches[1]);
                }
	}
  }
  //echo "type=$rts_type rid=$rids[0] from=$un@$edn\n";

  if(!$un || !$edn)
	local_exit("Could not get sender's email from email headers.");

  // Parse Body
  $body = mime_first_text_part($structure);
  if(!$body)
	local_exit('Empty comment.');
  //echo "$body\n";

  if(strlen($body)>$climit)
  	$body=substr($body,$climit);

  if($asp_mode)
	require_once('includes/asp/parse_mail_inc.php');

  if($rts_type == "bts"){
	require_once('includes/rts/bts.inc');
	$rts=new BugTracker($cid);
  }elseif($rts_type == "tts"){
	require_once('includes/itrts/tts.inc');
	$rts=new TicketTracker($cid);
  }elseif($rts_type == "kbs"){
	require_once('includes/kbs/kbs.inc');
	$rts=new KnowledgeBase($cid);
  }else
	local_exit('Unknown rts type.');


  if($check_from){
	if(isset($translate_domain[$edn]))
		$edn=$translate_domain[$edn];

		$uid = simple_query("SELECT uid FROM users WHERE email='".escape_data("$un@$edn")."' LIMIT 1");
		if(!$uid)
			local_exit("Unknown sender: $un@$edn");
  }else
  // When we do not check From: address, assume $uid=1.
  	$uid=1;

  if(count($rids)<1)
	local_exit('No request/bug IDs specified');

  if((count($rids)<2) && !$rids[0]){
	// File a new support request
	// Will add in the future
	exit(0);
  }

  foreach($rids as $rid){
  	if(!$rid)
		continue;

  	$xstr2uid=array();
  	$RInfo=$rts->request_info($rid,$xstr2uid);
  	if(!$RInfo)
		// Unknown rid
		continue;

	if($hide_comment)
		$RInfo['options']&=~(REQ_OPT_VBC|REQ_OPT_VBC2);

  	$rts->log_history($RInfo,$uid,"Action: Email comment\n\n".$body);
  }
  exit(0);

  function local_exit($msg){
  	// Sent to PHP's system logger
  	error_log($msg,0);
  	exit(0);
  }

  function mime_first_text_part(&$part){
	$ctp=$part->ctype_primary;
	if(($ctp=="text") || ($ctp=="message")){
		return $part->body;
	}elseif($l=count($part->parts)){
		for($i=0;$i<$l;$i++)
			if($body=mime_first_text_part($part->parts[$i]))
				return $body;
	}
  }
?>
